%%                           The Demo Code for 2D Tomography 
%
% threshold settings:
%
%   1) 'W1'  : Employing a stand-alone AWGN denoiser with λ = λ^{dflt} to
%              filter out degradations which are modeled as AWGN. The noise 
%              s.t.d. (flat noise root-PSD) is adaptively estimated from the
%              highest-frequency components of the spectrum of the noisy signal z_k.
%
%   2) 'C'   : Modeling degradations as stationary CORRELATED noise and
%              adaptively approximating the noise PSD from the surrogate 
%              and latent noise, which is then employed in a stand-alone 
%              filter for stationary correlated noise, with λ = λ^{dflt} . 
%
%   3) 'W1*' : The same noise s.t.d. estimation strategy as in 'W_1' but
%              employing an AWGN denoiser with a λ that may differ from λ^{dflt} .
%
%   4) 'W2*' : Modeling degradations as AWGN with the adopted noise s.t.d. of 
%              sqrt(2γ*ρ*) within a stand-alone AWGN denoiser with λ = λ^{dflt}  .
%
%   5) 'C*'  : The same noise PSD estimation strategy as in 'C' but employing
%              a (stationary correlated noise) denoiser with a λ that may
%              differ from λ^{dflt} . 
%
%  More details can be found in [*].
%
%
%            Authors :         Nasser Eslahi & Alessandro Foi
%                                   Tampere University
%        Contact info:          firstname.lastname@tuni.fi
%
%                          Release ver. 1.0  (April 2022)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Copyright (c) 2006-2022 Tampere University.
% All rights reserved.
% This work (software, material, and documentation) shall only be used for nonprofit 
% noncommercial purposes.
% Any unauthorized use of this work for commercial or for-profit purposes is prohibited.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clear;
addpath(genpath('./'));

problem          = '2D Tomography'; % 2D tomography   

%%% Loading ground-truth BrainWeb data %%% 
ImgsOrg          = load_data(problem);                                         
x                = ImgsOrg(:,:,1);  % selecting the ground-truth image

samp_strat       = 'radial';        % radial line sampling of the 2D FFT spectrum of the underlying signal.
% samp_strat     = 'lim_angle'      % limited angle sampling of 2D FFT spectrum of the underlying signal.
% samp_strat     = 'pseudo-random'  % pseudo-random sampling of 2D FFT spectrum of the underlying signal. 


snr              = inf; % SNR(dB) of the measurement (inf: noise-free measurements, otherwise noisy) 
samp_rate        = 30;  % number of radial lines for 'radial' sampling (or sampling rate for 'pseudo-random')

%%% Acquiring the Measurements
[y, M, pinvM, mask, std_eps] = Acquisition(x, samp_strat, samp_rate, snr);
%%% y          : measurement (sampled 2D FFT spectrum of x), i.e. y = M(x) + measurement_error
%%% M          : forward sampling operator
%%% pinvM      : pseudo inverse of M
%%% mask       : binary sampling mask of the 2D FFT spectrum, formed based on the sampling strategy
%%% std_eps    : standard deviation of measurement error



%%% Parameter Settings  (select/insert your desired set of parameters)
PnP_framework    = 'ADMM'; % other frameworks: 'ISTA', 'FISTA'
thr_setting      = 'C*';   % other threshold settings: 'W1', 'W1*', 'W2*', 'C'
rho              = 10;     % step-size
param            = struct; % the structure array containing some of the parameters/variables
param.lambda     = 4.6;    % thresholding parameter (scale of the root-PSD)
param.snr        = snr;    % used only in printed information
param.std_eps    = std_eps;% used for computing the PSD of back-projected measurement error ...
                           % (if not given, it will be estimated internally from y)
param.k_final    = 100;    % number of iterations for the recovery
param.visualize  = true;   % visualizing the recovered signal and the root-PSDs
param.save       = true;   % saving the results
% param.x_0      = randn(size(x)); % you can set your desired initial estimate, otherwise ...
                                   % it will be considered as param.x_0 = real(pinvM(y));
% iter_show      = sort(unique(min([1:10 10:10:100],param.k_final)));  % ...
                                    % iterations in which the recovery is illustrated                                    

                             

%%% Iterative sparse recovery
[x_rec, PSNR_out, param] = Iterative_PnP(x, samp_strat, PnP_framework, thr_setting, ...
                                                      y, M, pinvM, mask, rho, param);
                                                  
%%% if the ground-truth is not given: 
% [x_rec, PSNR_out, param] = Iterative_PnP([], samp_strat, PnP_framework, thr_setting, ...
%                                                       y, M, pinvM, mask, rho, param);