%%                    The Demo Code for Compressive Spectral Imaging 
%
% threshold settings:
%
%   1) 'W1'  : Employing a stand-alone AWGN denoiser with λ = λ^{dflt} to
%              filter out degradations which are modeled as AWGN. The noise 
%              s.t.d. (flat noise root-PSD) is adaptively estimated from the
%              highest-frequency components of the spectrum of the noisy signal z_k.
%
%   2) 'C'   : Modeling degradations as stationary CORRELATED noise and
%              adaptively approximating the noise PSD from the surrogate 
%              and latent noise, which is then employed in a stand-alone 
%              filter for stationary correlated noise, with λ = λ^{dflt} . 
%
%   3) 'W1*' : The same noise s.t.d. estimation strategy as in 'W_1' but
%              employing an AWGN denoiser with a λ that may differ from λ^{dflt} .
%
%   4) 'W2*' : Modeling degradations as AWGN with the adopted noise s.t.d. of 
%              sqrt(2γ*ρ*) within a stand-alone AWGN denoiser with λ = λ^{dflt}  .
%
%   5) 'C*'  : The same noise PSD estimation strategy as in 'C' but employing
%              a (stationary correlated noise) denoiser with a λ that may
%              differ from λ^{dflt} . 
%
%  More details can be found in [*].
%
%
%            Authors :         Nasser Eslahi & Alessandro Foi
%                                   Tampere University
%        Contact info:          firstname.lastname@tuni.fi
%
%                          Release ver. 1.0  (April 2022)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Copyright (c) 2006-2022 Tampere University.
% All rights reserved.
% This work (software, material, and documentation) shall only be used for nonprofit 
% noncommercial purposes.
% Any unauthorized use of this work for commercial or for-profit purposes is prohibited.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clear
addpath(genpath('./'));



problem          = 'Compressive Spectral Imaging'; % compressive spectral imaging 
samp_strat       = 'CSI';  % spectral multiplexing
transmittance    = 0.25;   % transmittance level of the coded aperture
idx_img          = 1;      % index of underlying multispectral image 
                               % 1: chart_and_stuffed_toy, 
                               % 2: feathers, 
                               % 3: flowers                               
file_names = {'chart_and_stuffed_toy', 'feathers', 'flowers'};

%%% loading ground-truth multispectral image %%%
x                = load_data(problem, idx_img);                  
x                = x(1:64,150:150+63,:); % using a 64x64x16 hyperspectral image for faster execution 

snr              = 25;     % SNR(dB) of the measurements

%%% Acquiring the Measurements
[y, M, pinvM, mask, std_eps, transpM] = Acquisition(x, samp_strat, transmittance, snr);
%%% y          : single-shot spectrally compressed measurements, i.e. y = M(x) + measurement_error
%%% M          : forward sampling operator
%%% pinvM      : pseudo-inverse of M
%%% mask       : 3D binary sampling mask
%%% std_eps    : standard deviation of measurement error
%%% transpM    : transpose of M


%%% Parameter Settings  (select/insert your desired set of parameters)
PnP_framework    = 'ADMM'; % other frameworks: 'ISTA', 'FISTA'
thr_setting      = 'C';    % other threshold settings: 'W1', 'W1*', 'W2*', 'C'
rho              = 5.5;    % step-size 
param            = struct; % the structure array containing some of the parameters/variables
param.snr        = snr;    % used only in printed information
param.std_eps    = std_eps;% used for computing the PSD of back-projected measurement error ...
                           % (if not given, it will be estimated internally from y)
% param.lambda     = 3.3;  % thresholding parameter (scale of the root-PSD)
param.k_final    = 50;     % number of iterations for the recovery
param.visualize  = true;  % visualizing the recovered signal and the root-PSDs
param.save       = true;   % saving the results
param.x_0        = transpM(y); % you can set your desired initial estimate, otherwise ...
                               % it will be considered as param.x_0 = real(pinvM(y));
% param.iter_show = sort(unique(min([1:10 10:10:100],param.k_final)));  % ...
                               % iterations in which the recovery is illustrated                                    
param.file_name  = file_names{idx_img}; % name of the hyperspectral image (only for illustration)
% param.Spctr_bands = 400:20:700;% Range of the wavelength of the signal x in nm (only for illustration)



%%% Iterative sparse recovery
[x_rec, PSNR_out, param] = Iterative_PnP(x, samp_strat, PnP_framework, thr_setting, ...
                                                      y, M, pinvM, mask, rho, param);
                                                  
%%% if the ground-truth is not given: 
% [x_rec, PSNR_out, param] = Iterative_PnP([], samp_strat, PnP_framework, thr_setting, ...
%                                                       y, M, pinvM, mask, rho, param);