function [x_rec, info_rec, param] = Iterative_PnP(x, samp_strat, PnP_framework, thr_setting, ...
                                                            y, M, pinvM, mask, rho, param)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Iterative_PnP performs iterative sparse signal recovery within a PnP
%  framework (e.g., ADMM, ISTA, or FISTA). The inverse imaging problems 
%  which can be tackled by this version of the code are 2D/3D Tomography 
%  and compressive spectral/temporal imaging, which are specified by their
%  corresponding sampling "samp_strat" (see the syntax of Acquisition.m 
%  for this matter). For the recovery, a family of BMxD filters (e.g., 
%  BM3D, BM4D and RF3D) are employed within PnP, which operate based on 
%  the selected threshold setting "thr_setting".
%
%   More details can be found in:
% [*] N. Eslahi and A. Foi, "Improved sparse Signal Recovery via Adaptive Correlated Noise Model",
%     submitted to IEEE TCI, 2022.
% _________________________________________________________________________________________________
% INPUTS:         |   TYPE:         |    DESCRIPTION:
% -------------------------------------------------------------------------------------------------
%   x             |  (2D/3D array)  |  Ground-truth image. This is only used to compute PSNR(dB) 
%                 |                 |  of the recovery. If the ground-truth is set to empty (i.e. 
%                 |                 |  x=[]), then mean-squared residual error (dB) is computed.
% -------------------------------------------------------------------------------------------------
%   samp_strat    |      (char)     |  Sampling strategy for each image/modality. 
%                 |        OR       |
%                 |     (string)    |                   <(APPLICATION -- BMxD)>
%                 |                 | <OPTIONS>      :  < DESCRIPTION>
%                 |                 |  ---   ---   ---   ---   ---   ---   ---
%                 |                 |                  (for "2D Tomography" -- "BM3D")
%                 |                 | 'radial'       : radial line sampling of 2D FFT spectrum. 
%                 |                 |                  
%                 |                 | 'lim_angle'    : limited angle sampling of 2D FFT spectrum.
%                 |                 |                  
%                 |                 | 'pseudo-random': pseudo-random sampling of 2D FFT spectrum. 
%                 |                 |  ---   ---   ---   ---   ---   ---   ---
%                 |                 |                  (for "3D Tomography" -- "BM4D")
%                 |                 | '3D_radial'    : 3D radial line sampling of 3D FFT spectrum.
%                 |                 |  ---   ---   ---   ---   ---   ---   --- 
%                 |                 |                  (for "Compressive Spectral Imaging" -- "BM4D")
%                 |                 | 'CSI'          : spectral multiplexing (acquiring single-shot 
%                 |                 |                  spectrally compressed measurements 
%                 |                 |                  given hyperspectral images).
%                 |                 | ---   ---   ---   ---   ---   ---   ---  
%                 |                 |                  (for "Compressive Temporal Imaging" -- "RF3D")
%                 |                 | 'CTI'          : temporal multiplexing (acquiring multi-chunk 
%                 |                 |                  temporally compressed measurements given a video).
% -------------------------------------------------------------------------------------------------
%   PnP_framework |      (char)     | The selected PnP framework for the iterative recovery
%                 |        OR       | <OPTIONS>      : < DESCRIPTION>
%                 |     (string)    | ---   ---   ---   ---   ---   ---   ---
%                 |                 | 'ADMM'         : PnP-ADMM  (alternating direction method of multipliers)
%                 |                 | 'ISTA'         : PnP-ISTA  (iterative shrinkage/thresholding algorithm)
%                 |                 | 'FISTA'        : PnP-FISTA (fast ISTA)
% -------------------------------------------------------------------------------------------------
%   thr_setting   |      (char)     | The selected threshold setting to filter out the degradations within recovery
%                 |        OR       | 
%                 |     (string)    | <OPTIONS>      : < DESCRIPTION>
%                 |                 | ---   ---   ---   ---   ---   ---   ---
%                 |                 | 'W1'           : Employing a stand-alone (AWGN) denoiser with λ = λ^{dflt}
%                 |                 |                  to filter out degradations which are modeled as "AWGN".
%                 |                 |                  The noise s.t.d. (flat noise root-PSD) is adaptively estimated 
%                 |                 |                  from the highest-frequency components of the spectrum of 
%                 |                 |                  the noisy signal.
%                 |                 | **Note: λ^{dflt} is the default value of thresholding parameter λ used by
%                 |                 | the filter, e.g., λ^{dflt}=2.7 for 8x8 blocks in BM3D. 
%                 |                 | ---   ---   ---   ---   ---   ---   ---                 
%                 |                 | 'W1*'          : Employing an AWGN denoiser with λ = λ* and the same            
%                 |                 |                  noise s.t.d. estimation strategy as in 'W1'.
%                 |                 | ---   ---   ---   ---   ---   ---   ---                    
%                 |                 | 'W2*'          : Modeling degradations as AWGN with the adopted noise s.t.d. of             
%                 |                 |                  "sqrt(2γ*ρ*)" within a stand-alone denoiser with λ = λ^{dflt}. 
%                 |                 | ---   ---   ---   ---   ---   ---   ---                
%                 |                 | 'C'            : Modeling degradations as "stationary CORRELATED noise" and
%                 |                 |                  adaptively approximating the noise PSD from the "surrogate" 
%                 |                 |                  and "latent" noise, which is then employed in a stand-alone 
%                 |                 |                 (stationary correlated noise) denoiser with λ = λ^{dflt}.
%                 |                 | ---   ---   ---   ---   ---   ---   ---                  
%                 |                 | 'C*'           : The same noise PSD estimation strategy as in 'C' but employing               
%                 |                 |                  a (stationary correlated noise) denoiser with λ = λ*.
% -------------------------------------------------------------------------------------------------
%   y             |  (2D/3D array)  |  The measurements obtained based on the sampling strategy "samp_strat".
% -------------------------------------------------------------------------------------------------                   
%   M             |(function_handle)|  The forward sampling operator 
% ------------------------------------------------------------------------------------------------- 
%   pinvM         |(function_handle)|  The pseudo-inverse of "M" 
%                 |                 |  (**Note:for more information regarding "y", "M" and "pinvM",
%                 |                 |  please see the syntax of Acquisition.m)
% ------------------------------------------------------------------------------------------------- 
%   mask          |     (double)    |   The sampling mask based on the sampling strategy "samp_strat".
%                 |        OR       |   ** Note: "mask" can be set to empty when threshold setting is 
%                 |     (logical)   |   either of W1, W1* or W2*. "mask" can also be set to empty in
%                 |                 |   '2D/3D Tomography' applications, thus it is computed internally.
% ------------------------------------------------------------------------------------------------- 
%   rho           |    (double)     |   The step-size parameter in ADMM and (F)ISTA. Please check
%                 |                 |   \rho in Eq.(3a),(3b) for ADMM, and in Eq.(5) for (F)ISTA.
% ------------------------------------------------------------------------------------------------- 
%   param         |    (struct)     |  The set of parameters used for the recovery
%                 |  .............  |  param.<OPTIONS> : < DESCRIPTION>
%                 |    (double)     |  param.lambda    : the thresholding parameter; if it is not given, 
%                 |                 |                    it will be set to the default value used by 
%                 |                 |                    the employed filter within PnP.
%                 |    (double)     |  param.std_eps   : the standard deviation of the measurement error
%                 |                 |                    (here we only have considered AWGN measurement 
%                 |                 |                    error)used for computing the PSD of its back-projection.
%                 |  (2D/3D array)  |  param.x_0       : the initial estimate; if it is not provided,
%                 |                 |                    it will be set to "real(pinvM(y))".
%                 |    (double)     |  param.k_final   : the number of iterations for recovery; if 
%                 |                 |                    it is not provided it will be set to 100.
%                 |    (double)     |  param.gamma     : the positive regularization parameter used ONLY
%                 |                 |                    under the threshold setting ''W2*'' in order to
%                 |                 |                    adopt the noise std based on a non-adaptive procedure
%                 |                 |                    as "sqrt(2γ*ρ*)/λ^{dflt}". It is ignored under 
%                 |                 |                    {'W1','W1*','C','C*'} threshold setting.
%                 |    (double)     |  param.sigma     : as an alternative to "param.gamma", one can directly 
%                 |                 |                    provide the noise std to be adopted by the filter in  
%                 |                 |                    'W2*' threshold setting.
%                 |    (logical)    |  param.visualize : visualizing the ground-truth and the recovered images,
%                 |                 |                    as well as the effective noise "Tr"-root-PSD computed
%                 |                 |                    from an oracle effective noise "e_k = z_k - x" at
%                 |                 |                    specific iterations determined in the parameter 
%                 |                 |                    "param.iter_show". It is disabled by default.
%                 |                 |                    (0:disable, 1: enable)
%                 |    (double)     |  param.iter_show : iterations in which the recovered images and root-PSD(s)                   
%                 |                 |                    are shown. Note: visualization makes the recovery slow,
%                 |                 |                    that's why we don't show for every single iteration!
%                 |    (logical)    |  param.save      : saves the recovered signal, the parameters/info and 
%                 |                 |                    PSNR(dB)/MSRE(dB) of the recovery. It is disabled by default. 
%                 |                 |                    (0:disable, 1: enable) 
%
%
%  
% _________________________________________________________________________________________________
% OUTPUTS:        |   TYPE:         |    DESCRIPTION:
% -------------------------------------------------------------------------------------------------
%   x_rec         |  (2D/3D array)  |  The recovered signal at the final iteration.
% -------------------------------------------------------------------------------------------------
%   info_rec      |    (double)     |  The PSNR(dB) or MSRE(dB) of the recovery through iterations.
% -------------------------------------------------------------------------------------------------
%   param         |    (struct)     |  The set of parameters used for the recovery
%
%
%  EXTERNAL DEPENDENCY FUNCTIONS:
%    I)   Est_effective_noise_PSD.m  (compute the approximated PSD of effective noise).
%         I.1) Est_rootPSD.m  (computes the root-PSD via local moving window).
%         I.2) FFT_PSD_Upsampling.m (upsamples a local FFT-PSD to constitute its global version).
%         I.3) FFT_PSD_domain_convertor.m (converts an FFT-PSD into a Tr-PSD).
%    II)  Est_std.m  (computes the std of an AWGN from the noisy signal).
%    III) PSD_Auxiliaries.m  (computes the root-PSD of back-projected measurement error and
%         the PSD compensation ratio).
%    IV)  Visualization.m  (an axuiliary function for visualizing the recovered signal and root-PSDs).
%    V)   BMxD filters (BM3D_colored.m , bm4d_colored.m , RF3D.m).
% ** for more information on each of the above functions, please see their syntaxes.
%
%
%
%            Authors :         Nasser Eslahi & Alessandro Foi
%                                   Tampere University
%        Contact info:          firstname.lastname@tuni.fi
%
%                          Release ver. 1.0  (April 2022)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Copyright (c) 2006-2022 Tampere University.
% All rights reserved.
% This work (software, material, and documentation) shall only be used for nonprofit 
% noncommercial purposes.
% Any unauthorized use of this work for commercial or for-profit purposes is prohibited.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% CHECKING THE NECESSARY INPUTS %%%

if nargin < 9
    error(['Necessary inputs are not given! ' ...
           'Please check the <a href="matlab:help %s">syntax</a>'],mfilename)
elseif nargin == 9
   param    = struct;   % the structure array containing some of the parameters/variables 
end

param.PnP_framework = PnP_framework;   % PnP framework
param.thr_setting   = thr_setting;     % threshold setting
param.samp_strat    = samp_strat;      % sampling strategy
  

%%%% Checking the inputs, and setting the parameters %%%%
param    = check_input_and_set_param(param, y, M, pinvM, mask, rho);


%%% compute PSNR(dB) if the ground-truth is provided, otherwise ...
%%% mean-squared residual error (dB) is computed.
if ~exist('x','var') || isempty(x)
    param.compute_PSNR = false;        % mean-squared residual error is computed
    param.assmt   = 'MSRE';            % selected objective quality assessment
    if sum(cell2mat(param.check_strat(1:5).'),1) % for 2D/3D Tomography experiments
        ind_assmt = mask~=0;           % locations which are sampled in the FFT plane
        vec       = @(X) X(ind_assmt); % extracts the elements of a vector w.r.t. the indices "ind_psnr"
    else
        vec       = @(X) X(:);         % PSNR(dB) is computed over all pixels
    end
else
    param.compute_PSNR = true;         % PSNR is computed
    param.assmt   = 'PSNR';            % selected objective quality assessment
    if sum(cell2mat(param.check_strat(1:5).'),1) % for 2D/3D Tomography experiments
        ind_assmt = x > .04*max(x(:)); % pixel indices where PSNR(dB) is computed
                                       % (to consider only the more informative parts).
        vec       = @(X) X(ind_assmt); % extracts the elements of a vector w.r.t. the indices "ind_psnr"                           
    else
        vec       = @(X) X(:);         % PSNR(dB) is computed over all pixels
    end
end
info_rec = zeros(param.k_final,1);     % PSNR(dB)/MSRE(dB) of the recovery



%%% initializing the auxiliary signals
b_k     = zeros(param.SizeX);
r_k     = zeros(param.SizeY);
x_k     = param.x_0;

if sum(cell2mat(param.check_strat(6:7).'),1)
    %%% precomputed signal to be used when updating x_k under PnP-ADMM ...
    %%% in compressive spectral/temporal imaging (see Appendix C-A of [*]). 
    v_0 = pinvM(y)./(1+1/rho); 
end
%%%%%%%%%%%%%%%%%%%%
%%% SETTING ENDS %%%
%%%%%%%%%%%%%%%%%%%%



if param.compute_PSNR
    %%% PSNR of the initial estimate "x_0"
    PSNR_init = 10*log10(max(vec(x)).^2/mean(vec(x-x_k).^2));
else
    PSNR_init = [];
end

%%% displaying the information of the understudying problem
param = verbose_init(param, PSNR_init);


% ======================================================================
% ======================================================================
%      Iterative Sparse Signal Recovery via (BMxD) PnP-ADMM/PnP-ISTA
% ======================================================================
% ======================================================================
for k = 1:param.k_final
    
    %%% updating noisy image %%%
    z_k           = x_k + b_k;
    
    if k == 1                           % The first iteration is modeled as AWGN
        sigma_e_1  = 0.2.*(max(z_k(:))-min(z_k(:))); % s.t.d. of the modeled AWGN at k=1
        rootPSD_HT = sigma_e_1.*ones(param.Tr_Sz);   % flat root-PSD
        lambda_thr = param.lambda_dflt; % default thresholding parameter at k=1
        thr_stng   = 'W';               % threshold setting (filtering mode)
    
    else
        lambda_thr = param.lambda;      % thresholding parameter
        thr_stng   = param.thr_setting; % threshold setting (selected filtering mode)
        
        %%%%% estimating/adopting noise effective root-PSD for filtering %%%%%
        if    sum(strcmpi(param.thr_setting,{'W1','W1*'})) == 1 %%% AWGN model %%%
            %%% Estimating the noise standard deviation (std) by applying ... 
            %%% the median absolute deviation (MAD) over the high-passed ...
            %%% noisy image"z_k" (see Section IV-B of [*]).
            sigma_e_k   = Est_std(z_k, param.Tr_Sz, param.HT_Tr,'MAD'); % estimated noise std
            rootPSD_HT  = sigma_e_k.*ones(param.Tr_Sz); % flat root-PSD
            
        elseif strcmpi(param.thr_setting,{'W2*'})               %%% AWGN model %%%
            %%% Adopting the noise s.t.d. based on a non-adaptive procedure ...
            %%% as "sqrt(2γ*ρ*)/λ^{dflt}" (see Section IV-A of [*]).
            if isfield(param,'sigma')
                sigma_e_k = param.sigma; % adopted noise std 
            elseif isfield(param,'gamma')
                sigma_e_k = sqrt(2*param.gamma*rho)/param.lambda_dflt; % adopted noise std
            end
            rootPSD_HT = sigma_e_k.*ones(param.Tr_Sz); % flat root-PSD

        elseif sum(strcmpi(param.thr_setting,{'C','C*'})) == 1  %%% stationary correlated noise model %%%
            %%%% Approximating the effective noise root-PSD (see Section IV-C of [*]).
            d_k         = -real(pinvM(r_k));
            [~, PSD_HT] = Est_effective_noise_PSD(param.problem, ...
                            param.HT_Tr, param.Tr_Sz, param.locWin, b_k, d_k, ...
                            param.loc_FFT_f_rate, param.loc_FFT_rootPSD_pinvM_eps);
            rootPSD_HT  = sqrt(PSD_HT);   % approximated noise root-PSD   

        end
        

        
    end
    
    if strcmpi(PnP_framework,'FISTA')
        u_k_1 = u_k;
    end
    
    %%% filtering the noisy image "z_k" %%%
    u_k = denoising(z_k, rootPSD_HT, lambda_thr, param.denoiser);
    

        
    if     strcmpi(PnP_framework,'ADMM')     %%% PnP-ADMM %%%
        %%% updating x_k based on the forward model and sampling strategy (see Appendix C of [*])
        if sum(cell2mat(param.check_strat(1:5).'),1) %% 2D/3D Tomography
            x_k   = real(ifftn( (y + fftn(u_k - b_k)./rho)./(mask.^2 + 1/rho) ));
            x_k   = max(x_k,0);              % enforcing the non-negativity constraint
            b_k   = b_k + (x_k - u_k);       % updating b_k
        else %% compressive spectral/temporal imaging
            b_k   = b_k + (x_k - u_k);       % updating b_k
            v_k   = pinvM(M(u_k - b_k));
            x_k   = v_0 + u_k - b_k - v_k./(1+1/rho); 
        end
        r_k       = y - M(x_k);              % computing residual measurements

    elseif sum(strcmpi(PnP_framework,{'ISTA','FISTA'}))     %%% PnP-(F)ISTA %%%
        if strcmpi(PnP_framework,'FISTA')
            alpha_k_1 = alpha_k;
            alpha_k   = .5 + .5*sqrt(1+4*alpha_k^2);
            t_k   = (alpha_k_1-1)/alpha_k;
            x_k   = u_k + t_k.*(u_k-u_k_1); 
        else
            x_k   = u_k;
        end
        x_k       = max(x_k,0);              % enforcing the non-negativity constraint
        r_k       = y - M(x_k);              % computing residual measurements
        b_k       = rho.*real(pinvM(r_k));   % updating b_k

    end
 
    
    
    if param.compute_PSNR
        %%% compute PSNR of the recovered signal "x_k" if the ground-truth is provided
        info_rec(k) = 10*log10(max(vec(x)).^2/mean(vec(x-x_k).^2));
    else
        %%% compute MSRE (mean-squared residual error) of "x_k" if the ground-truth is not provided
        info_rec(k) = 10*log10(mean(abs(vec(r_k)).^2));
    end

    if param.visualize ==0
        %%% displaying the PSNR(dB)/MSRE(dB) value at every iteration %%%
        if k==1
            fprintf('Iter.    %s(dB)   threshold     λ\n', param.assmt );
            fprintf('                     setting       \n');
        end
        if info_rec(k)<10
            fprintf('%3g \t %0.3f \t\t %s \t %0.1f\n', k, info_rec(k), thr_stng, lambda_thr);
        else
            fprintf('%3g \t %0.3f \t %s \t %0.1f\n', k, info_rec(k), thr_stng, lambda_thr);
        end
        
    elseif param.visualize && ismember(k, param.iter_show)
        %%% visualizing the noisy and the recovered signal (as well as the ground-truth ...
        %%% signal and the oracle effective noise root-PSD if "x" is provided) at ...
        %%% iterations specified in "param.iter_show".
        Visualization
    end
        
        
        
end

clearvars -except  x_k  info_rec  k  param





x_rec = x_k;


fprintf('============================================================= \n');
    fprintf('The recovery is done! %s = %0.3f dB\n', param.assmt , info_rec(k));
fprintf('============================================================= \n');

if param.save  %%% saving the recovered signal and the PSNR/MSRE values
    save_results(x_rec, info_rec, param)
end


return











function param = check_input_and_set_param(param, y, M, pinvM, mask, rho)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PARAMETER CHECKING AND SETTING
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%
%%% CHECK BEGINS %%%
%%%%%%%%%%%%%%%%%%%%
if ~isa(param.samp_strat,'char') || ~isa(param.samp_strat,'char')
     error(['the data type of the selected sampling strategy "samp_strat" ' ...
            'should be ''string'' or ''char'' !'])    
end

samp_modes  = {'radial', 'lim_angle', 'pseudo-random','pseudo_random',...
               '3D_radial', 'CSI', 'CTI'}; % available sampling modes (Don't touch this part ...
                                           % unless you know what you are doing!)
            
check_strat = cellfun(@(c) strcmpi(c,param.samp_strat), samp_modes, ...
                         'UniformOutput',false);

checked_idx = sum(cell2mat(check_strat.'),1);                     

if checked_idx == 0 
    error(['Your chosen sampling strategy ''%s'' does not exist.\n' ...
           'Please check the <a href="matlab:help %s">syntax of this function</a> ' ...
           'and/or the <a href="matlab:help %s">syntax of Acquisition.m</a>'], param.samp_strat, mfilename, 'Acquisition');
end

param.check_strat = check_strat;

if ~isa(param.PnP_framework,'char') || ~isa(param.PnP_framework,'char')
     error(['the data type of the selected PnP framework "PnP_framework" ' ...
            'should be ''string'' or ''char'' !'])    
end

if sum(strcmpi(param.PnP_framework,{'ADMM','ISTA','FISTA'})) == 0 
    error(['your selected PnP framework does not exist!' ...
           ' we only have ''ADMM'', ''ISTA'' and ''FISTA''. \n' ...
           'Add your ''%s'' framework by your own!'] , param.PnP_framework)
end


if ~isa(param.thr_setting,'char') || ~isa(param.thr_setting,'char')
     error(['the data type of the selected threshold setting "thr_setting" ' ...
            'should be ''string'' or ''char'' !'])    
end

if sum(strcmpi(param.thr_setting,{'W1','W1*','W2*','C','C*'})) == 0 
    error(['your selected threshold setting does not exist!' ...
           'Please check the <a href="matlab:help %s">syntax</a>'],mfilename)
end


if ~isa(y,'double')
    error('the data type of measurements "y" should be ''double'' !') 
elseif isempty(y)
    error('"y" cannot be set as empty!')
elseif isvector(y) || isscalar(y)
    error('"y" should be a 2D or 3D array!')
end


if ~isa(M, 'function_handle')
    error('the data type of forward sampling operator "M" should be ''function_handle'' !')
end


if ~isa(pinvM, 'function_handle')
    error('the data type of "pinvM" (i.e. pseudo-inverse of "M") should be ''function_handle'' !')
end


if ~(isa(mask, 'logical') || isa(mask, 'double'))
    error('the data type of "mask" should be either ''logical'' or ''double'' !')
elseif isempty(mask)
    error('"mask" cannot be set as empty!') 
elseif isvector(mask) || isscalar(mask)
    error('"mask" should be a 2D or 3D logical/double array!')    
end
param.mask = mask;


if ~isa(rho, 'double') && ~isscalar(rho) 
    error('the step-size "rho" should be a scalar of type ''double'' !')
end


if strcmpi(param.thr_setting,{'W2*'}) && ~isfield(param,'gamma')
    if ~isfield(param,'sigma')
    error('u:stuffed:it', ['the regularization parameter "gamma" ' ...
          '(i.e. a positive scalar of type ''double'') is not provided! \n' ...
          'It is required by the threshold setting ''W2*'' in order to ' ...
          'adopt the noise std based on a non-adaptive procedure ' ... 
          'as "sqrt(2γ*ρ*)/λ^{dflt}". '])
    end
end
%%%%%%%%%%%%%%%%%%%%
%%%  CHECK ENDS  %%%
%%%%%%%%%%%%%%%%%%%%



%%% INITIALIZATION AND PARAMETER SETTINGS %%%
%%%%%%%%%%%%%%%%%%%%
%%%SETTING BEGINS%%%
%%%%%%%%%%%%%%%%%%%%

%%% initial estimate
if ~isfield(param,'x_0') || isempty(param.x_0)
    param.x_0   = real(pinvM(y));  % pseudo-inverse
end


%%% size of the underlying signal
param.SizeX = size(param.x_0);
%%% size of the measurements
param.SizeY = size(y);


%%% number of iterations for the recovery
if ~isfield(param,'k_final') || isempty(param.k_final)
    param.k_final = 100;           
end



%%% visualizing the ground-truth and the recovered images, as well as ...
%%% the effective noise "Tr"-root-PSD computed from an oracle effective ...
%%% noise "e_k = z_k - x" at specific iterations determined in the ...
%%% parameter "param.iter_show".
if ~isfield(param,'visualize') || isempty(param.visualize)
    param.visualize = 0;    % (0: disable, 1:enable) 
end


if param.visualize
    %%% iterations in which the recovered images and root-PSD(s) are shown.
    %%% Note: visualization makes the recovery slow, that's why we don't ...
    %%% show for every single iteration!
    if ~isfield(param,'iter_show') || isempty(param.iter_show)
        param.iter_show = unique(min([1:5 10:5:50 60:10:90 100:50:1000 ...
                                      param.k_final], param.k_final));   
    end
end


%%% saving the final recovered signal
if ~isfield(param,'save') || isempty(param.save) 
    param.save = 0;    % (0: disable, 1:enable) 
end


%%% setting the inverse problem, denoiser for PnP, transform used in ...
%%% the filter, size of local moving window, and thresholding parameter ...
%%% based on the selected sampling strategy 'samp_strat'.
if sum(cell2mat(param.check_strat(1:4).'),1)
    param.problem  = '2D Tomography'; % selected sparse recovery problem
    param.denoiser = 'BM3D';          % selected denoiser
    param.Tr_Sz    = min(8, max(4,round(param.SizeX./16)));  % the block size in BM3D
    param.HT_Tr    = {'bior1.5','bior1.5'}; % the 2D (spatial) sparsifying transform ...
                                     % in the hard-thresholding step of BM3D
    param.locWin   = max(floor(param.SizeX./10), param.Tr_Sz);  % the size of sliding 2D window ...
                                                                % for computing local 2D FFT-PSDs 
    param.lambda_dflt = 2.7; % the default thresholding parameter in BM3D
                                                
elseif cell2mat(param.check_strat(5))
    param.problem  = '3D Tomography'; % selected sparse recovery problem
    param.denoiser = 'BM4D';          % selected denoiser
    param.Tr_Sz    = min(8, max(4,round(param.SizeX./16)));  % the cube size in BM4D
    param.HT_Tr    = {'dct','dct','dct'}; % the 3D (spatial) sparsifying transform  ...
                                   % in the hard-thresholding step of BM4D
    param.locWin   = min(4*param.Tr_Sz-1 , floor(param.SizeX./2)); % the size of sliding 3D window ...
                                                                   % for computing local 3D FFT-PSDs 
    param.lambda_dflt = 2.7; % the default thresholding parameter in BM4D     
    
elseif cell2mat(param.check_strat(6))
    param.problem  = 'Compressive Spectral Imaging';   % selected sparse recovery problem
    param.denoiser = 'BM4D';  % selected denoiser
    param.Tr_Sz    = [min(8, max(4,round(param.SizeX(1:2)./16))) param.SizeX(3)]; % the cube size in BM4D
    param.HT_Tr    = {'dct','dct','dct'}; % the 3D (2D spatial + 1D spectral) sparsifying transform  ...
                                   % in the hard-thresholding step of BM4D
    param.locWin   = [min(4*param.Tr_Sz(1:2)-1 , ceil(param.SizeX(1:2)./2)), ...
                                 floor(param.SizeX(3)*.85)]; % the size of sliding 3D window ...
                                                             % for computing local 3D FFT-PSDs 
    param.lambda_dflt = round(sqrt(2*log(prod(param.Tr_Sz))),1);  % the default thresholding parameter in BM4D                                                
                                        
else
    param.problem  = 'Compressive Temporal Imaging';   % selected sparse recovery problem
    param.denoiser = 'RF3D';  % selected denoiser
    param.Tr_Sz    = [8 8];   % the block size in RF3D (must be 8x8 !)
    param.HT_Tr    = {'dct','dct'};   % the 2D (spatial) sparsifying transform  ...
                               % in the hard-thresholding step of RF3D
    param.locWin   = min(4*param.Tr_Sz-1 , ceil(param.SizeX(1:2)./2)); % the size of sliding 2D window ...
                                                             % for computing local 2D FFT-PSDs  
    param.lambda_dflt = 2.7; % the default thresholding parameter in RF3D 
end


if ~isfield(param,'lambda')
    if sum(strcmpi(param.thr_setting,{'W1','W2*','C'})) == 1 
        param.lambda = param.lambda_dflt;
    else
        warning(['the thresholding parameter lambda (i.e. a positive scalar ', ...
                 'of type ''double'') is not given for the selected threshold setting %s. \n', ... 
                 'It is is set to the default value used by %s: param.lambda=%.01f '], ...
                 param.thr_setting, param.denoiser, param.lambda_dflt);
        param.lambda = param.lambda_dflt;
    end
end


if sum(strcmpi(param.thr_setting,{'W1','W2*','C'})) == 1 && param.lambda ~= param.lambda_dflt
    warning(['the thresholding parameter lambda for the selected threshold setting %s ', ...
        'is automatically set to the default value used by %s: param.lambda=%.01f '], ...
        param.thr_setting, param.denoiser, param.lambda_dflt);
    param.lambda = param.lambda_dflt;
    
elseif sum(strcmpi(param.thr_setting,{'W1*','C*'})) == 1 && param.lambda == param.lambda_dflt
    % the selected filter for PnP coincides with ...
    % a stand-alone filter for correlated noise if 'C*' is selected, or
    % a stand-alone AWGN denoiser if 'W1*' is selected.
    param.thr_setting = regexprep(param.thr_setting,'*','');
end
    


if ~isfield(param,'std_eps')
    %%% standard deviation of the measurement error used for ...
    %%% computing the PSD of back-projected measurement error
    param.std_eps = [];  % it is estimated internally
end



%%%%% computing PSD compensation vectors only for C and C^* settings (Appendix B) %%%%%
if sum(strcmpi(param.thr_setting,{'C','C*'})) == 1

    %%% computing PSD compensation vectors only for C and C^* settings (see Appendix B of [*]) %%%
    [param.loc_FFT_f_rate, param.loc_FFT_rootPSD_pinvM_eps] = ...
                          PSD_Auxiliaries(param.samp_strat, mask, param.locWin, param.std_eps, y);
    %%% loc_FFT_f_rate           : ratio between the local FFT-PSD compensation vectors of perpM and MdaggerM
    %%% loc_FFT_rootPSD_pinvM_eps: local FFT-root-PSD of the back-projected measurement error
    %%% For more information, please check the syntax of PSD_Auxiliaries.m
end


return













function denoise_Img = denoising(noisy_Img, rootPSD_HT, lambda_thr, denoiser)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% "denoising" takes the noisy image and performs BMxD filtering.
%
%
%  FUNCTION INTERFACE:
%   denoise_Img = denoising(noisy_Img, rootPSD_HT, lambda_thr, denoiser)
%
% ________________________________________________________________________
% INPUTS:           |  TYPE:         |    DESCRIPTION:
% ------------------------------------------------------------------------
%  noisy_image      | (2D/3D array)  | the input noisy data
% ------------------------------------------------------------------------
%  PSD_HT           | (2D/3D array)  | the noise PSD for
%                   |                | hard-thresholding stage of BMxD
% ------------------------------------------------------------------------
%  lambda_thr       | (double)       | thresholding parameter
% ------------------------------------------------------------------------
%  denoiser         |(char OR string)| selected filter
%                   |                | **right now we have listed only BMxDs
%                   |                | 'BM3D', 'BM4D', 'RF3D'
%
% ________________________________________________________________________
% OUTPUTS:          |  TYPE:         |    DESCRIPTION:
% ------------------------------------------------------------------------
%  denoise_Img      | (2D/3D array)  | the denoised data
%
%
% **Besides BMxDs, one can easily add other transform-domain filters
%   to the above-mentioned list of denoisers, and modify the call of 
%   "denoiser" in "denoising.m".   
%
%
% Note: If someone aims to add the new version of BM3D filter
% (Mäkinen, Azzari, and Foi, IEEE TIP 2020) available at: 
% https://webpages.tuni.fi/foi/GCF-BM3D/#ref_software
% to the list of available filters, should note that it takes
% the global FFT-PSD as the input, on the contrary to the older
% version of BM3D (employed here) which takes the root-PSD defined
% w.r.t. the transform used internally by BM3D, i.e. 8x8 bior1.5 in
% hard-thresholding stage. The global FFT-PSD of effective noise is
% computed by "Est_effective_noise_PSD.m", which is its first 
% output argument "glob_FFT_PSD_e_k".
% Also note that the Wiener filtering is not disabled in the new 
% version of BM3D. You can disable it by setting
% WIENER_FILTERING = int8(0); in "BM3DProfile.m"
% 
%  
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


if     strcmpi(denoiser,'BM3D') %%%% BM3D filtering stage %%%%

    min_noisy_Img = min(noisy_Img(:));
    range         =(max(noisy_Img(:)) - min_noisy_Img);
    normal_noisy_Img = (noisy_Img - min_noisy_Img)./range;     % normalizing the noisy image (required by BM3D)
    rPSD_HT       = rootPSD_HT/range;     % the normalized noise root-PSD for hard-thresholding (required by BM3D)
    rPSD_Wie      = zeros(size(rPSD_HT)); % the root-PSD for Wiener filtering stage (disabled)
    denoise_Img   = BM3D_colored(normal_noisy_Img, rPSD_HT, rPSD_Wie, lambda_thr); %%%% BM3D filtering %%%%
    denoise_Img   = (denoise_Img.*range) + min_noisy_Img; % mapping the filtered image to the pre-normalized range
    
%%%------------------------------------------------------------------------%%%    
elseif strcmpi(denoiser,'BM4D') %%%% BM4D filtering stage %%%%
    
    sigma_e_k     = 1; % because root-PSD is not normalized.
    denoise_Img   = bm4d_colored(noisy_Img, sigma_e_k, lambda_thr, rootPSD_HT); %%%% BM4D filtering %%%%

%%%------------------------------------------------------------------------%%%    
elseif strcmpi(denoiser,'RF3D') %%%% RF3D filtering stage %%%%
    
    if size(rootPSD_HT,3)==1 % AWGN modeling
        sigma_fpn = 0; % no FPN
        sigma_rnd = rootPSD_HT(end); % std of AWGN        
        PSD_FPN_nrm = zeros(size(rootPSD_HT)); % no FPN
        PSD_RND_nrm = ones(size(rootPSD_HT));  % normalized flat PSD of AWGN       
    else
        sigma_fpn = -1;  % scaling factor of the FPN root-PSD (computed internally by the filter) 
        sigma_rnd = -1;  % scaling factor of the RND root-PSD (computed internally by the filter)
% one can set the scaling factor as the following (the results can be different from the internal estimation)
%         sigma_fpn = rootPSD_HT(end,end,1);
%         sigma_rnd = rootPSD_HT(end,end,2);
        PSD_FPN_nrm = (rootPSD_HT(:,:,1)./rootPSD_HT(end,end,1)).^2; % the normalized PSD of FPN component
        PSD_RND_nrm = (rootPSD_HT(:,:,2)./rootPSD_HT(end,end,2)).^2; % the normalized PSD of RND component
    end
   
    denoise_Img   = RF3D(noisy_Img, PSD_RND_nrm, PSD_FPN_nrm, ...
                                  sigma_rnd, sigma_fpn, lambda_thr);  %%%% RF3D filtering %%%%
     
%%%------------------------------------------------------------------------%%%    
else
    error(['Your selected filter does not exit yet! ' ...
          'Please check the <a href="matlab:help denoising">syntax</a>'])
    
end

        


return
 
















function param = verbose_init(param, PSNR_init)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% verbose_init displays information on the understudying problem 
%  INPUTS:
%        param (struct)        param.snr : snr(dB) of the measurements
%                              param.problem: (string/char) '2D Tomography', ...
%                                         '3D Tomography',
%                                         'Compressive Spectral Imaging', ...
%                                         'Compressive Temporal Imaging'
%                              param.file_name: (string/char) name of the signal
%                              .
%                              .
%                              .
%        PSNR_init  (double)   the computed PSNR(dB) based on the initial
%                              estimate (if ground-truth image is given)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
    if isfield(param,'snr')
        if isinf(param.snr)
            param.txt_snr = ['The measurements are noise-free  (SNR = ' num2str(param.snr) ' dB).'];
        else
            param.txt_snr = ['The measurements are noisy (AWGN, SNR = ' num2str(param.snr) ' dB).'];
        end

    elseif ~isempty(param.std_eps)
        if param.std_eps == 0
            param.txt_snr = ['The measurements are noise-free  (std of ' ...
                                 'measurement error = ' num2str(param.std_eps) ').'];
        else
            param.txt_snr = ['The measurements are noisy (AWGN, std of ' ...
                                 'measurement error = ' num2str(param.std_eps) ').'];
        end
        
    else
        param.txt_snr = [];
    end
    
    
    if     sum(strcmpi(param.problem, {'2D Tomography','3D Tomography'}))==1
        if ~isfield(param,'file_name') || isempty(param.file_name)
            param.file_name = [];
        else
            param.file_name = [' of ' param.file_name];
        end
        fprintf(['%s.\n'...
            'Recovery%s from %0.1f%% %s sampling of the FFT spectra ' ...
            'of the signal via %s PnP-%s.\n',...
            '%s \n'],...
            param.problem, param.file_name, mean(param.mask(:))*100, ...
            strrep(param.samp_strat, '_' ,' '), param.denoiser, ...
            param.PnP_framework, param.txt_snr);
%%%------------------------------------------------------------------------%%%
    elseif strcmpi(param.problem, 'Compressive Spectral Imaging')
        if isfield(param,'file_name') && ~isempty(param.file_name)
            param.file_name = [strrep(param.file_name, '_' ,' ') ' multispectral image'];
        else
            param.file_name = 'multispectral image';
        end
        fprintf(['%s.\n' ...
            'Iterative sparse recovery of %dx%dx%d %s from \n',...
            'single-shot spectrally compressed measurements via %s PnP-%s.\n' ...
            '%s \n'],...
            param.problem, param.SizeX(1), param.SizeX(2), param.SizeX(3), ...
            param.file_name, param.denoiser, param.PnP_framework, param.txt_snr);
%%%------------------------------------------------------------------------%%%
    elseif strcmpi(param.problem, 'Compressive Temporal Imaging')
        if ~isfield(param,'file_name') || isempty(param.file_name)
            param.file_name = 'video';
        else
            param.file_name = [param.file_name ' video'];
        end
        fprintf(['%s.\n' ...
            'Iterative sparse recovery of %dx%dx%d %s from \n',...
            '%dx%dx%d temporally multiplexed measurements via %s PnP-%s.\n' ...
            '%s. \n'],...
            param.problem, param.SizeX(1), param.SizeX(2), param.SizeX(3), ...
            param.file_name, param.SizeX(1), param.SizeX(2), param.SizeY(3),...
            param.denoiser, param.PnP_framework, param.txt_snr);
    end
%%%------------------------------------------------------------------------%%%
    if param.k_final>100 && strcmpi(param.denoiser,'BM3D') && prod(param.SizeX)>512^2
        warning('This might take several minutes!')
        
    elseif param.k_final>200 && strcmpi(param.denoiser,'BM4D') && prod(param.SizeX)>=64^3
        warning('This might take several minutes! Please be more patient!')
        
    elseif param.k_final>50 && strcmpi(param.denoiser,'BM4D') && prod(param.SizeX)>=16^5
        warning('This might take several minutes! Please be more patient!')
        
    elseif param.k_final>20 && strcmpi(param.denoiser,'RF3D') && prod(param.SizeX)>=128^3
        warning('This might take several minutes! Please be patient!')
    end

    
    fprintf('============================================================= \n');
    if ~isempty(PSNR_init)
        if sum(strcmpi(param.problem, {'2D Tomography','3D Tomography'}))==1
           txt_psnr = sprintf(['   10*log10((max(x(ind))).^2/mean((x(ind)-x_k(ind)).^2)),\n'...
                                   'where the indices "ind = x > .04*max(x(:))" is to exclude\n' ...
                                   'the sizable uninformative background.']); 
        else
           txt_psnr = sprintf('    10*log10((max(x(:))).^2/mean((x(:)-x_k(:)).^2))'); 
        end
        fprintf('PSNR(dB) is computed as:\n %s \n', txt_psnr)
        fprintf('============================================================= \n');
        fprintf('Initial Reconstruction, PSNR = %0.3f dB\n', PSNR_init);
        fprintf('============================================================= \n');
    else
        txt = sprintf('where "r_k = y - M(x_k)" is the residual measurements');
        if sum(strcmpi(param.problem, {'2D Tomography','3D Tomography'}))==1
           txt_msre = sprintf(['   10*log10(mean(abs(r_k(ind)).^2)),\n'...
                                   '%s, and \nthe indices "ind" indicates the locations ' ...
                                   'of the FFT planes which are sampled.'],txt); 
        else

           txt_msre = sprintf('    10*log10(mean(abs(r_k(:)).^2)) \n %s.', txt); 
        end
        fprintf('MSRE(dB) is computed as:\n %s \n', txt_msre)
        fprintf('============================================================= \n');
    end
    


return









function save_results(x_rec, info, param)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% save_results saves the recovered signal and the PSNR(dB)/MSRE(dB) of 
% all iterations in a folder called "Results".
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if ~isfolder('Results')
    mkdir Results
end

if (isfield(param,'snr')&& isinf(param.snr)) || (~isempty(param.std_eps) && param.std_eps == 0)
    txt = 'noise-free';
else
    if     isfield(param,'snr')     && ~isempty(param.snr)
        txt = sprintf('noisy(AWGN,SNR=%sdB)',num2str(param.snr));
    elseif isfield(param,'std_eps') && ~isempty(param.std_eps)
        txt = sprintf('noisy(AWGN,std_eps=%sdB)',num2str(param.std_eps));
    else
        txt = 'noisy(AWGN)';
    end
end

getname = @(x) inputname(1);
save_dir = sprintf('./Results/%s__%s_PnP_%s__%s__%s.mat',...
        strrep(param.problem, ' ' ,'_'), param.denoiser, param.PnP_framework, param.thr_setting, txt);
if param.compute_PSNR
    PSNR_out = info;
    save(save_dir, getname(x_rec), getname(PSNR_out), getname(param))
else
    MSRE_out = info;
    save(save_dir, getname(x_rec), getname(MSRE_out), getname(param))
end

fprintf('The results are saved in %s \n', save_dir)
fprintf('============================================================= \n');
return