%%                          Figure 6 -- Compressive Temporal Imaging
%
% This code reproduces Figure 6 of the paper:
% [*] N. Eslahi and A. Foi, Improved sparse Signal Recovery via Adaptive Correlated Noise Model,
%     submitted to IEEE TCI, 2022.
%
% It performs compressive video recovery from noise-free/noisy temporally
% multiplexed measurement(with AWGN measurement error) under additive
% stationary white/colored(correlated) noise modeling of degradations
% within recovery -- i.e. the noise to be filtered at each iteration.
% For the recovery, RF3D filter is employed as the denoiser within PnP-ADMM and PnP-ISTA.
%
%
% threshold settings:
%
%   1) 'W1'  : Employing a stand-alone (AWGN) denoiser with λ = λ^{dflt} = 2.7 to
%              filter out degradations which are modeled as AWGN. The noise 
%              s.t.d. (flat noise root-PSD) is adaptively estimated from the
%              highest-frequency components of the spectrum of the noisy signal z_k.
%                                                                                                                                  
%   2) 'C'   : Modeling degradations as stationary CORRELATED noise and
%              adaptively approximating the noise PSD from the surrogate 
%              and latent noise, which is then employed in a stand-alone 
%              (stationary correlated noise) denoiser with λ = λ^{dflt} = 2.7 . 
%
%   3) 'W1*' : The same noise s.t.d. estimation strategy as in 'W_1' but
%                     employing an AWGN denoiser with λ = λ* >= λ^{dflt} .
%
%   4) 'W2*' : Modeling degradations as AWGN with the adopted noise s.t.d. of 
%              sqrt(2γ*ρ*) within a denoiser with λ = λ^{dflt} = 2.7 .
%
%   5) 'C*'  : The same noise PSD estimation strategy as in 'C' but employing
%              a (stationary correlated noise) denoiser with λ = λ* >= λ^{dflt} . 
%
%  More details can be found in [*].
%
%
%            Authors :         Nasser Eslahi & Alessandro Foi
%                                   Tampere University
%        Contact info:          firstname.lastname@tuni.fi
%
%                          Release ver. 1.0  (April 2022)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Copyright (c) 2006-2022 Tampere University.
% All rights reserved.
% This work (software, material, and documentation) shall only be used for nonprofit 
% noncommercial purposes.
% Any unauthorized use of this work for commercial or for-profit purposes is prohibited.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clear;
addpath(genpath('./'));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%DO NOT TOUCH THIS PART!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
param              = struct;             % the structure array containing some of the parameters/variables
problem            = 'Compressive Temporal Imaging'; % compressive video recovery
samp_strat         = 'CTI';              % temporal multiplexing
transmittance      = 0.5;                % transmittance level of the coded aperture
SNRs               = [inf, 15];          % noise-free and noisy measurement cases
PnP_frameworks     = {'ADMM', 'ISTA'};   % PnP frameworks for recovery  
threshold_settings = {'W1';'C';...
                      'W1*';'W2*';'C*'}; % the threshold setting (see the above syntax)                      
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

param.visualize    = 0;                  % visualizing a frame of the ground-truth and the recovered
                                         % video, together with the effective noise 8x8 DCT-root-PSDs 
                                         % computed from an oracle (e_k = z_k - x), and the estimated
                                         % noise root-PSD or the estimated noise std at iterations
                                         % specified by param.iter_show.
                                         % (0 or false: disable, 1 or true: enable) 
                                
param.iter_show    = [1:10, 15:5:95, ...
                         100:50:1000];   % iterations where visualization is performed.                           
                                
param.save         = 0;                  % saves the recovered signal, the parameters and the PSNR(dB) 
                                         % of the recovery in a separate folder of Results under 
                                         % a specific name for each experiment.
                                         % (0 or false: disable, 1 or true: enable) 

param.k_final      = 1000;               % number of iterations (1000 is more than necessary!). we left
                                         % it to 1000 iterations to produce the numbers in Table II.
                                         % One can set a smaller number of iterations, probably costing
                                         % a slightly number.
 
                                         
warning('u:stuffed:it',['Running the whole experiment on A SINGLE MACHINE for 1000 iterations ' ...
         '(i.e. the current format) can take SEVERAL DAYS to be completed!\n' ...
         'We Strongly recommend to run it on a CLUSTER for a multitude of tasks!\n']);
pause(5)



% =====================================================
%             Loading the Ground-truth
% =====================================================                                                      
%%% loading ground-truth video %%%  
x        = load_data(problem); 
chunkSz  = 8; % chunk size for temporal multiplexing (temporal compression ratio: 1/chunkSz)


%%% PSNRs of all recovered images at every iteration
PSNR_all = zeros(param.k_final, numel(threshold_settings),...
                 numel(SNRs), numel(PnP_frameworks)); 
%%% thresholding parameters of all recovery scenarios
Lambdas  = zeros(numel(threshold_settings), numel(SNRs), numel(PnP_frameworks));             


%%% index of SNR, [1-2]: [inf, 15] (dB)
for idx_snr       = 1:numel(SNRs) 
    param.snr     = SNRs(idx_snr); % selected SNR(dB) of the measurement
    
        
%%% index of recovery framework, {1-2}: {'ADMM', 'ISTA'}
for idx_framework = 1:numel(PnP_frameworks)      
    PnP_framework = PnP_frameworks{1,idx_framework}; % selected PnP framework
    
    
%%% index of threshold setting, % {1-5}: {'W1'; 'C'; 'W1*'; 'W2*'; 'C*'} (see the above syntax)    
for idx_thr_set   = 1:numel(threshold_settings) 
    thr_setting   = threshold_settings{idx_thr_set,1};  % selected threshold setting                         


% =====================================================
%                 Parameter Settings 
% =====================================================
load_params  %%% Loading parameters of the recovery %%% 

Lambdas(idx_thr_set, idx_snr, idx_framework) = param.lambda; % thresholding parameters                                                        


% =====================================================
%              Acquiring the Measurements
% =====================================================
[y, M, pinvM, mask, param.std_eps] = Acquisition(x, samp_strat, [1/chunkSz transmittance], param.snr);
%%% y          : temporally compressed measurements, i.e. y = M(x) + measurement_error
%%% M          : forward sampling operator
%%% pinvM      : pseudo-inverse of M
%%% mask       : 3D binary sampling mask
%%% std_eps    : standard deviation of measurement error


% =====================================================
% Iterative Sparse Recovery via BM4D PnP-ADMM/PnP-ISTA
% =====================================================
[x_rec, PSNR_out] = Iterative_PnP(x, samp_strat, PnP_framework, thr_setting, ...
                                  y, M, pinvM, mask, rho, param);
                            
                              
  PSNR_all(:, idx_thr_set, idx_snr, idx_framework) = PSNR_out;  
                              

end %%end idx_thr_set

end %%end idx_framework

end %%end idx_snr

clear idx_img idx_thr_set idx_framework idx_snr ...
      PSNR_out PnP_framework rad_line thr_setting

 
  
%%% illustrating the PSNR (dB) progression vs. iteration number ...
%%% corresponding to the results of Figure 6 in [*]
figure(777)
cnt = 0;
plt = {':m','k','--r','.-b','--k'};
legend_txt = {'$\textrm{W}_1$','$\textrm{C}$','$\textrm{W}_1^*$',...
    '$\textrm{W}_2^*$','$\textrm{C}^*$'};
for idx_framework = 1:numel(PnP_frameworks)
for idx_snr       = 1:numel(SNRs)
    cnt = cnt+1;
    subplot(3,2,cnt)
    for idx_thr_set   = 1:numel(threshold_settings)
        plot(squeeze(PSNR_all(:,idx_thr_set, idx_snr, idx_framework)),plt{idx_thr_set}); hold on
    end
    xlabel('iteration number k')
    ylabel('PSNR(dB)')
    title(sprintf('RF3D PnP-%s, (SNR=%d dB)',PnP_frameworks{idx_framework},SNRs(idx_snr)))
end
end
subplot(3,2,5:6),plot(1:param.k_final, nan), axis off
legend(legend_txt,...
    'FontSize',12,'interpreter','latex',...
    'Location','SouthOutside','Orientation','horizontal','Box','on')
hold off

clear PSNRs idx_framework idx_snr idx_thr_set plt cnt legend_txt