%%                          Table III -- 3D Tomography 
%
% This code reproduces Table III of the paper:
% [*] N. Eslahi and A. Foi, Improved sparse Signal Recovery via Adaptive Correlated Noise Model,
%     submitted to IEEE TCI, 2022.
%
% It performs the recovery of 3D BrainWeb phantom from incomplete noise-free and noisy measurements 
% (with AWGN measurement error) under additive stationary white/colored(correlated) noise modeling of 
% degradations during the recovery -- the noise to be filtered at each iteration. The measurements
% are acquired under 13.6% and 27.7% 3D radial line sampling of the FFT spectra of 3D BrainWeb phantom.
% For the recovery, BM4D filter is employed as the denoiser within PnP-ADMM and PnP-ISTA.
%
%
% threshold settings:
%
%   1) 'W1'  : Employing a stand-alone (AWGN) denoiser with λ = λ^{dflt} = 2.7 to
%              filter out degradations which are modeled as AWGN. The noise 
%              s.t.d. (flat noise root-PSD) is adaptively estimated from the
%              highest-frequency components of the spectrum of the noisy signal z_k.
%                                                                                                                                  
%   2) 'C'   : Modeling degradations as stationary CORRELATED noise and
%              adaptively approximating the noise PSD from the surrogate 
%              and latent noise, which is then employed in a stand-alone 
%              (stationary correlated noise) denoiser with λ = λ^{dflt} = 2.7 . 
%
%   3) 'W1*' : The same noise s.t.d. estimation strategy as in 'W_1' but
%                     employing an AWGN denoiser with λ = λ* >= λ^{dflt} .
%
%   4) 'W2*' : Modeling degradations as AWGN with the adopted noise s.t.d. of 
%              sqrt(2γ*ρ*) within a denoiser with λ = λ^{dflt} = 2.7 .
%
%   5) 'C*'  : The same noise PSD estimation strategy as in 'C' but employing
%              a (stationary correlated noise) denoiser with λ = λ* >= λ^{dflt} . 
%
%  More details can be found in [*].
%
%
%            Authors :         Nasser Eslahi & Alessandro Foi
%                                   Tampere University
%        Contact info:          firstname.lastname@tuni.fi
%
%                          Release ver. 1.0  (April 2022)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Copyright (c) 2006-2022 Tampere University.
% All rights reserved.
% This work (software, material, and documentation) shall only be used for nonprofit 
% noncommercial purposes.
% Any unauthorized use of this work for commercial or for-profit purposes is prohibited.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clear
addpath(genpath('./'));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%% DO NOT TOUCH THIS PART! %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
param              = struct;             % the structure array containing some of the parameters/variables
problem            = '3D Tomography';    % 3D tomography experiments
samp_strat         = '3D_radial';        % 3D radial line sampling of the 3D FFT spectrum of the underlying signal
samp_coverages     = [15 , 30];          % 3D radial lines sampling coverages (%)
SNRs               = [inf, 20];          % noise-free and noisy measurement cases
PnP_frameworks     = {'ADMM', 'ISTA'};   % PnP frameworks for recovery  
threshold_settings = {'W1';'C';...
                      'W1*';'W2*';'C*'}; % the threshold setting (see the above syntax)                    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

param.visualize    = 0;                  % visualizing the ground-truth and the recovered images, 
                                         % the progression of PSNR over iterations, as well as 
                                         % the effective noise 4x4x4 DCT-root-PSD computed from
                                         % an oracle (e_k = z_k - x) and the estimated noise  
                                         % root-PSD or the estimated noise std at iterations
                                         % k = [1:10, 15:5:95, 100:50:1000] 
                                         % (0 or false: disable, 1 or true: enable) 
                                
param.iter_show    = [1:10, 15:5:95, ...
                         100:50:1000];   % iterations where visualization is performed.                           
                                
param.save         = 0;                  % saves the recovered signal and the PSNR(dB) of the recovery
                                         % in a separate folder of Results under a specific name for
                                         % each experiment.
                                         % (0 or false: disable, 1 or true: enable) 

param.k_final      = 1000;               % number of iterations (1000 is more than necessary!), we left
                                         % it to 1000 iterations to produce the numbers in Table II.
                                       
 
                                         
warning('u:stuffed:it',['Running the whole experiment on a single machine for ' ...
         '1000 iterations (i.e. the current format) can take HOURS ' ...
         '(perhaps one to three DAYS) to be completed!\n ' ...
         'We recommend to run it on a cluster for a multitude of tasks!']);

                                                    
% =====================================================
%             Loading the Ground-truth
% =====================================================                                                  

x               = load_data(problem);   %%% Loading ground-truth 3D BrainWeb phantom %%% 
%%% name of image (optional -- only for illustration)
param.file_name = '3D brainWeb Phantom'; 
    

%%% PSNRs of all recovered images at every iteration
PSNR_all = zeros(param.k_final, numel(threshold_settings),...
                 numel(samp_coverages),numel(SNRs), numel(PnP_frameworks)); 

%%% thresholding parameters of all recovery scenarios
Lambdas  = zeros(numel(threshold_settings), numel(samp_coverages), ...
                 numel(SNRs), numel(PnP_frameworks));  
             
             
%%% index of SNR, [1-2]: [inf, 20] (dB)
for idx_snr       = 1:numel(SNRs) 
    param.snr     = SNRs(idx_snr); % selected SNR(dB) of the measurement
    
    
%%% index of radial line sampling, [1-2]: [15, 30] radial line sampling    
for idx_samp_cvg  = 1:numel(samp_coverages) 
    samp_cvg      = samp_coverages(idx_samp_cvg); % selected number of radial lines for sampling
 
    
%%% index of recovery framework, {1-2}: {'ADMM', 'ISTA'}
for idx_framework = 1:numel(PnP_frameworks)      
    PnP_framework = PnP_frameworks{1,idx_framework}; % selected PnP framework
    
    
%%% index of threshold setting, % {1-5}: {'W1'; 'C'; 'W1*'; 'W2*'; 'C*'} (see the above syntax)    
for idx_thr_set   = 1:numel(threshold_settings) 
    thr_setting   = threshold_settings{idx_thr_set,1};  % selected threshold setting                         

    
% =====================================================
%                 Parameter Settings 
% =====================================================
load_params  %%% Loading parameters of the recovery %%%  
                                                         
Lambdas(idx_thr_set, idx_samp_cvg, idx_snr, idx_framework) = param.lambda;   % thresholding parameters  

% =====================================================
%              Acquiring the Measurements
% =====================================================
[y, M, pinvM, mask, param.std_eps] = Acquisition(x, samp_strat, samp_cvg, param.snr);
%%% y             : measurement (sampled 3D FFT spectrum of x), i.e. y = M(x) + measurement_error
%%% M             : forward sampling operator of 3D FFT spectrum
%%% pinvM         : pseudo inverse of M
%%% mask          : binary sampling mask of the 3D FFT spectrum, formed based on the samp_strat
%%% std_eps       : standard deviation of measurement error
        
    
% =====================================================
% Iterative Sparse Recovery via BM4D PnP-ADMM/PnP-ISTA
% =====================================================
[x_rec, PSNR_out] = Iterative_PnP(x, samp_strat, PnP_framework, thr_setting, ...
                                  y, M, pinvM, mask, rho, param);
                            
                              
  PSNR_all(:, idx_thr_set, idx_samp_cvg, idx_snr, idx_framework) ...
           = PSNR_out;
                              

end %%end idx_thr_set

end %%end idx_framework

end %%end idx_samp_cvg

end %%end idx_snr

clear idx_img idx_thr_set idx_framework idx_samp_cvg idx_snr ...
      PSNR_out PnP_framework samp_coverages thr_setting
  
  

  
%%% displaying the PSNR (dB) of the recovered 3D BrainWeb MR phantom ...
%%% corresponding to the results of Table III in [*]
PSNR_avgs = round(squeeze(PSNR_all(end,:,:,:,:)),1);

for idx_framework = 1:numel(PnP_frameworks)
for idx_samp_cvg  = 1:numel(samp_coverages)  
 
     txt_snr   = {'SNR_inf' 'SNR_20dB'};
    lambdas = squeeze(Lambdas(:,idx_samp_cvg,:,idx_framework));
    
    PSNRs    = squeeze(PSNR_avgs(:,idx_samp_cvg,:,idx_framework));
    
    Table_lmbds = array2table(lambdas, 'VariableNames',txt_snr,...
                    'RowNames',threshold_settings);
    Table_psnrs = array2table(PSNRs,    'VariableNames',txt_snr,...
                    'RowNames',threshold_settings);
    
    fprintf('Table III: Recovery via BM4D PnP-%s (%d radial lines)\n',...
                       PnP_frameworks{1,idx_framework}, ...
                       samp_coverages(idx_samp_cvg));
    
    disp(table(Table_lmbds, Table_psnrs,'VariableNames',{'lambda','PSNR_dB'}));  

end
end
clear PSNRs lambdas PSNR_avgs Table_lmbds Table_psnrs txt_snr ...
      idx_framework idx_samp_cvg

  