function [y, M, pinvM, D, std_eps, transpM, x_backproj] = Acquisition(x, samp_strat, samp_rate, snr, visualize)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
% Acquisition computes the (compressed) measurements given the signal,  
%  sampling strategy, sampling ratio, and SNR of the measurements.
%
%
%  FUNCTION INTERFACE:
%   [y, M, pinvM, D, std_eps, transpM, x_backproj] = Acquisition(x, samp_strat, samp_rate, snr)
%
%  *** In case of multimodal / multi-sensor / multi-acquisition, the inputs can be given as cells ...
%      and the outputs will be in "cell" format. Otherwise, the inputs and outputs can be given in 
%      "double" format. ***
% ___________________________________________________________________________________________________________
% INPUTS:       |  TYPE: (if not "cell",     |    DESCRIPTION:
%               |         then "double")     |
% -----------------------------------------------------------------------------------------------------------
%   x:          |(1xI cell,                  |  each cell contains an image.
%               | a double within each cell) |  "I" is the number of images/modalities.
% -----------------------------------------------------------------------------------------------------------
%   samp_strat: |(1xI cell,                  |  sampling strategy (string) for each image/modality.
%               | a string within each cell) |  Options: 
%               |                            |           'radial'       : radial line sampling of 
%               |                            |                            2D FFT spectrum (for 2D images).
%               |                            |           'lim_angle'    : limited angle sampling of
%               |                            |                            2D FFT spectrum (for 2D images).
%               |                            |           'pseudo_random' OR
%               |                            |           'pseudo-random': pseudo random sampling of
%               |                            |                            2D FFT spectrum (for 2D images).
%               |                            |           '3D_radial'    : 3D radial line sampling of 
%               |                            |                            3D FFT spectrum 
%               |                            |                            (for volumetric 3D images).
%               |                            |           'CSI'          : spectral multiplexing
%               |                            |                            (To obtain single-shot 
%               |                            |                             spectrally compressed measurements 
%               |                            |                             given hyperspectral images).
%               |                            |           'CTI'          : temporal multiplexing
%               |                            |                            (To obtain multi-chunk 
%               |                            |                             temporally compressed measurements 
%               |                            |                             given a video).
% -----------------------------------------------------------------------------------------------------------              
%   samp_rate:  |(1xI cell,                  |  sampling ratio
%               | a double within each cell) |  Options: 
%               |                            |           an integer for the number of radial lines, or
%               |                            |           percentage of random sampling
% -----------------------------------------------------------------------------------------------------------
%   snr:        |(1xI cell,                  |  SNR (dB) of the measurements (injecting AWGN).
%               | a double within each cell) |  if snr = inf, then measurements are noise-free. 
% -----------------------------------------------------------------------------------------------------------
%   visualize:  | (logical)                  |  visualization of the ground-truth and back-projected 
%               |                            |  images as well as the sampling mask.  
%               |                            |  options:   1 or true : visualize
%               |                            |             0 or false: no visualization (default)
% 
%
% ___________________________________________________________________________________________________________
% OUTPUTS:      |  TYPE: (if not "cell",     |    DESCRIPTION:
%               |         then "double")     |
% -----------------------------------------------------------------------------------------------------------
%   y:          |(1xI cell,                  |  measurements obtained based on the sampling strategy
%               | a double within each cell) |  "samp_strat".
% -----------------------------------------------------------------------------------------------------------
%   M:          |(1xI cell, a function_handle|  forward sampling operator
%               | within each cell)          |  
% -----------------------------------------------------------------------------------------------------------
%   pinvM:      |(1xI cell, a function_handle|  backward sampling operator
%               | within each cell)          |  (pseudo-inverse of M)
% -----------------------------------------------------------------------------------------------------------
%   D:          |(1xI cell,                  |  the sampling mask based on the sampling strategy "samp_strat".
%               | a double within each cell) |  
% -----------------------------------------------------------------------------------------------------------
%   std_eps:    |(1xI cell,                  |  standard deviation of the measurement error.
%               | a double within each cell) |  
% -----------------------------------------------------------------------------------------------------------
%   transpM:    |(1xI cell, a function_handle|  transpose of the sampling operator M
%               | within each cell)          |  
% -----------------------------------------------------------------------------------------------------------   
%   x_backproj: |(1xI cell,                  |  back-projected measurements (i.e. "x_backproj=pinvM(y)" )
%               | a double within each cell) |  
%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Copyright (c) 2006-2022 Tampere University.
% All rights reserved.
% This work (software, material, and documentation) shall only
% be used for nonprofit noncommercial purposes.
% Any unauthorized use of this work for commercial or for-profit purposes
% is prohibited.
%
% AUTHORS:
%                      Nasser Eslahi and Alessandro Foi
%     email:             {firstname.lastname}@tuni.fi
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if nargin < 4
    error(['Necessary inputs are not given! ' ...
           'Please check the <a href="matlab:help Acquisition">syntax</a>'])
end

if ~exist('snr','var')||isempty(snr)
    snr = inf;
end

if ~exist('visualize','var')||isempty(visualize)
    visualize = 0;
end



trajectories = {'radial', 'lim_angle', 'pseudo-random', ...
                         'pseudo_random', '3D_radial', 'CSI', 'CTI'}; % available sampling modes
                     
check_strat  = cellfun(@(c) strcmpi(c,samp_strat), trajectories, ...
                         'UniformOutput',false);

checked_idx  = sum(cell2mat(check_strat.'),1);


if (iscell(x)==0 && checked_idx==0) 
    error(['Your chosen sampling strategy "%s" does not exist.\n' ...
           'Please check the <a href="matlab:help Acquisition">syntax</a>'], samp_strat);

elseif (iscell(x) && any(checked_idx==0) )
    if sum(checked_idx==0)>1;  plural=1; else, plural=0; end

    error(['Your chosen sampling %s "%s" %s not exist.' ...
        'Please check the <a href="matlab:help Acquisition">syntax</a>'],...
        ['strateg', repmat('y',1-plural), repmat('ies',plural)],...
        strjoin(samp_strat(1,check_strat==0),', '),...
        ['do', repmat('es',1-plural), repmat('es',plural)]);
end
                  
                    


if iscell(x)        %%% multimodal, multi-sensor or multi-acquisition cases
    
    I          = size(snr,2);   % number of modalities
    M          = cell(1,I);     % forward sampling operator
    pinvM      = M;             % pseudo-inverse of the forward sampling operator M
    transpM    = M;             % transpose of the forward sampling operator M
    D          = M;             % sampling mask
    y          = M;             % measurement (sampled)
    x_backproj = M;             % back-projected signal
    std_eps    = M;             % standard deviation of the measurement error

    for i = 1:I
        
        %%% defining the sampling operator #i
        [M{i}, pinvM{i}, transpM{i}, D{i}, y{i}] = Sampling(x{i}, samp_rate{i}, samp_strat{i});
        %%% y{i}          = M{i}(x{i});
        

        %%% injecting AWGN to the measurement #i
        if ismember(samp_strat{i}, {'CSI', 'CTI'})
            noise_inj = true(size(D{i}));
        else
            noise_inj = D{i}~=0;
        end
        [y{i}(noise_inj), std_eps{i}] = Additive_WGN(y{i}(noise_inj),snr{i});
        
        
        %%% back-projection (inverse FFT)
        x_backproj{i} = pinvM{i}(y{i});
        
    end

    
else                %%% monomodal or single-acquisition case
    
        %%% defining the sampling operator 
        [M, pinvM, transpM, D, y] = Sampling(x, samp_rate, samp_strat);
        %%% y         = M(x);
        
        
        %%% injecting AWGN to the measurement
        if ismember(samp_strat, {'CSI', 'CTI'})
            noise_inj = true(size(y));
        else
            noise_inj = D~=0;
        end
        [y(noise_inj), std_eps] = Additive_WGN(y(noise_inj), snr);
        
        
        %%% back-projection (inverse FFT)
        x_backproj = pinvM(y);
 
end


if visualize  %%% visualizing the ground-truth image(s), back-projected image(s), and sampling mask(s)
    Visualizer(x, y, x_backproj, D, samp_strat, samp_rate, snr)
end


return






%%
%%
                %% defines sampling and acquire measurements
%%
%%


function [M, pinvM, transpM, S, y] = Sampling(x, smpl, trajectory, seed)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  FUNCTION INTERFACE:
%   [M, pinvM, transpM, S, y] = Sampling(x, smpl, trajectory, seed)
%
% INPUTS:
%   x:                (double)           input 2D/3D image.
%
%   smpl:             (double)           radial line numbers in the case of pseudo-radial line 
%                                        or limited-angle sampling; 
%                                        or measurement ratio in the case of pseudo-random sampling.
%
%   trajectory:       (string)           'radial' : radial line sampling
%                                        'lim_angle': limited angle sampling 
%                                        'pseudo_random': pseudo random sampling
%                                        '3D_radial' : 3D radial line sampling (for volumetric images).
%
% OUTPUTS:
%   M:                (function_handle)  forward operator (sampling in FFT domain).
%   pinvM:            (function_handle)  pseudo-inverse of the operator "M".
%   transpM:          (function_handle)  transpose of the operator "M".
%   S:                (logic)            binary sampling indicator.
%   y:                (double)           measurement vector "y = M(x)"
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


if nargin == 3
    seed  = 0;
end

if     strcmpi(trajectory,'radial')==1 ||...
       strcmpi(trajectory,'lim_angle')==1      % radial line or limited angel sampling
    [M, pinvM, transpM, S] =  Radial_Line(x, smpl, seed, trajectory );
    
elseif strcmpi(trajectory,'pseudo_random')==1 ||...
        strcmpi(trajectory,'pseudo-random')==1 % pseudo-random sampling
    [M, pinvM, transpM, S] =  Pseudo_Random(x, smpl, seed);
    
elseif strcmpi(trajectory,'3D_radial')==1      % 3D radial line sampling
    [M, pinvM, transpM, S] =  Volumetric_Radial_Line(x, smpl, seed );
    
elseif strcmpi(trajectory,'CSI')==1     % compressive spectral sampling
    [M, pinvM, transpM, S] =  Compressive_Spectral_Sampling(x, smpl, seed);
    
elseif strcmpi(trajectory,'CTI')==1     % compressive temporal sampling
    [M, pinvM, transpM, S] =  Compressive_Temporal_Sampling(x, smpl, seed);   
    
else
    error('check your sampling trajectory')
end

y = M(x);

return;

          

function [M, pinvM, transpM, S] =  Radial_Line( im, L, seed, Trajectory )
rng(seed,'v4');
[height, width]  =    size(im);

if strcmpi(Trajectory,'radial')==1
    aperture  = (pi/180)*180;   % aperture encompassing all scanning angles (aperture<pi is limited angle)
    direction = (pi/180)*0;     % direction of the scanning beam (middle axis)
else
    aperture  = (pi/180)*90;    % aperture encompassing all scanning angles (aperture<pi is limited angle)
    direction = (pi/180)*45;    % direction of the scanning beam (middle axis)    
end
S       =    LineMaskLimitedAngle(L,height,width,aperture,direction,seed); % sampling mask
M       =    @(z) S.*fft2(z);          % forward sampling operator (of the 2D FFT spectrum)
pinvM   =    @(z) ifft2(z);            % pseudo-inverse of the operator M
transpM =    @(z) numel(z)*ifft2(z);   % transpose of the operator M
% pinvM    = @(z) iff2(z./(S+eps));    % when 'S' is not binary
% transp   = @(z) numel(z)*iff2(S.*z); % when 'S' is not binary
return



function S = LineMaskLimitedAngle(LineNum, height, width, aperture, direction, seed)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% LineMaskLimitedAngle.m
%
% Returns the indicator of the domain in 2D Fourier space for the
% specified line geometry.
% FUNCTION INTERFACE:   
%       S = LineMaskLimitedAngle(LineNum, height, width, aperture, direction, seed)
%
% INPUTS:
% LineNum   : (double)  number of lines
% height    : (double)  height of the output mask
% width     : (double)  width of the output mask
% aperture  : (double)  aperture of fan
% direction : (double)  central direction of fan
%
% OUTPUTS:
% S         : (double)  2D radial line sampling mask of size height x width
%
% Alessandro Foi, Tampere University of Technology (2006-2011)
%
% Based on the original file LineMask.m written by Justin Romberg (rev. 12/2/2004) 
% as included in the L1magic package (http://www.acm.caltech.edu/l1magic/)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

rng(seed,'v4');
if nargin<3
    aperture=pi;
end
if nargin<4
    direction=0;
end

if (pi-aperture)>(aperture/LineNum)
    thc = linspace(-direction-aperture/2, -direction+aperture/2, LineNum);
else
    thc = linspace(-direction-pi/2, -direction+pi/2-pi/LineNum, LineNum);
end

thc=mod(thc,pi);

S = zeros(height,width);
for ll = 1:LineNum
    if ((thc(ll) <= pi/4) || (thc(ll) > 3*pi/4))
        yr        = round(tan(thc(ll))*(-width/2+1:width/2-1)+height/2+1);
        yr(yr<=0) = NaN;
        yr(yr>height)  = NaN;
        for nn = 1:width-1
            if ~isnan(yr(nn))
                S(yr(nn),nn+1) = 1;
            end
        end
    else
        xc        = round(cot(thc(ll))*(-height/2+1:height/2-1)+width/2+1);
        xc(xc<=0) = NaN;
        xc(xc>width)  = NaN;
        for nn = 1:height-1
            if ~isnan(xc(nn))
                S(nn+1,xc(nn)) = 1;
            end
        end
    end
end

S = ifftshift(S);    % sampling mask
return



function [M, pinvM, transpM, S]   =  Pseudo_Random( im, ratio, seed )
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Written by Weisheng Dong 
% http://see.xidian.edu.cn/faculty/wsdong/NLR_Exps.htm
% W. Dong, G. Shi, X. Li, Y. Ma, and F. Huang, "Compressive sensing via 
% Nonlocal Low-rank Regularization," IEEE Trans. Image Process., vol. 23, 
% no. 8, pp. 3618-3632, 2014.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rng(seed,'v4');
[height, width]       =    size(im);
if     ratio == 0.2
    factor   =    4.427;
elseif ratio == 0.25
    factor   =  4;
else
    factor   =  sqrt(1/ratio)*2;
end
picks        =  RandMask_rect( double(height/factor), double(width/factor), height, width);
S            =  zeros(height*width,1);
S(picks)     =  1;
S            =  reshape(S,[height width]); % sampling mask
M            =  @(z) S.*fft2(z);        % forward sampling operator (of the 2D FFT spectrum)
pinvM        =  @(z) ifft2(z);          % pseudo-inverse of the operator M
transpM      =  @(z) numel(z)*ifft2(z); % transpose of the operator M
% pinvM    = @(z) iff2(z./(S+eps));     % when 'S' is not binary
% transp   = @(z) numel(z)*iff2(S.*z);  % when 'S' is not binary
return



function indeces = RandMask_rect(tolm, toln, m, n)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Shiqian Ma, Wotao Yin, Yin Zhang and Amit Chakraborty, "An Efficient Algorithm 
% for Compressed MR Imaging Using Total Variation and Wavelets", in IEEE CVPR 2008. 
% Author: Shiqian Ma
% Date : 09/05/2007
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rng(0,'v4');
M = zeros(m,n);

for i=1:m
	for j=1:n
		if ( ( abs(round(i-m/2)-(rand-0.5)*m/2)  < tolm ) &&  ( abs(round(j-n/2)-(rand-0.5)*n/2)  < toln ) )
			M(i,j) = 1;
		end
	end
end

cm         = round(m/2);
cn         = round(n/2);
M(cm,cn)   = 1;
M          = ifftshift(M);
M(1)       = 1;
indeces    = find(M);
return



function [M, pinvM, transpM, S] =  Volumetric_Radial_Line( im, coverage, seed )
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Returns the indicator of the domain in 3D Fourier space for the 3D radial
% line sampling.
% M. Maggioni, V. Katkovnik, K. Egiazarian, A. Foi, "A Nonlocal Transform-Domain Filter 
% for Volumetric Data Denoising and Reconstruction", IEEE Trans. Image Process.,
% vol. 22, no. 1, pp. 119-133, January 2013. doi:10.1109/TIP.2012.2210725
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

rng(seed,'v4');
coverage = coverage/100;       % coverage of the 3D FFT space for 3D radial line sampling
tol      = 1/100;          % tolerance error of coverage 
line_std = 0/100;          % noise in subsampling trajectories
history  = zeros(2);
stop     = 0;
S        = zeros(size(im));
n        = size(im,1);
scale    = 8;
p0       = [0 0 0];
p1       = size(S);
t = 0:1/n:1;
while (sum(S(:))==0 || abs(mean(S(:))-coverage)>tol) && ~stop
    if mean(S(:))>coverage
        scale = scale+1;
    else
        scale = scale-1;
    end
    if scale==history(1,1)
        scale = history(1,history(2,:)==min(history(2,:)));
        stop = 1;
    end
    S = zeros(size(im));
    for sx=0:scale:p1(2)
        for sy=0:scale:p1(1)
            for sz=0:scale:p1(3)
                a = p0;
                b = p1;
                a(1) = max(1,min(p1(2),a(1)+sx));
                a(2) = max(1,min(p1(1),a(2)+sy));
                a(3) = max(1,min(p1(3),a(3)+sz));
                
                b(1) = max(1,min(p1(2),b(1)-sx));
                b(2) = max(1,min(p1(1),b(2)-sy));
                b(3) = max(1,min(p1(3),b(3)-sz));
                d = b-a;
                x = max(1,min(p1(2),round(a(1)+d(1).*t+line_std*randn)));
                y = max(1,min(p1(1),round(a(2)+d(2).*t+line_std*randn)));
                z = max(1,min(p1(3),round(a(3)+d(3).*t+line_std*randn)));
                for i=1:length(t)
                    S(y(i),x(i),z(i)) = 1;
                end
            end
        end
    end
    history = circshift(history,[0 1]);
    history(1,end) = scale;
    history(2,end) = abs(mean(S(:))-coverage);
end
for i=1:size(S,3)
    S(:,:,i) = ifftshift(S(:,:,i));
end
for i=1:size(S,1)
    for j=1:size(S,2)
        S(i,j,:) = squeeze(ifftshift(S(i,j,:)));
    end
end

M            = @(z) S.*fftn(z);        % forward sampling operator (of the 3D FFT spectrum)
pinvM        = @(z) ifftn(z);          % pseudo-inverse of the operator M
transpM      = @(z) numel(z)*ifftn(z); % transpose of the operator M
% pinvM    = @(z) iffn(z./(S+eps));    % when 'S' is not binary
% transp   = @(z) numel(z)*iffn(S.*z); % when 'S' is not binary
return




function [M, pinvM, transpM, mask] = Compressive_Spectral_Sampling(x, samp_rate, seed)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Returns a sampling operator "M" by which we obtain a single-shot spectrally
% compressed measurements. Each band is masked by a shifting pseudo-random
% binary mask with "samp_rate" transmittance; the masked bands are then
% summed into a a single measurement array.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

rng(seed,'v4');
[height, width, depth] = size(x);

if depth==1
   error('the input data should be multi-channel!') 
end

Tr                = samp_rate;                          % Transmittance level of the coded aperture
mask              = zeros([height width]);                     
rnd_binary_arr    = 1*(rand([height width])<Tr);        % uniformly distributed random binary 2-D coded aperture
for j = 1:depth
    mask(:,:,j)   = circshift(rnd_binary_arr,j-1,2);    % 3-D binary mask
end
clear Tr rnd_binary_arr j
sum_mask = sum(mask,3);

M        = @(X) sum(X.*mask,3);                         % Forward sampling operator M
pinvM    = @(X) (X./max(sum_mask,1)).*mask;             % pseudo-inverse of the operator M 
transpM  = @(X) (X./max(sum_mask,1)).*mask.*sum_mask;   % transpose operator of the operator M 


return





function [M, pinvM, transpM, mask] = Compressive_Temporal_Sampling(x, samp_rate, seed)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Returns a sampling operator "M" by which we obtain a single-shot spectrally
% compressed measurements. Each band is masked by a shifting pseudo-random
% binary mask with "samp_rate" transmittance; the masked bands are then
% summed into a a single measurement array.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

rng(seed,'v4');
[height, width, depth] = size(x);

if depth==1
   error('the input data should be multi-frame (video)!') 
end

chunkSz  = 1/samp_rate(1); % chunkSz = 1/temporal_compression_ratio
NumMeas  = depth./chunkSz;

%%% the height x width x chunkSz random binary mask constituted as horizontally
%%% shifted versions of a single random binary mask drawn from Bernoulli
%%% distribution with probability "Tr";

Tr             = samp_rate(2); %Tr =0.5; % Transmittance level of the coded aperture
shift          = 1;  % one pixel horizontal shift
for t = 1:chunkSz
    if t==1
        mask   = binornd(1, Tr, [height, width chunkSz]); % 3-D binary mask
    else
        mask(:,1+(t-1)*shift:width,t) = ...
                  mask(:,1+(t-2)*shift:width-shift,t-1);  % 3-D binary mask (shifting horizontally)
%         mask(1+(t-1)*shift:height,:,t) = ...
%                   mask(1+(t-2)*shift:height-shift,:,t-1);  % 3-D binary mask (shifting vertically)
    end
end

sum_mask = sum(mask,3); 
% Forward sampling operator M for the whole video frames 
M        = @(X) Multi_Chunk_Op(X, 'Forward',   mask, NumMeas, chunkSz, sum_mask);   
% pseudo-inverse operator of M 
pinvM    = @(X) Multi_Chunk_Op(X, 'Pinv',      mask, NumMeas, chunkSz, sum_mask); 
% transpose operator of M 
transpM  = @(X) Multi_Chunk_Op(X, 'Transpose', mask, NumMeas, chunkSz, sum_mask); 


return


function output = Multi_Chunk_Op(input, operator, mask, NumMeas, chunkSz, sum_mask)

if     strcmpi('Forward',  operator)
    OPR = @(X) sum(X.*mask,3);   % Forward sampling operator for each chunk of the video
    
elseif strcmpi('Pinv',     operator) 
    OPR = @(X) (X./max(sum_mask,1)).*mask; % pseudo-inverse operator for each sampled chunk of the video 
                                         
elseif strcmpi('Transpose',operator)
    OPR = @(X) (X./max(sum_mask,1)).*mask.*sum_mask; % transpose operator for each sampled chunk of the video
    
else
    error('Your "operator" does not exist!')
end

if strcmpi('Forward',  operator)
    output  = zeros([size(input,1), size(input,2), NumMeas]);
    for t = 1:NumMeas
        output(:,:,t) = OPR(input(:,:,(t-1)*chunkSz+(1:chunkSz)));
    end
    
elseif ismember(operator, {'Pinv','Transpose'})
    output  = zeros([size(input,1), size(input,2), NumMeas*chunkSz]);
    for t = 1:NumMeas
        output(:,:,(t-1)*chunkSz+(1:chunkSz)) = OPR( input(:,:,t) );
    end
    
end

return







%%
%%
%%                   %% injects noise to the measurement %%
%%
%%


                     
                                          
function [x_noisy, std_WGN] = Additive_WGN(x,snr)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
% Additive_WGN Adds zero-mean real-/complex-valued white Gaussian noise to
% the signal "x" yielding signal-to-noise ratio of "snr" dB with respect to 
% the power of signal, equivalent to:
%
%               x_noisy = awgn(x,snr,'measured')
%
% (useful when "Communication System Toolbox" is not installed in your MATLAB.)
% 
%  FUNCTION INTERFACE:
%  x_noisy = Additive_WGN(x,snr)
%
%  INPUTS:
%     1) x               :Input signal
%     2) snr             :Signal-to-noise ratio (in dB)
%
%
%  OUTPUTS:
%     1) x_noisy         :Noisy signal
%     2) std_WGN         :Noise standard deviation
%
%-------------------------------------------------------------------------
%  EXAMPLES:
%
% x   = im2double(imread('cameraman.tif'));
% snr = 10;
% rng(0,'v4')
% x_noisy1 = awgn(x,snr,'measured');
% 
% rng(0,'v4')
% x_noisy = Additive_WGN(x,snr);
% 
% diff = x_noisy1 - x_noisy;
% [max(diff(:)) min(diff(:))]
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    if isreal(x)==1
        WGN      = randn(size(x));
        std_WGN  = sqrt(mean(x(:).^2)./(10.^(snr/10)));
        x_noisy  = x + std_WGN .* WGN;
    else
        real_WGN = randn(size(x));
        imag_WGN = randn(size(x));
        std_WGN  = sqrt(mean(abs(x(:)).^2)./(10.^(snr/10)));
        x_noisy  = x + (std_WGN/sqrt(2)) .* (real_WGN + 1i.* imag_WGN);
    end
return







%%
%%
%%                        functions for visualizations
%%
%%


function Visualizer(x, y, x_backproj, D, samp_strat, samp_rate, snr)

if iscell(x)==1   %%% multimodal, multi-sensor or multi-acquisition cases
    
    %%% only visualizing those of 'radial', 'lim_angle', 'pseudo-random' sampling strategies
    check_strat_vis = cellfun(@(c)strcmpi(c,samp_strat),...
        {'radial', 'lim_angle','pseudo-random', 'pseudo_random'},'UniformOutput',false);
    checked_idx_vis = sum(cell2mat(check_strat_vis.'),1);
    idx_vis = find(checked_idx_vis==1);
    I = sum(checked_idx_vis);
    if I~=0; figure, end
    
    for i = 1:I
        ivi = idx_vis(i);
        subplot(3,I,I*i-2),imshow(x{ivi},[]),colorbar
        title(['$\bf{x}_' sprintf('%d',ivi) '$'],'Interpreter','latex')
        subplot(3,I,I*i-1),imshow(real(x_backproj{ivi}),[]),colorbar,
        title(['$Real\left(\bf{M}^{\dagger}_' sprintf('%d',ivi) '\bf{y}_' sprintf('%d',ivi) '\right)$'],'Interpreter','latex')
        subplot(3,I,I*i),imshow(fftshift(D{ivi}),[]),colorbar,
        title({['$\bf{D}_' sprintf('%d, %s~%s',ivi, num2str(samp_rate{ivi}),samp_strat{ivi}) '$'],...
            sprintf('SNR: %d~(dB)',snr{ivi})},'Interpreter','latex')
    end
    
    
    
else                %%% monomodal or single-acquisition case
    
    check_strat_vis = cellfun(@(c)strcmpi(c,samp_strat),...
        {'radial', 'lim_angle','pseudo-random', 'pseudo_random',...
         '3D_radial', 'CSI', 'CTI'},'UniformOutput',false);
    
    fg=figure(1);
    fg.WindowState = 'Maximized';  % making the whole figure as size as your screen
    
    
    if sum(cell2mat(check_strat_vis(1:4).'),1)==1
        %%% visualizing that of 'radial', or 'lim_angle', or 'pseudo-random' sampling strategy
        
        subplot(2,2,1),imshow(x,[]),colorbar
        title('$\bf{x}$','Interpreter','latex')
        
        subplot(2,2,3),imshow(real(x_backproj),[]),colorbar,
        title('$Real\left(\bf{M}^{\dagger}\bf{y}\right)$','Interpreter','latex')
        
        subplot(2,2,4),imshow(imag(x_backproj),[]),colorbar,
        title('$Imag\left(\bf{M}^{\dagger}\bf{y}\right)$','Interpreter','latex')
        
        subplot(2,2,2),imshow(fftshift(D),[]),colorbar,
        title('FFT sampling mask','Interpreter','latex')
        
        suptitle(sprintf(['\n 2D %s sampling \n',...
            '(%0.2f%s of full coverage) \n',...
            'SNR = %02d dB \n'],...
            samp_strat,100*mean(D(:)),'%' , snr));
        
    elseif cell2mat(check_strat_vis(5).')==1
        %%% visualizing that of '3D_radial' line sampling strategy
        
        clip = @(x1,x2) max(min(x1(:)),min(max(x1(:)),x2));   % clipping function only for visualization 
        
        subplot(1,4,1)
        visualize3DXsect(x)
        title('$\mathbf{x}$','Interpreter','Latex','fontsize',14);
        
        subplot(1,4,3)
        visualize3DXsect(clip(x,real(x_backproj)))
        title('$Real\left(\mathbf{M}^{\dagger}\mathbf{y}\right)$','Interpreter','Latex','fontsize',14);

        subplot(1,4,4)
        visualize3DXsect(clip(x,imag(x_backproj)))
        title('$Imag\left(\mathbf{M}^{\dagger}\mathbf{y}\right)$','Interpreter','Latex','fontsize',14);
        
        subplot(1,4,2)
        visualize3DXsect(fftshift(D))
        title('$\textrm{3D radial sampling mask}$','Interpreter','Latex','fontsize',12);
        
        suptitle(sprintf(['\n\n\n\n 3D radial lines sampling \n',...
            '(%0.2f%s of full coverage) \n',...
            'SNR = %02d dB \n'],...
            100*mean(D(:)),'%' , snr));
        
    elseif cell2mat(check_strat_vis(6).')==1
        %%% visualizing that of 'CSI' line sampling strategy
        
        gammaCorr     = .5; % gamma correction factor for visualization
        clip          = @(x1,x2) max(min(x1(:)),min(max(x1(:)),x2));   % clipping function only for visualization
        RGB_x         = MS2RGB(x).^gammaCorr;
        RGB_pinvM_y = MS2RGB(clip(x,x_backproj)).^gammaCorr;
   
        
        subplot(1,3,1)
        imshow(RGB_x)
        title('\textrm{RGB rep. of ground-truth} $\mathbf{x}$','Interpreter','Latex',...
            'HorizontalAlignment','center','VerticalAlignment','bottom','fontsize',16);
        
        subplot(1,3,2)
        imshow(y,[])
        title('\textrm{single-shot measurements} $\mathbf{y}$',...
            'Interpreter','Latex',...
            'HorizontalAlignment','center','VerticalAlignment','bottom','fontsize',16);
        
        subplot(1,3,3)
        imshow(RGB_pinvM_y)
        title('\textrm{RGB rep. of pseudo-inverse} $(\mathbf{M}^{\dagger}\mathbf{y})$',...
            'Interpreter','Latex',...
            'HorizontalAlignment','center','VerticalAlignment','bottom','fontsize',16);
        


        
        
       
    elseif cell2mat(check_strat_vis(7).')==1
        %%% visualizing that of 'CTI' line sampling strategy

        Frame_no = max(2, round(size(x,3)/2));
        meas_no  = max(1, round(size(y,3)/2));
        
        subplot(1,3,1)
        imshow(x(:,:,Frame_no),[]), colorbar
        title(['\textrm{ground-truth} $\mathbf{x}$ \textrm{(frame \#' sprintf('%d',Frame_no) ')}'],...
            'Interpreter','latex','fontsize',16);
        
        subplot(1,3,2)
        imshow(y(:,:,meas_no),[]), colorbar
        title(['$\mathbf{y}$ \textrm{(measurements frame \#' sprintf('%d',meas_no) ')}'],...
            'Interpreter','latex','fontsize',16);
        
        subplot(1,3,3)
        imshow(x_backproj(:,:,Frame_no),[]), colorbar
        title(['\textrm{pseudo-inverse} $(\mathbf{M}^{\dagger}\mathbf{y})$ \textrm{(frame \#' sprintf('%d',Frame_no) ')}'],...
            'Interpreter','latex','fontsize',16);
        
        
    end
    
end

return





function visualize3DXsect( D )
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% visualize3DXsect visualizes a 3D cross-section of a 3D data
%-------------------------------------------------------------------------
%  EXAMPLE:
%
%       load mri;  % load 3D MRI data
%       D = double(squeeze(D));  % of size 128x128x27
%       visualize3DXsect( D ) ;  % visualization of a 3D cross-section
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if ndims(D)~=3
    error('Input must be 3-D.')
end

D = (D-min(D(:)))./(max(D(:))-min(D(:))); % normalizing the range of D into the range [0, 1]

% show 3-D cross-sections
%             h = figure;
%             set(h, 'color', 'white');
%             set(h, 'renderer', 'zbuffer');

    x = round(size(D)./2);
    
    p1 = squeeze(D(x(1),:,:));
    p2 = squeeze(D(:,x(2),:))';
    p3 = squeeze(D(:,:,x(3)));
    
    offset = 1;
    p1(:,1:x(3)-offset) = [];
    p1(1:x(2)-offset,:) = [];
    
    p3(:,1:x(2)-offset) = [];
    p3(1:x(1)-offset,:) = [];
    
    p2(:,1:x(1)-offset) = [];
    p2(1:x(3)-offset,:) = [];
    
    x(:) = offset;
    
    x1 = zeros(size(p1)) + x(1);
    x2 = zeros(size(p2)) + x(2);
    x3 = zeros(size(p3)) + x(3);
    [X1_1,X1_2] = meshgrid(1:size(p1,2),1:size(p1,1));
    [X2_1,X2_2] = meshgrid(1:size(p2,2),1:size(p2,1));
    [X3_1,X3_2] = meshgrid(1:size(p3,2),1:size(p3,1));
    hold all
    warp(X1_2,x1,X1_1,p1)
    warp(x2,X2_1,X2_2,p2)
    warp(X3_1,X3_2,x3,p3)
    view([35 30])
    axis off vis3d;
    camproj('persp')

return




function RGB_im = MS2RGB(x, WaveLengths)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% MS2RGB provides the RGB representation of the input multispectral image
%
%  FUNCTION INTERFACE:
%  bar_gray(rootPSD, txt, sigma_k)
% 
%  INPUTS:
%     1) x              (3D array) : the input 3D multispectral image
%     2) WaveLengths    (1D array) : the wavelengths of the input image 
%                                    (spectral bands interval)
%  Output:
%     1) RGB            (3D array) : the RGB representation of the input image
%-------------------------------------------------------------------------
%  EXAMPLE:
%
%  load('../Data/chart_and_stuffed_toy_ms_16bit.mat')  % the 512x512x31 "chart_and_stuffed_toy" 
%                                                      % multispectral image, available at:
%                                                      % http://www.cs.columbia.edu/CAVE/databases/multispectral/
%  WaveLengths       = 400:10:700;   % wavelengths
%
%  RGB_im = MS2RGB(x, WaveLengths);
%  figure, imshow(RGB_im.^.5)        % visualization (the power 0.5 is for gamma-correction)
%
%   nasser.eslahi@tuni.fi
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


if ~exist('WaveLengths','var')||isempty(WaveLengths)
    WaveLengths = linspace(400,700,size(x,3));
end

R = mean(x(:,:,(WaveLengths<=700 & WaveLengths>=635) ), 3);
G = mean(x(:,:,(WaveLengths<=560 & WaveLengths>=520) ), 3);
B = mean(x(:,:,(WaveLengths<=490 & WaveLengths>=450) ), 3);

RGB_im = cat(3,R,G,B);
RGB_im =( RGB_im - min(RGB_im(:)) )./(max(RGB_im(:)) -  min(RGB_im(:)));

return
