function [glob_FFT_PSD_e_k, Tr_PSD_e_k] = Est_effective_noise_PSD(...
                                          problem, Tr, Tr_Sz, locWin, b_k, d_k, ...
                                          loc_FFT_f_rate, loc_FFT_rootPSD_pinvM_eps, ...
                                          PnP_framework)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Est_effective_noise_PSD computes the FFT and transform-domain PSDs of ...
%  effective noise according to Section IV-C and Appendix B of [*]
% [*] N. Eslahi and A. Foi, "Improved sparse Signal Recovery via Adaptive Correlated Noise Model",
%     submitted to IEEE TCI, 2022.
%
%
%  FUNCTION INTERFACE:
%        [glob_FFT_PSD_e_k, Tr_PSD_e_k] = Est_effective_noise_PSD(...
%                                         problem, Tr, Tr_Sz, locWin, b_k, ...
%                                         d_k, loc_FFT_f_rate, loc_FFT_rootPSD_pinvM_eps, ...
%                                         PnP_framework)
%
% _________________________________________________________________________________________
% INPUTS:          |      TYPE:      |  DESCRIPTION:
% -----------------------------------------------------------------------------------------
%  problem         |      (char)     |  selected sparse recovery problem. 
%                  |        OR       |  <OPTIONS>      
%                  |     (string)    |       '2D Tomography'
%                  |                 |       '3D Tomography'
%                  |                 |       'Compressive Spectral Imaging'
%                  |                 |       'Compressive Temporal Imaging'
% -----------------------------------------------------------------------------------------
%  Tr              |(char OR string) | selected transform for computing the PSD
%                  |                 | in its domain. This transform should be the one
%                  |                 | used in the shrinkage step by the filter.
% -----------------------------------------------------------------------------------------
%  Tr_Sz           |     (double)    | the support size of the transform "Tr".
% -----------------------------------------------------------------------------------------
%  locWin          |     (double)    | the support size of the FFT for computing
%                  |                 | the local FFT-PSDs using the local moving window.
% -----------------------------------------------------------------------------------------
%  b_k             |     (double)    | the signal which is added to the previous 
%                  |                 | estimate of the underlying unknown image.
% -----------------------------------------------------------------------------------------
%  d_k             |     (double)    | the back-projection of the measurement residual. 
% -----------------------------------------------------------------------------------------
%  loc_FFT_f_rate  |     (double)    | ratio between the local FFT-PSD compensation vectors of
%                  |                 | perpM and MdaggerM (see the syntax of PSD_Auxiliaries.m).
% -----------------------------------------------------------------------------------------
% loc_FFT_rootPSD_pinvM_eps |(double)| local FFT-root-PSD of the back-projected measurement 
%                  |                 | error (see the syntax of PSD_Auxiliaries.m).
% -----------------------------------------------------------------------------------------
%  PnP_framework   |      (char)     |  (optional) selected PnP framework. 
%                  |        OR       |  <OPTIONS>      
%                  |     (string)    |       'ADMM'
%                  |                 |       'ISTA'
%                  |                 |       'FISTA'
%                  |                 | **Note: This is used only for fast computation of 
%                  |                 | the PSD of "w_k = b_k + d_k" under ISTA/FISTA.
%                  |                 | It can be set to empty or not given, for a normal
%                  |                 | computation of such PSD.                 
%
%
%
%
% _________________________________________________________________________________________
% OUTPUTS:         |      TYPE:      |  DESCRIPTION:
% -----------------------------------------------------------------------------------------
% glob_FFT_PSD_e_k |     (double)    |  The approximated global FFT-PSD of effective noise.
% -----------------------------------------------------------------------------------------
%  Tr_PSD_e_k      |     (double)    |  The approximated transform-domain PSD of effective noise. 
%
%
%  EXTERNAL DEPENDENCY FUNCTIONS: 
%    1) Est_rootPSD.m  (computes the root-PSD via local moving window)
%    2) FFT_PSD_Upsampling.m (upsamples a local FFT-PSD to constitute its global version)
%    3) FFT_PSD_domain_convertor.m (converts an FFT-PSD into a Tr-PSD)
% ** for more information on each of the above functions, please see their syntaxes.
% 
%
%
%            Authors :         Nasser Eslahi & Alessandro Foi
%                                   Tampere University
%        Contact info:          firstname.lastname@tuni.fi
%
%                          Release ver. 1.0  (April 2022)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Copyright (c) 2006-2022 Tampere University.
% All rights reserved.
% This work (software, material, and documentation) shall only be used for nonprofit 
% noncommercial purposes.
% Any unauthorized use of this work for commercial or for-profit purposes is prohibited.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if nargin < 8
    error(['Necessary inputs are not given! ' ...
           'Please check the <a href="matlab:help %s">syntax</a>'],mfilename)

elseif nargin == 9
    if sum(strcmpi(PnP_framework, {'ISTA', 'FISTA'}))==1
        fast_computation = true; % fast computation of the PSD of "w_k = b_k + d_k"
        %%% Note that in PnP-ISTA/FISTA: "b_k=rho*pinvM(r_k)", "d_k=-pinvM(r_k)" ...
        %%% and therefore "w_k=(rho-1)*d_k"
        rho = -sum(b_k(:))/sum(d_k(:));
    else
        fast_computation = false;
    end

elseif nargin == 8
        fast_computation = false;
end

SizeX   = size(b_k);

FFT_Tr  = repelem({'fft'},size(locWin,2)); % the 2D/3D FFT for computing the local PSDs
w_k     = b_k + d_k;


if     sum(strcmpi(problem, {'2D Tomography', '3D Tomography', 'Compressive Spectral Imaging'}))==1
    %%% estimating the local FFT-root-PSDs of "d_k" and "w_k" as the sample std ...
    %%% of their local FFT spectrum computed over a moving window, as each ...
    %%% position of the window can be treated as different realization of the noise.
    loc_FFT_rootPSD_d_k = Est_rootPSD(d_k, locWin, FFT_Tr, 0, 'sstd');
    if fast_computation
        loc_FFT_rootPSD_w_k = abs(rho-1).*loc_FFT_rootPSD_d_k; % in PnP-ISTA/FISTA: "w_k=(rho-1)*d_k"
    else
        loc_FFT_rootPSD_w_k = Est_rootPSD(w_k, locWin, FFT_Tr, 0, 'sstd');
    end

    
    %%% approximating the local FFT-PSD of surrogate noise (according to (17) in [*])
    loc_FFT_PSD_s_k = (loc_FFT_rootPSD_w_k + loc_FFT_rootPSD_pinvM_eps).^2;
    %%% approximating the local FFT-PSD of latent noise (according to (20) in [*])
    loc_FFT_PSD_l_k =  loc_FFT_f_rate.*(loc_FFT_rootPSD_d_k - loc_FFT_rootPSD_pinvM_eps).^2;
    
    %%% approximating the local FFT-PSD of effective noise as sum of the approximated ...
    %%% local FFT-PSDs of surrogate and latent noise (according to (20) in [*])
    loc_FFT_PSD_e_k =  loc_FFT_PSD_s_k + loc_FFT_PSD_l_k;
    
    
    %%% constituting a global FFT-PSD given the local FFT-PSD
    if     sum(strcmpi(problem, {'2D Tomography', '3D Tomography'}))==1
        glob_FFT_PSD_e_k  = FFT_PSD_Upsampling(loc_FFT_PSD_e_k, SizeX); % via upsampling
    else
        glob_FFT_PSD_e_k = abs(reshape(...
            imresize(...
            reshape(loc_FFT_PSD_e_k,[prod(locWin(1:2)) locWin(3)]),...
            [prod(locWin(1:2)) SizeX(3)])*(SizeX(3)/locWin(3)),...
            [locWin(1:2) SizeX(3)])); % via resizing (interpolating)
    end
    
    %%% converting the global FFT-PSD into Tr-PSD compatible with ...
    %%% the selected denoiser for filtering within PnP
    Tr_PSD_e_k  = FFT_PSD_domain_convertor(glob_FFT_PSD_e_k, Tr_Sz, Tr);
    
    
else
    
    %%% computing the FPN and RND components of w_k and d_k
    w_k_FPN = mean(w_k,3);
    w_k_RND = (w_k - w_k_FPN);
    d_k_FPN = mean(d_k,3);
    d_k_RND = (d_k - d_k_FPN);
    
    %%% estimating the FFT-root-PSDs of RND and FPN components of "d_k" and "w_k" ...
    %%% as the sample std of their local FFT spectrum computed over a moving window, ...
    %%% as each position of the window can be treated as different realization of the noise.    
    loc_FFT_rootPSD_d_k_FPN = Est_rootPSD(d_k_FPN, locWin, FFT_Tr, 0, 'sstd');
    loc_FFT_rootPSD_d_k_RND = Est_rootPSD(d_k_RND, locWin, FFT_Tr, 0, 'sstd');
    
    if fast_computation
        loc_FFT_rootPSD_w_k_FPN = abs(rho-1)*loc_FFT_rootPSD_d_k_FPN;
        loc_FFT_rootPSD_w_k_RND = abs(rho-1)*loc_FFT_rootPSD_d_k_RND;
    else
        loc_FFT_rootPSD_w_k_FPN = Est_rootPSD(w_k_FPN, locWin, FFT_Tr, 0, 'sstd');
        loc_FFT_rootPSD_w_k_RND = Est_rootPSD(w_k_RND, locWin, FFT_Tr, 0, 'sstd');        
    end
    
    clear d_k w_k w_k_FPN w_k_RND d_k_FPN d_k_RND
    
    %%% FFT-PSD compensation ratios
    loc_FFT_f_FPN = loc_FFT_f_rate(:,:,1);
    loc_FFT_f_RND = loc_FFT_f_rate(:,:,2);
    
    %%% local FFT-root-PSDs of FPN and RND component of back-projected measurement error
    loc_FFT_rootPSD_pinvM_eps_FPN = loc_FFT_rootPSD_pinvM_eps(:,:,1);
    loc_FFT_rootPSD_pinvM_eps_RND = loc_FFT_rootPSD_pinvM_eps(:,:,2);
    
    
    %%% approximating the local FFT-PSDs of  RND and FPN components of effective noise as ...
    %%% sum of their corresponding approximated PSDs of surrogate and latent noise.
    loc_FFT_PSD_s_k_FPN  = ( loc_FFT_rootPSD_w_k_FPN + loc_FFT_rootPSD_pinvM_eps_FPN).^2;
    loc_FFT_PSD_l_k_FPN  = loc_FFT_f_FPN.*abs(loc_FFT_rootPSD_d_k_FPN - loc_FFT_rootPSD_pinvM_eps_FPN).^2;
    loc_FFT_PSD_e_k_FPN  = loc_FFT_PSD_s_k_FPN + loc_FFT_PSD_l_k_FPN;
    
    loc_FFT_PSD_s_k_RND  = ( loc_FFT_rootPSD_w_k_RND + loc_FFT_rootPSD_pinvM_eps_RND).^2;
    loc_FFT_PSD_l_k_RND  = loc_FFT_f_RND.*abs(loc_FFT_rootPSD_d_k_RND - loc_FFT_rootPSD_pinvM_eps_RND).^2;
    loc_FFT_PSD_e_k_RND  = loc_FFT_PSD_s_k_RND + loc_FFT_PSD_l_k_RND;
    
    clear loc_FFT_rootPSD_d_k_FPN  loc_FFT_rootPSD_d_k_RND ...
          loc_FFT_rootPSD_w_k_FPN  loc_FFT_rootPSD_w_k_RND ...
          loc_FFT_PSD_s_k_FPN      loc_FFT_PSD_s_k_RND     ...
          loc_FFT_PSD_l_k_FPN      loc_FFT_PSD_l_k_RND     ...
          loc_FFT_f_FPN            loc_FFT_f_RND
    
    glob_FFT_PSD_e_k = zeros([SizeX(1:2), 2]);  
    %%% constituting global FFT-PSDs given the local FFT-PSDs (via upsampling)
    glob_FFT_PSD_e_k(:,:,1) = FFT_PSD_Upsampling(loc_FFT_PSD_e_k_FPN, SizeX(1:2)); % global FFT-PSD of FPN part
    glob_FFT_PSD_e_k(:,:,2) = FFT_PSD_Upsampling(loc_FFT_PSD_e_k_RND, SizeX(1:2)); % global FFT-PSD of RND part
    
    
    Tr_PSD_e_k = zeros([Tr_Sz, 2]);
    %%% converting the global FFT-PSDs into Tr-PSDs compatible with ...
    %%% the selected denoiser for filtering withing PnP.
    Tr_PSD_e_k(:,:,1)   = FFT_PSD_domain_convertor(glob_FFT_PSD_e_k(:,:,1), Tr_Sz, Tr); % Tr-PSD of FPN part 
    Tr_PSD_e_k(:,:,2)   = FFT_PSD_domain_convertor(glob_FFT_PSD_e_k(:,:,2), Tr_Sz, Tr); % Tr-PSD of RND part 
    
    
end

clearvars -except  Tr_PSD_e_k  glob_FFT_PSD_e_k