function rootPSD = Est_rootPSD(eta, rPSD_size, Tr, visualize, estimator, Pr1, Pr2)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Est_rootPSD estimates 2D/3D noise root-PSD (i.e. square-root of PSD) via 
%  an estimator (e.g., IQR, MAD, sample mean, and sample standard deviation) 
%  over the local transform spectrum of "eta" computed over a moving window,
%  as each position of the window can be treated as a different realization 
%  of the noise. The local transform spectrum of "eta" refers to the spectra 
%  of all blocks/cubes extracted from "eta". Est_rootPSD assumes normality 
%  of the noise; although the distribution of "eta" may diverge from being 
%  Gaussian, each subband of its local 2D/3D Tr-spectra follows a Gaussian 
%  distribution according to the central-limit theorem.
%
%  [1]. N. Eslahi and A. Foi, "Anisotropic spatiotemporal regularization in
%  compressive vide recovery by adaptively modeling the residual errors as
%  correlated noise", in Proc. IEEE IVMSP 2018.
%
%  [2]. N. Eslahi and A. Foi, "Improved sparse signal recovery via adaptive 
%  correlated noise model", submitted to IEEE TCI, 2022.
%
%
%  FUNCTION INTERFACE:
%  rootPSD = Est_rootPSD(eta, rPSD_size, Tr, visualize, estimator, Pr1, Pr2)
% 
%  INPUTS:
%     1) eta              (2D/3D array) : real-/complex-valued (surrogate) noise
%     2) rPSD_size        (double)      : the size of output root-PSD
%     3) Tr               (cell/char)   : sparsifying transform corresponding to 
%                                         those used internally by the filter
%     4) visualize        (logical)     : 0 --> do not visualize output information
%                                       : 1 --> visualize information to screen
%                                         (default is 0)
%     5) estimator        (char)        : 'IQR'    --> interquantile range
%                                       : 'MAD'    --> median absolute deviation
%                                       : 'meanAD' --> mean absolute deviation
%                                       : 'sstd'   --> sample standard deviation
%                                         (default is 'MAD')
%     6) Pr1              (double)      : Pr1-quantile (default is 0) --
%     7) Pr2              (double)      : Pr2-quantile (default is 0.5)
%
%     * inputs Pr1 and Pr2 are optional and only used for IQR
%
%  OUTPUTS:
%     1) rootPSD          (2D/3D array) : the Tr-root-PSD of z
%
%-------------------------------------------------------------------------
%  EXAMPLES:
%      
%     1) Estimating 2D FFT root-PSD of a real valued
%      correlated Gaussian noise:  
%
%       N          = 1023;
%       w          = randn(N); % white Gaussian noise
%       g1         = fspecial('gaussian',7,1.5);
%       g2         = fspecial('gaussian',7,0.5);
%       g          = g1-g2;    % correlation kernel
%       eta        = imfilter(w,g,'circular');   % correlated noise
%       rPSD_size  = 31; % size of 2D root-PSD
%       FFT_rootPSD_eta = sqrt(abs(fft2(g,rPSD_size,rPSD_size)).^2*(rPSD_size^2));  % noise root-PSD in Fourier (FFT) domain                                                                    % FFT-PSD
%       estimator  = 'sVar'; % or other options e.g., 'MAD', 'IQR', 'meanAD' 
%       FFT_rootPSD_eta_est = Est_rootPSD(eta, [rPSD_size rPSD_size], {'fft','fft'}, 0, 'sVar');
%
%       figure, subplot(1,2,1), imagesc(FFT_rootPSD_eta), colormap('jet'),
%       title('oracle FFT root-PSD via kernel')
%       subplot(1,2,2), imagesc(FFT_rootPSD_eta_est), colormap('jet'),
%       title('estimated FFT root-PSD')
%
%
%     2) Estimating 3D DCT root-PSD of a 3D complex-valued
%      correlated Gaussian noise:  
%
%        w1        = randn(256,256,16); w1 = (w1-mean(w1(:)))./std(w1(:)); % standard WGN
%        w2        = randn(256,256,16); w2 = (w2-mean(w2(:)))./std(w2(:)); % standard WGN
%        w         = w1 + 1i*w2; % complex WGN
%        g         = randn(7,7,11);   % convolutional kernel, here we used random kernel
%        eta       = imfilter(w,g,'circular');  % correlated noise
%        estimator = 'MAD'; % or other options e.g., 'sVar', 'IQR', 'meanAD' 
%        rootPSD   = Est_rootPSD(eta, [8 8 16], {'dct','dct', 'dct'}, 1, estimator);
%
%-------------------------------------------------------------------------
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Copyright (c) 2006-2022 Tampere University.
% All rights reserved.
% This work (software, material, and documentation) shall only
% be used for nonprofit noncommercial purposes.
% Any unauthorized use of this work for commercial or for-profit purposes
% is prohibited.
%
%               AUTHORS: Nasser Eslahi & Alessandro Foi
%                 contact: firstname.lastname@tuni.fi
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%  Parameter settings
if ~exist('rPSD_size','var')||isempty(rPSD_size)
    rPSD_size= [8 8];
end

s1 = rPSD_size(1); s2 = rPSD_size(2);
if numel(rPSD_size)==2
    s3 = 1;
else
    s3  = rPSD_size(3);
end



if ~exist('Tr','var')||isempty(Tr)
    if numel(rPSD_size)==2
        Tr  = {'dct', 'dct'};         
    else
        Tr  = {'dct', 'dct', 'dct'}; 
    end 
end


if ~exist('estimator','var')||isempty(estimator)
    estimator='MAD';
end


if ~exist('Pr2','var')||isempty(Pr2)
    Pr2=0.5; % equal to MAD
end


if ~exist('Pr1','var')||isempty(Pr1)
    Pr1=0; % equal to MAD
end


if ~exist('visualize','var')||isempty(visualize)
    visualize=0; 
end

    TM_1 = single( GetTransformMatrix( s1, Tr{1,1} ) );
    TM_2 = single( GetTransformMatrix( s2, Tr{1,2} ) );  

if s3~=1
    TM_3 = single( GetTransformMatrix( s3, Tr{1,3} ) );
else
    TM_3 = 1;
end

    %% collect transform statistics using a moving window
    
step1   = min(max(round(s1/4),1),15);
step2   = min(max(round(s2/4),1),15);
step3   = 1;

% % %     if s3 == 1
% % % %%% The following is the old version of computing the 2D spectrum using a 2D moving window. 
% % % %%% Here we keep it only to show how the local moving window approach works!
% % % 
% % %         j1ToDo = unique(round([(1:step1:size(eta,1)-s1+1),size(eta,1)-s1+1]));
% % %         j2ToDo = unique(round([(1:step2:size(eta,2)-s2+1),size(eta,2)-s2+1]));
% % %         j3ToDo = unique(round([(1:step3:size(eta,3)),size(eta,3)]));
% % %         AllTrCoeffs = zeros(s2*s1,numel(j1ToDo)*numel(j2ToDo)*numel(j3ToDo),'single');
% % %         j12count = 0;
% % %         j1count  = 0;
% % %         for j1 = j1ToDo
% % %             j1count = j1count+1;
% % %             for j2 = j2ToDo
% % %                 for j3 = j3ToDo
% % %                     j12count = j12count+1;
% % %                     blk_Tr_coeff = TM_1*eta(j1:j1+s1-1,j2:j2+s2-1,j3)*conj(TM_2');
% % %                     AllTrCoeffs(:,j12count) = blk_Tr_coeff(:);
% % %                 end
% % %             end
% % %         end
% % %     else
% % % %%% The following is the old version of computing the 3D spectrum using a 3D moving window.            
% % % 
% % %         j1ToDo = unique(round([(1:step1:size(eta,1)-s1+1),size(eta,1)-s1+1]));
% % %         j2ToDo = unique(round([(1:step2:size(eta,2)-s2+1),size(eta,2)-s2+1]));
% % %         j3ToDo = unique(round([(1:step3:size(eta,3)-s3+1),size(eta,3)-s3+1]));
% % %         AllTrCoeffs = zeros(s3*s2*s1,numel(j1ToDo)*numel(j2ToDo)*numel(j3ToDo),'double');
% % %         j12count = 0;
% % %         j1count  = 0;
% % %         for j1 = j1ToDo
% % %             j1count = j1count+1;
% % %             for j2 = j2ToDo
% % %                 for j3 = j3ToDo
% % %                     j12count = j12count+1;
% % %                     j3count = 0;
% % %                     BB = zeros(s1*s2,s3);
% % %                     for j3b = j3:j3+s3-1
% % %                         j3count = j3count+1;
% % %                         B = TM_1*eta(j1:j1+s1-1,j2:j2+s2-1,j3b)*conj(TM_2');
% % %                         BB(:,j3count) = B(:);
% % %                     end
% % %                     Cube_Tr_Coeff = BB*conj(TM_3');
% % %                     AllTrCoeffs(:,j12count) = Cube_Tr_Coeff(:);
% % %                 end
% % %             end
% % %         end
% % %     end
    
    

%%%%% The following is faster than the above version
dmy     = PARSER(eta, [s1 s2 s3], [step1 step2 step3]);
dmy     = TM_1*reshape(dmy, s1, []);
dmy     = conj(TM_2)*reshape( permute( reshape(dmy, s1, s2, []), [3 1 2]), [], s2)';
dmy     = conj(TM_3')*reshape(dmy', s3, []);
AllTrCoeffs = reshape(dmy',[], s1*s2*s3)';
clear dmy
    
    
    %% compute root-PSD based on the selected estimator
if strcmpi(estimator, 'MAD') == 1
    if isreal(AllTrCoeffs)==1   % MAD of real-valued variates
        rootPSD = median(abs(AllTrCoeffs-median(AllTrCoeffs,2)),2)*1.482602218505601;  % 1/icdf('half normal',0.5,0,1)
    else                           % MAD of complex-valued variates        
        rootPSD_r = median(abs(real(AllTrCoeffs)-median(real(AllTrCoeffs),2)),2)*1.482602218505601;
        rootPSD_i = median(abs(imag(AllTrCoeffs)-median(imag(AllTrCoeffs),2)),2)*1.482602218505601;
        rootPSD = sqrt(rootPSD_r.^2+rootPSD_i.^2);
    end
elseif strcmpi(estimator, 'IQR') == 1
    if isreal(AllTrCoeffs)==1
        rootPSD = stdEstIQR(AllTrCoeffs,2,Pr1,Pr2);
    else
        rootPSD_r = stdEstIQR(real(AllTrCoeffs),2,Pr1,Pr2);
        rootPSD_i = stdEstIQR(imag(AllTrCoeffs),2,Pr1,Pr2);
        rootPSD = sqrt(rootPSD_r.^2+rootPSD_i.^2);
    end
    
    
elseif strcmpi(estimator, 'meanAD') == 1
    rootPSD = mean(abs(AllTrCoeffs-mean(AllTrCoeffs,2)),2)*sqrt(pi/2);
    
elseif strcmpi(estimator, 'sstd') == 1
%%%     rootPSD = sqrt(mean(abs(AllTrCoeffs-mean(AllTrCoeffs,2)).^2,2));
    rootPSD = sqrt(var(AllTrCoeffs,[],2));
end
    rootPSD = reshape(rootPSD,[s1,s2,s3]);
    clearvars -except rootPSD visualize
     %% Visualization

    if visualize
        if s3==1
            %%% 2D root-PSD visualization
            figure(777)
            bar3(rootPSD);
            title('\boldmath{${\Psi}$}$_{\eta}^{\frac{1}{2}}$','Interpreter','latex','fontsize',21)
            view([61,31]);colormap(0.9*ones(64,3));
            axis tight square, camproj('persp'),
            box on,  set(gca,'color',[1,1,1])
            xticks([1 s1]); xticklabels([0 s1-1]);
            yticks([1 s2]); yticklabels([0 s2-1]);
            xlabel('$\xi_1$','Interpreter','latex','fontsize',21);
            ylabel('$\xi_2$','Interpreter','latex','fontsize',21);
            X_pos = get(get(gca, 'XLabel'), 'Position');
            set(get(gca, 'XLabel'), 'Position', X_pos + [-floor(s1/2) 0 0]);
            Y_pos = get(get(gca, 'YLabel'), 'Position');
            set(get(gca, 'YLabel'), 'Position', Y_pos + [0 -floor(s2/4) 0]);
        else
            %%% 3D root-PSD visualization (only the lowest- and highest-frequency faces 
            %%% of the 3D root-PSD cube is shown)
            CubeFaceVis(rootPSD)
        end
        
    end
    
end



%%
%%           auxiliary function for extracting the blocks/cubes
%%


function vec_parsed_image = PARSER( full_image, partition_size, step_size )
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PARSER extracts 2D blocks or 3D cubes of size "partition_size" from ...
% the input signal "full_image" with step size "step_size"
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if ~ismember(ndims(full_image),[2 3])
    error('the input image should be either 2D or 3D!')
end

if ~ismember(size(partition_size,2),[2 3])
    error('our partitioning is either block-based or cube-based!')
end

if nargin == 2 || isempty(step_size)
   step_size = max(round(partition_size./2),1); 
end

if ischar(step_size)
   if strcmpi(step_size,'distinct') % non-overlapping parsing
       step_size = partition_size;
       
   elseif strcmpi(step_size,'sliding')  % fully-overlapped parsing
       step_size = ones(1,size(partition_size,2));
       
   else
       error('available options for "step_size": ''%s'' and ''%s''', 'distinct', 'sliding')
       
   end
       
end


if size(step_size,2) == 2
   step_size(3) = 1; 
end

if size(partition_size,2) == 2
   partition_size(3) = 1; 
end


if  size(step_size,2) ~= size(partition_size,2)
    error('please check the step-size!')
end



s1    = partition_size(1); % horizontal size of partition
s2    = partition_size(2); % vertical size of partition
s3    = partition_size(3); % depth size of partition (if S3=1 then partitioning is block-based, otherwise cube-based)

step1 = step_size(1);      % step-size for sliding across horizontal dimension
step2 = step_size(2);      % step-size for sliding across vertical dimension
step3 = step_size(3);      % step-size for sliding across depth

[Sz1,Sz2,Sz3] = size(full_image);

% indices of the first element in the depth dimension, ...
% starting from 0 which indicates the first slice
ind_1st_element_depth = unique([0:step3:Sz3-s3, Sz3-s3])*Sz1*Sz2;

% indices of the first element within each partition (block or cube), ...
% where each index is the representative of the index of top left corner ...
% within each block or cube
ind_1st_element = ( unique([1:step1:Sz1-s1+1, Sz1-s1+1])' +...
                    unique([0:step2:Sz2-s2, Sz2-s2])*Sz1 )+...
                    reshape(ind_1st_element_depth, [1 1 length(ind_1st_element_depth)]); 
          
% computing the row indices of partitions
ind_plus_rows = permute((ind_1st_element(:)+(0:s1-1))', [1 3 2]); 


% computing the row-and-column indices of the partition
ind_both_rows_cols = reshape( ind_plus_rows +(0:s2-1)*Sz1, s1*s2, 1, []); 

% indices of partitioned elements 
all_parsing_indices = ind_both_rows_cols + Sz1*Sz2*(0:s3-1);
% indices of partitioned elements (vectorized)
all_parsing_indices = reshape(all_parsing_indices, [], size(all_parsing_indices,3));

% extract the vectorized partitioned image from the computed indices
vec_parsed_image = full_image(all_parsing_indices);
end






%%
%% Estimating standard deviation using interquantile range
%%



function stdEst = stdEstIQR(Z,dim,q1,q2)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% stdEstIQR is the interquantile range estimator of the root-PSD applied
% over the "dim" dimension of the input signal "Z" with quantiles "q1" and
% "q2".
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if ~exist('dim','var')||isempty(dim)
    dim=1;
end
if ~exist('q1','var')||isempty(q1)
    if ~exist('q2','var')||isempty(q2)
        est_type=0;
        q1=0.25;
        q2=0.75;
    else
        est_type=2;
        q1=0;
    end
else
    if ~exist('q2','var')||isempty(q2)
        est_type=0;
        q2=1-q1;
    else
        est_type=0;
    end
end

if q1==0
    est_type=2;
end
if  isreal(Z)==1
    
    
    if est_type==1
        distName='half normal';
        icdf_q1=icdf(distName,q1,0,1);
        icdf_q2=icdf(distName,q2,0,1);
        stdEst=diff(quantile(abs(Z-median(Z,2)),cat(dim,q1,q2),dim),1,dim)./(icdf_q2-icdf_q1);  % inter-quantile range of absolute
    elseif est_type==2
        distName='half normal';
        icdf_q2=icdf(distName,q2,0,1);
        stdEst=quantile(abs(Z-median(Z,2)),q2,dim)./icdf_q2;  % direct generalization of median absolute deviation
    else
        distName='normal';
        icdf_q1=icdf(distName,q1,0,1);
        icdf_q2=icdf(distName,q2,0,1);
        stdEst=diff(quantile(Z-median(Z,2),cat(dim,q1,q2),dim),1,dim)./(icdf_q2-icdf_q1);  % inter-quantile range std estimator
        %     stdEst = diff(quantile(Z,cat(dim,q1,1-q1),dim),1,dim)./diff(icdf('normal',cat(dim,q1,1-q1),0,1),1,dim);
    end
    

        %%% see Section 9.1.3 (pp. 326-327) of A. DasGupta, "Probability for Statistics and Machine Learning: Fundamentals and Advanced Topics", Springer Texts in Statistics, 2011. DOI 10.1007/978-1-4419-9634-3_9
        %%% and A. DasGupta and L. R. Haff, "Asymptotic values and expansions for the correlation between different measures of spread", Journal of Statistical Planning and Inference, vol. 136, no. 7, pp. 2197-2212, 2006. DOI 10.1016/j.jspi.2005.08.028

    
else
    if est_type==1
        distName='rayl';
        icdf_q1=icdf(distName,q1,1);
        icdf_q2=icdf(distName,q2,1);
        stdEst=diff(quantile(abs(Z-mean(Z,2)),cat(dim,q1,q2),dim),1,dim)./(icdf_q2-icdf_q1);  % inter-quantile range of absolute
    else
        distName='rayl';
        icdf_q2=icdf(distName,q2,1);
        stdEst=quantile(abs(Z-mean(Z,2)),q2,dim)./icdf_q2;  % direct generalization of median absolute deviation   
    end
    
end



end


%% Auxiliary function for creating the sparsifying transform basis used internally by the filter

function [Tforward, Tinverse] = GetTransformMatrix(N, transform_type, dec_levels)
%
% Create forward and inverse transform matrices, which allow for perfect
% reconstruction. The forward transform matrix is normalized so that the
% l2-norm of each basis element is 1.
%
% [Tforward, Tinverse] = getTransfMatrix (N, transform_type, dec_levels)
%
%  INPUTS:
%
%   N               --> Size of the transform (for wavelets, must be 2^K)
%
%   transform_type  --> 'dct', 'dst', 'hadamard', or anything that is
%                       listed by 'help wfilters' (bi-orthogonal wavelets)
%                       'DCrand' -- an orthonormal transform with a DC and all
%                       the other basis elements of random nature
%
%   dec_levels      --> If a wavelet transform is generated, this is the
%                       desired decomposition level. Must be in the
%                       range [0, log2(N)-1], where "0" implies
%                       full decomposition.
%
%  OUTPUTS:
%
%   Tforward        --> (N x N) Forward transform matrix
%
%   Tinverse        --> (N x N) Inverse transform matrix
%

if exist('dec_levels','var') ~= 1
    dec_levels = 0;
end

if N == 1
    Tforward = 1;
elseif strcmpi(transform_type, 'hadamard') == 1
    Tforward    = hadamard(N);
elseif (N == 8) && strcmpi(transform_type, 'bior1.5')==1 % hardcoded transform so that the wavelet toolbox is not needed to generate it
    Tforward =  [ 0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274;
        0.219417649252501   0.449283757993216   0.449283757993216   0.219417649252501  -0.219417649252501  -0.449283757993216  -0.449283757993216  -0.219417649252501;
        0.569359398342846   0.402347308162278  -0.402347308162278  -0.569359398342846  -0.083506045090284   0.083506045090284  -0.083506045090284   0.083506045090284;
        -0.083506045090284   0.083506045090284  -0.083506045090284   0.083506045090284   0.569359398342846   0.402347308162278  -0.402347308162278  -0.569359398342846;
        0.707106781186547  -0.707106781186547                   0                   0                   0                   0                   0                   0;
        0                   0   0.707106781186547  -0.707106781186547                   0                   0                   0                   0;
        0                   0                   0                   0   0.707106781186547  -0.707106781186547                   0                   0;
        0                   0                   0                   0                   0                   0   0.707106781186547  -0.707106781186547];
elseif (N == 8) && strcmpi(transform_type, 'dct')==1 % hardcoded transform so that the signal processing toolbox is not needed to generate it
    Tforward = [ 0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274;
        0.490392640201615   0.415734806151273   0.277785116509801   0.097545161008064  -0.097545161008064  -0.277785116509801  -0.415734806151273  -0.490392640201615;
        0.461939766255643   0.191341716182545  -0.191341716182545  -0.461939766255643  -0.461939766255643  -0.191341716182545   0.191341716182545   0.461939766255643;
        0.415734806151273  -0.097545161008064  -0.490392640201615  -0.277785116509801   0.277785116509801   0.490392640201615   0.097545161008064  -0.415734806151273;
        0.353553390593274  -0.353553390593274  -0.353553390593274   0.353553390593274   0.353553390593274  -0.353553390593274  -0.353553390593274   0.353553390593274;
        0.277785116509801  -0.490392640201615   0.097545161008064   0.415734806151273  -0.415734806151273  -0.097545161008064   0.490392640201615  -0.277785116509801;
        0.191341716182545  -0.461939766255643   0.461939766255643  -0.191341716182545  -0.191341716182545   0.461939766255643  -0.461939766255643   0.191341716182545;
        0.097545161008064  -0.277785116509801   0.415734806151273  -0.490392640201615   0.490392640201615  -0.415734806151273   0.277785116509801  -0.097545161008064];
elseif (N == 8) && strcmpi(transform_type, 'dst')==1 % hardcoded transform so that the PDE toolbox is not needed to generate it
    Tforward = [ 0.161229841765317   0.303012985114696   0.408248290463863   0.464242826880013   0.464242826880013   0.408248290463863   0.303012985114696   0.161229841765317;
        0.303012985114696   0.464242826880013   0.408248290463863   0.161229841765317  -0.161229841765317  -0.408248290463863  -0.464242826880013  -0.303012985114696;
        0.408248290463863   0.408248290463863                   0  -0.408248290463863  -0.408248290463863                   0   0.408248290463863   0.408248290463863;
        0.464242826880013   0.161229841765317  -0.408248290463863  -0.303012985114696   0.303012985114696   0.408248290463863  -0.161229841765317  -0.464242826880013;
        0.464242826880013  -0.161229841765317  -0.408248290463863   0.303012985114696   0.303012985114696  -0.408248290463863  -0.161229841765317   0.464242826880013;
        0.408248290463863  -0.408248290463863                   0   0.408248290463863  -0.408248290463863                   0   0.408248290463863  -0.408248290463863;
        0.303012985114696  -0.464242826880013   0.408248290463863  -0.161229841765317  -0.161229841765317   0.408248290463863  -0.464242826880013   0.303012985114696;
        0.161229841765317  -0.303012985114696   0.408248290463863  -0.464242826880013   0.464242826880013  -0.408248290463863   0.303012985114696  -0.161229841765317];
elseif strcmpi(transform_type, 'dct') == 1
    Tforward    = dct(eye(N));
elseif strcmpi(transform_type, 'fft') == 1
    Tforward    = fft(eye(N));    
elseif strcmpi(transform_type, 'dst') == 1
    Tforward    = dst(eye(N));
elseif strcmpi(transform_type, 'eye') == 1
    Tforward    = eye(N);    
elseif strcmpi(transform_type, 'DCrand') == 1
    x = randn(N); x(1:end,1) = 1; [Q,~] = qr(x);
    if (Q(1) < 0)
        Q = -Q;
    end
    Tforward = Q';
else %% a wavelet decomposition supported by 'wavedec'
    %%% Set periodic boundary conditions, to preserve bi-orthogonality
    dwtmode('per','nodisp');
    
    Tforward = zeros(N,N);
    for i = 1:N
        Tforward(:,i)=wavedec(circshift([1 zeros(1,N-1)],[dec_levels i-1]), log2(N), transform_type);  %% construct transform matrix
    end
end

%%% Normalize the basis elements
if strcmpi(transform_type, 'fft') == 0
    Tforward = (Tforward' * diag(sqrt(1./sum(Tforward.^2,2))))';
end
%%% Compute the inverse transform matrix
Tinverse = inv(Tforward);

end



%% Visualization

function CubeFaceVis(rootPSD)
%%% visualizing the outer and inner faces of a 3D cube


% extrapolation of 3D cube w.r.t its outer faces (top, right, front)
rPSD=zeros(size(rootPSD)+1);
rPSD(1:end-1,1:end-1,1:end-1)=rootPSD;
rPSD(end,1:end-1,1:end-1)=rootPSD(end,:,:);
rPSD(1:end-1,end,1:end-1)=rootPSD(:,end,:);
rPSD(:,:,end)=rPSD(:,:,end-1);
rPSD(end,end,:)=rPSD(end-1,end-1,:);
rPSD = permute(rPSD,[2 1 3]);

cMap = parula; %jet, hot


fResponse=@(x) log10(x); invfResponse=@(x) 10.^x;
rPSD_scale = fResponse(rPSD);

[xx1, xx2, xx3]=size(rPSD);
[xx, yy, zz] = meshgrid(1:xx1,1:xx2,1:xx3);


sq123{1}=[0 0 1 1 0;0 1 1 0 0;0 0 0 0 0];
sq123{2}=[0 0 1 1 0;0 1 1 0 0;1 1 1 1 1];
sq123{3}=sq123{1}([1 3 2],:);
sq123{4}=sq123{2}([1 3 2],:);
sq123{5}=sq123{1}([3 1 2],:);
sq123{6}=sq123{2}([3 1 2],:);
sPidx=[inf inf 1];
sPidxB=[0 0 1];

figure(777),


for j=1:2
    
    %maximizing the screen
%     set(groot, 'defaultFigureWindowState', 'maximized'); 
    %    set(get(handle(gcf),'JavaFrame'),'Maximized',1);
    
    if j==1
        ax1=subplot(1,2,j,'align');
        % inner faces of cube (left, rear, bottom)
        xslice = 1;
        yslice = 1;
        zslice = 1;
        hold on
        mHandle=plot3(1.2,1.2,1.2,'k.');
        set(mHandle,'markerSize',28);
    elseif j==2 
        ax2=subplot(1,2,j,'align');
        % outer faces of cube (right, front, top)
        xslice = xx1;
        yslice = xx2;
        zslice = xx3;
        hold on
        mHandle=plot3(size(rPSD,1),size(rPSD,2),size(rPSD,3),'k.');
        set(mHandle,'markerSize',28);
    end
    
    slice(xx,yy,zz,rPSD_scale,xslice,yslice, zslice);
    axis vis3d;
    
    axis equal
    axis off
    
    xticks([]);
    yticks([]);
    zticks([]);
    
    view([120 25]);
    set(gcf,'color',[1,1,1])
    colormap(cMap)
    cb=colorbar;
    
    if size(rPSD,3)>15
       cbTicks = 7;   
    else
       cbTicks = 5; 
    end
    
    cbL=get(cb,'Limits');
    set(cb,'ticks',linspace(cbL(1),cbL(2),cbTicks));
    set(cb,'TickLabels',num2str(invfResponse(linspace(min(rPSD_scale(:)),max(rPSD_scale(:)),cbTicks)'),2))
    

    if j==1
         set(cb,'Visible','off');
        for jj=1:3
            pHandle=patch(max(sq123{1}(1,:)*size(rPSD,1),1),max(sq123{1}(2,:)*size(rPSD,2),1),max(sq123{1}(3,:)*size(rPSD,3),1),1);
            set(pHandle,'faceColor','none','EdgeColor',[0 0 0],'EdgeAlpha',1,'LineWidth',3);
            pHandle=patch(max(sq123{1}(1,:)*size(rPSD,1)+sPidxB(1)*size(rPSD,1),1),max(sq123{1}(2,:)*size(rPSD,2)+sPidxB(2)*size(rPSD,2),1),max(sq123{1}(3,:)*size(rPSD,3)+sPidxB(3)*size(rPSD,3),1),1);
            set(pHandle,'faceColor','none','EdgeColor',[0 0 0],'EdgeAlpha',1,'LineWidth',.7);
            pHandle=patch(max(sq123{1}(1,:)*size(rPSD,1),1),max(sq123{1}(2,:)*size(rPSD,2),1),min(max(sq123{1}(3,:)*size(rPSD,3),1),1),1);
            set(pHandle,'faceColor','none','EdgeColor',[0 0 0],'EdgeAlpha',1,'LineWidth',3);
            pFaceAlpha=.7;
            set(pHandle,'FaceAlpha',pFaceAlpha);
            pO=[2 3 1];
            
            for kk=1:6
                sq123{kk}=sq123{kk}(pO,:);
            end
            sPidx=sPidx(pO);
            sPidxB=sPidxB(pO);
        end
    elseif j==2
        for jj=1:3
            pHandle=patch(max(sq123{1}(1,:)*size(rPSD,1),1),max(sq123{1}(2,:)*size(rPSD,2),1),max(sq123{1}(3,:)*size(rPSD,3),1),1);
            set(pHandle,'faceColor','none','EdgeColor',[0 0 0],'EdgeAlpha',1);
            pHandle=patch(max(sq123{1}(1,:)*size(rPSD,1)+sPidxB(1)*size(rPSD,1),1),max(sq123{1}(2,:)*size(rPSD,2)+sPidxB(2)*size(rPSD,2),1),max(sq123{1}(3,:)*size(rPSD,3)+sPidxB(3)*size(rPSD,3),1),1);
            set(pHandle,'faceColor','none','EdgeColor',[0 0 0],'EdgeAlpha',1,'LineWidth',3);
            pHandle=patch(max(sq123{1}(1,:)*size(rPSD,1)+sPidxB(1)*size(rPSD,1),1),max(sq123{1}(2,:)*size(rPSD,2)+sPidxB(2)*size(rPSD,2),1),max(sq123{1}(3,:)*size(rPSD,3)+sPidxB(3)*size(rPSD,3),1),1);
            set(pHandle,'faceColor','none','EdgeColor',[0 0 0],'EdgeAlpha',1,'LineWidth',3);
            pFaceAlpha=.7;
            set(pHandle,'FaceAlpha',pFaceAlpha);
            pO=[2 3 1];
            
            for kk=1:6
                sq123{kk}=sq123{kk}(pO,:);
            end
            sPidx=sPidx(pO);
            sPidxB=sPidxB(pO);
        end
    end
    
    if j==1
        cbP1=get(ax1,'Position');
        cbP1=cbP1.*[1.5 1.5 1.5 1]+[-.1 -.2 0 0];
        set(ax1,'Position',cbP1);
    else
        cbP1=get(ax2,'Position');
        cbP1=cbP1.*[1.5 1.5 1.5 1]+[-.4 -.2 0 0];
        set(ax2,'Position',cbP1);
    end
    
    L={'\xi_1','\xi_2','\xi_3'};
    tHandle=text(0,0,size(rPSD,3)+1,L{3});
    set(tHandle,'fontsize',22);
    tHandle=text(0,size(rPSD,2)+1,0,L{2});
    set(tHandle,'fontsize',22);
    tHandle=text(size(rPSD,1)+1.5,0,0,L{1});
    set(tHandle,'fontsize',22);
    
    
    if j==1
        tHandle=text(0,-2,1.3*size(rPSD,3),'LLL corner');
        set(tHandle,'fontsize',18);
    else
        tHandle=text(0,-2,1.3*size(rPSD,3),'HHH corner');
        set(tHandle,'fontsize',18);
    end
    
    
    cbP=get(cb,'Position');
    cbP=cbP.*[.8 .8 .5 0.45]+[0.15 cbP(4)/3.5 0 0];
    set(cb,'Position',cbP);
    set(cb,'FontSize',16)
    
end
end
