function sigma_hat = Est_std(z,kernel_size,Tr,estimator,Pr1,Pr2)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Est_std estimates the white Gaussian noise standard deviation (std) in the
%  "Tr" domain, i.e. AWGN flat Tr-root-PSD, by applying a robust estimator  
%  such as the median absolute deviation (MAD) over the highest frequency
%  components of the Tr-spectra of the noisy data "z".
%
%
%  The details of the algorithm can be found in:
%  N. Eslahi and A. Foi, "Improved sparse signal recovery via adaptive 
%  correlated noise modeling", submitted to IEEE Trans. Comput. Imag. (2022) 
%
%
%  FUNCTION INTERFACE:
% 
%  INPUTS:
%     1) z                (2D/3D array) : real-/complex-valued noisy data (or noise)
%     2) kernel_size      (double)      : the size of support of
%                                           the sparsifying transform "Tr"
%     3) Tr               (cell/char)   : sparsifying transform corresponding to 
%                                         those used internally by the filter
%     4) estimator        (char)        : 'MAD'  --> median absolute deviation
%                                       : 'sstd' --> sample standard deviation
%                                       : 'IQR'  --> interquantile range
%                                         (default is 'MAD')
%     5) Pr1              (double)      : Pr1-quantile (default is 0)
%     6) Pr2              (double)      : Pr2-quantile (default is 0.5)
%
%
%  OUTPUTS:
%     1) sigma_hat        (double)      : the std/scale of z in "Tr" domain
%
%
%  EXAMPLES:
%--------------------------------------------------------------------------
%      real-valued AWGN:           
%
%        x         = double(imread('cameraman.tif'));
%        eta       = randn(size(x)); eta = (eta-mean(eta(:)))./std(eta(:));
%        eta       = 16.65 .* eta + 7.6;
%        z         = x + eta;
%        sigma_hat = Est_std(z,[8 8],{'bior1.5','bior1.5'}); % estimation by MAD
%--------------------------------------------------------------------------
%      complex-valued AWGN:        
%
%        addpath([fileparts(pwd) '/Data'])
%        load('NBA.mat'); 
%%%        load mristack;  x = double(mristack); clear mristack %% note that data itself is manipulated by some noise
%        eta1       = randn(size(x)); eta1 = (eta1-mean(eta1(:)))./std(eta1(:));
%        eta2       = randn(size(x)); eta2 = (eta2-mean(eta2(:)))./std(eta2(:));
%        eta3       = 3*eta1 + 4*1i*eta2 +randn;
%        z          = x + eta3;
%        sigma_hat  = Est_std(z,[8 8 9],{'dct','dct','dct'});
%--------------------------------------------------------------------------
%      Rician noie:                
%      
%        addpath([fileparts(pwd) '/Data'])
%        load('NBA.mat'); 
%%%        load mristack;  x = double(mristack); clear mristack %% note that data itself is manipulated by some noise
%        eta1       = randn(size(x)); eta1 = (eta1-mean(eta1(:)))./std(eta1(:));
%        eta2       = randn(size(x)); eta2 = (eta2-mean(eta2(:)))./std(eta2(:));
%        eta3       = 3*eta1 + 4*1i*eta2 + randn;
%        z          = sqrt( real(x+eta3).^2 + imag(x+eta3).^2 );
%        sigma_hat  = Est_std(z,[8 8 9],{'dct','dct','dct'})*sqrt(2);
%--------------------------------------------------------------------------
%
%               AUTHORS:  Nasser Eslahi & Alessandro Foi
%                 contact:  firstname.lastname@tuni.fi
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Copyright (c) 2006-2022 Tampere University.
% All rights reserved.
% This work (software, material, and documentation) shall only be used for nonprofit 
% noncommercial purposes.
% Any unauthorized use of this work for commercial or for-profit purposes is prohibited.
%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Parameter Settings

if ~exist('kernel_size','var')||isempty(kernel_size)
    kernel_size = [8 8];
end

if ~exist('Tr','var')||isempty(Tr)
    if numel(kernel_size)==2
        Tr  = {'dct', 'dct'};         
    else
        Tr  = {'dct', 'dct', 'dct'};  
    end 
end

if ~exist('estimator','var')||isempty(estimator)
    estimator='MAD';
end

if ~exist('Pr1','var')||isempty(Pr1)
    Pr1=0; %equal to MAD
end

if ~exist('Pr2','var')||isempty(Pr2)
    Pr2=0.5; %equal to MAD
end


Tr_1      = GetTransformMatrix(kernel_size(1), Tr{1,1});
kernel_s1 = Tr_1(end,:);
kernel_s1 = kernel_s1-mean(kernel_s1(:));         % make kernel zero-mean
kernel_s1 = kernel_s1/sqrt(sum(kernel_s1(:).^2)); % normalize ell2


if mod(kernel_size(1),2)==0
    theta = convn(padarray(z,(kernel_size(1)-1),'pre','circular'),kernel_s1','valid');
else
    theta = convn(padarray(z,(kernel_size(1)-1)/2,'circular'),kernel_s1','valid');
end

if numel(kernel_size)>=2
    Tr_2      = GetTransformMatrix(kernel_size(2), Tr{1,2});
    kernel_s2 = conj(Tr_2(end,:));
    kernel_s2 = kernel_s2-mean(kernel_s2(:));         % make kernel zero-mean
    kernel_s2 = kernel_s2/sqrt(sum(kernel_s2(:).^2)); % normalize ell2
    if mod(kernel_size(2),2)==0
        theta = permute(convn(padarray(permute(theta,[2 1 3]),(kernel_size(2)-1),'pre','circular'),kernel_s2','valid'),[2 1 3]);
    else
        theta = permute(convn(padarray(permute(theta,[2 1 3]),(kernel_size(2)-1)/2,'circular'),kernel_s2','valid'),[2 1 3]);
    end
end
 

if numel(kernel_size)==3
    theta     = reshape(theta,[size(theta,1)*size(theta,2) size(theta,3)]);
    theta     = theta';
    Tr_3      = GetTransformMatrix(kernel_size(3), Tr{1,3});
    kernel_s3 = Tr_3(end,:);
    kernel_s3 = kernel_s3-mean(kernel_s3(:));
    if mod(kernel_size(3),2)==0
        theta = convn(padarray(theta,(kernel_size(3)-1),'pre','circular'),kernel_s3','valid');
    else
        theta = convn(padarray(theta,(kernel_size(3)-1)/2,'circular'),kernel_s3','valid');
    end
    theta = reshape(theta,size(z));
end



if strcmpi(estimator, 'MAD') == 1
    if isreal(theta)==1                      % MAD of real variates
        sigma_hat = median(abs(theta(:)-median(theta(:))))*1.482602218505601;  % 1/icdf('half normal',0.5,0,1)
    else                                     % MAD of complex variates
%%%             sigma_hat = sqrt(2)*median(abs(theta(:)-mean(theta(:))))*0.849321800288019;  % 1/icdf('rayl',0.5,1)  or  1/icdf('Rician',0.5,0,1)
        sigma_hat_r = median(abs(real(theta(:))-median(real(theta(:)))))*1.482602218505601;
        sigma_hat_i = median(abs(imag(theta(:))-median(imag(theta(:)))))*1.482602218505601;
        sigma_hat = sqrt(sigma_hat_r.^2+sigma_hat_i.^2);
    end
    
elseif  strcmpi(estimator, 'sstd') == 1
%%%     sigma_hat = sqrt(mean(abs(theta(:)-mean(theta(:))).^2));
    sigma_hat = sqrt(var(theta(:)));
    
elseif strcmpi(estimator, 'IQR') == 1
    sigma_hat = stdEstIQR(theta(:),1,Pr1, Pr2);
    
end

end



%%
%% Estimate standard deviation using inter-quantile range
%%

function stdEst = stdEstIQR(Z,dim,q1,q2)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% stdEstIQR is the interquantile range estimator of the standard deviation
% applied over the "dim" dimension of the input signal "Z" with quantiles 
% "q1" and "q2".
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if ~exist('dim','var')||isempty(dim)
    dim=1;
end
if ~exist('q1','var')||isempty(q1)
    if ~exist('q2','var')||isempty(q2)
        est_type=0;
        q1=0.25;
        q2=0.75;
    else
        est_type=2;
        q1=0;
    end
else
    if ~exist('q2','var')||isempty(q2)
        est_type=0;
        q2=1-q1;
    else
        est_type=0;
    end
end

if q1==0
    est_type=2;
end


if  isreal(Z)==1
    
    
    if est_type==1
        distName='half normal';
        icdf_q1=icdf(distName,q1,0,1);
        icdf_q2=icdf(distName,q2,0,1);
        stdEst=diff(quantile(abs(Z-median(Z)),cat(dim,q1,q2),dim),1,dim)./(icdf_q2-icdf_q1);  % inter-quantile range of absolute
    elseif est_type==2
        distName='half normal';
%         q1=0;
%         icdf_q1=0;
        icdf_q2=icdf(distName,q2,0,1);
        stdEst=quantile(abs(Z-median(Z)),q2,dim)./icdf_q2;  % direct generalization of median absolute deviation
    else
        distName='normal';
        icdf_q1=icdf(distName,q1,0,1);
        icdf_q2=icdf(distName,q2,0,1);
        stdEst=diff(quantile(Z-median(Z),cat(dim,q1,q2),dim),1,dim)./(icdf_q2-icdf_q1);  % inter-quantile range std estimator
    end
    
else
 
    if est_type==1
        distName='rayl';
        icdf_q1=icdf(distName,q1,1);
        icdf_q2=icdf(distName,q2,1);
        stdEst=diff(quantile(abs(Z-mean(Z)),cat(dim,q1,q2),dim),1,dim)./(icdf_q2-icdf_q1);  % inter-quantile range of absolute
    else
        distName='rayl';
%         q1=0;
%         icdf_q1=0;
        icdf_q2=icdf(distName,q2,1);
        stdEst=quantile(abs(Z-mean(Z)),q2,dim)./icdf_q2;  % direct generalization of median absolute deviation   
    end
    
end

end



%% Auxiliary function for creating the sparsifying transform basis used internally by the filter

function [Tforward, Tinverse] = GetTransformMatrix(N, transform_type, dec_levels)
%
% Create forward and inverse transform matrices, which allow for perfect
% reconstruction. The forward transform matrix is normalized so that the
% l2-norm of each basis element is 1.
%
% [Tforward, Tinverse] = getTransfMatrix (N, transform_type, dec_levels)
%
%  INPUTS:
%
%   N               --> Size of the transform (for wavelets, must be 2^K)
%
%   transform_type  --> 'dct', 'dst', 'hadamard', or anything that is
%                       listed by 'help wfilters' (bi-orthogonal wavelets)
%                       'DCrand' -- an orthonormal transform with a DC and all
%                       the other basis elements of random nature
%
%   dec_levels      --> If a wavelet transform is generated, this is the
%                       desired decomposition level. Must be in the
%                       range [0, log2(N)-1], where "0" implies
%                       full decomposition.
%
%  OUTPUTS:
%
%   Tforward        --> (N x N) Forward transform matrix
%
%   Tinverse        --> (N x N) Inverse transform matrix
%

if exist('dec_levels','var') ~= 1
    dec_levels = 0;
end

if N == 1
    Tforward = 1;
elseif strcmpi(transform_type, 'hadamard') == 1
    Tforward    = hadamard(N);
elseif (N == 8) && strcmpi(transform_type, 'bior1.5')==1 % hardcoded transform so that the wavelet toolbox is not needed to generate it
    Tforward =  [ 0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274;
        0.219417649252501   0.449283757993216   0.449283757993216   0.219417649252501  -0.219417649252501  -0.449283757993216  -0.449283757993216  -0.219417649252501;
        0.569359398342846   0.402347308162278  -0.402347308162278  -0.569359398342846  -0.083506045090284   0.083506045090284  -0.083506045090284   0.083506045090284;
        -0.083506045090284   0.083506045090284  -0.083506045090284   0.083506045090284   0.569359398342846   0.402347308162278  -0.402347308162278  -0.569359398342846;
        0.707106781186547  -0.707106781186547                   0                   0                   0                   0                   0                   0;
        0                   0   0.707106781186547  -0.707106781186547                   0                   0                   0                   0;
        0                   0                   0                   0   0.707106781186547  -0.707106781186547                   0                   0;
        0                   0                   0                   0                   0                   0   0.707106781186547  -0.707106781186547];
elseif (N == 8) && strcmpi(transform_type, 'dct')==1 % hardcoded transform so that the signal processing toolbox is not needed to generate it
    Tforward = [ 0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274   0.353553390593274;
        0.490392640201615   0.415734806151273   0.277785116509801   0.097545161008064  -0.097545161008064  -0.277785116509801  -0.415734806151273  -0.490392640201615;
        0.461939766255643   0.191341716182545  -0.191341716182545  -0.461939766255643  -0.461939766255643  -0.191341716182545   0.191341716182545   0.461939766255643;
        0.415734806151273  -0.097545161008064  -0.490392640201615  -0.277785116509801   0.277785116509801   0.490392640201615   0.097545161008064  -0.415734806151273;
        0.353553390593274  -0.353553390593274  -0.353553390593274   0.353553390593274   0.353553390593274  -0.353553390593274  -0.353553390593274   0.353553390593274;
        0.277785116509801  -0.490392640201615   0.097545161008064   0.415734806151273  -0.415734806151273  -0.097545161008064   0.490392640201615  -0.277785116509801;
        0.191341716182545  -0.461939766255643   0.461939766255643  -0.191341716182545  -0.191341716182545   0.461939766255643  -0.461939766255643   0.191341716182545;
        0.097545161008064  -0.277785116509801   0.415734806151273  -0.490392640201615   0.490392640201615  -0.415734806151273   0.277785116509801  -0.097545161008064];
elseif (N == 8) && strcmpi(transform_type, 'dst')==1 % hardcoded transform so that the PDE toolbox is not needed to generate it
    Tforward = [ 0.161229841765317   0.303012985114696   0.408248290463863   0.464242826880013   0.464242826880013   0.408248290463863   0.303012985114696   0.161229841765317;
        0.303012985114696   0.464242826880013   0.408248290463863   0.161229841765317  -0.161229841765317  -0.408248290463863  -0.464242826880013  -0.303012985114696;
        0.408248290463863   0.408248290463863                   0  -0.408248290463863  -0.408248290463863                   0   0.408248290463863   0.408248290463863;
        0.464242826880013   0.161229841765317  -0.408248290463863  -0.303012985114696   0.303012985114696   0.408248290463863  -0.161229841765317  -0.464242826880013;
        0.464242826880013  -0.161229841765317  -0.408248290463863   0.303012985114696   0.303012985114696  -0.408248290463863  -0.161229841765317   0.464242826880013;
        0.408248290463863  -0.408248290463863                   0   0.408248290463863  -0.408248290463863                   0   0.408248290463863  -0.408248290463863;
        0.303012985114696  -0.464242826880013   0.408248290463863  -0.161229841765317  -0.161229841765317   0.408248290463863  -0.464242826880013   0.303012985114696;
        0.161229841765317  -0.303012985114696   0.408248290463863  -0.464242826880013   0.464242826880013  -0.408248290463863   0.303012985114696  -0.161229841765317];
elseif strcmpi(transform_type, 'dct') == 1
    Tforward    = dct(eye(N));
elseif strcmpi(transform_type, 'dst') == 1
    Tforward    = dst(eye(N));
elseif strcmpi(transform_type, 'fft') == 1
    Tforward    = fft(eye(N));     
elseif strcmpi(transform_type, 'DCrand') == 1
    x = randn(N); x(1:end,1) = 1; [Q,~] = qr(x);
    if (Q(1) < 0)
        Q = -Q;
    end
    Tforward = Q';
else %% a wavelet decomposition supported by 'wavedec'
    %%% Set periodic boundary conditions, to preserve bi-orthogonality
    dwtmode('per','nodisp');
    
    Tforward = zeros(N,N);
    for i = 1:N
        Tforward(:,i)=wavedec(circshift([1 zeros(1,N-1)],[dec_levels i-1]), log2(N), transform_type);  %% construct transform matrix
    end
end

%%% Normalize the basis elements
%%% Normalize the basis elements
if strcmpi(transform_type, 'fft') == 0
    Tforward = (Tforward' * diag(sqrt(1./sum(Tforward.^2,2))))';
end

%%% Compute the inverse transform matrix
Tinverse = inv(Tforward);

end
