
function [loc_FFT_f_rate, loc_FFT_rootPSD_pinvM_eps] = PSD_Auxiliaries(samp_strat, D, locWin, std_eps, y)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  prerequisite | 1)  pinvM is the pseudo-inverse of the sampling matrix/operator M, 
%  information  |     and perpM is the orthogonal projection onto the null space of M.
%     on the    | 
%   notations   | 2)  we use "pinvM",  "pinvM_M"    and "perpM" as prefix/suffix
%  and the name |     to denote variables which go respectively
%  of variables |     under   pinvM(.), pinvM(M(.)) and  perpM(.) = . - pinvM(M(.)) operation. 
%  used in this |     For instance, we shall name: 
%    piece of   |         pinvM_M_eta = pinvM(M(eta));  %if M is an operator, or 
%      code     |         pinvM_M_eta = pinv(M)*M*eta;  %if M is a matrix
%               |
%               | 3) "loc" stands for local.
% -------------------------------------------------------------------------
%
%  We approximate the PSD of effective noise (i.e. e_k = z_k - x = b_k + x_k -x ) as 
%  sum of the PSDs of surrogate noise (i.e. s_k = b_k + pinvM(M(x_k - x)) ) and
%  latent noise (i.e. l_k = perpM(x_k - x) = x_k - x - pinvM(M(x_k - x)) ).
%
%  To this aim, we approximate the PSD of latent noise through multiplying 
%  the approximated PSD of pinvM(M(x_k - x)) with a PSD compensation ratio.
%
%  "CompensationVectors" computes the local FFT-PSD compensation ratio, 
%   and the local FFT-root-PSD of back-projected measurement noise.
%
% -------------------------------------------------------------------------
%  For pinvM(M(.)), we can compute an array (denoted by loc_f_pinvM_M in
%  "CompensationVectors"), such that for an AWGN "eta" with a flat local FFT-PSD
%  denoted by loc_FFT_PSD_eta, we have:
%             loc_FFT_PSD_pinvM_M_eta = loc_FFT_PSD_eta .* loc_f_pinvM_M; 
%  where loc_FFT_PSD_pinvM_M_eta denotes the local FFT-PSD of pinvM_M_eta.
%  Therefore, an approximate of loc_FFT_PSD_eta can be computed as
%             approx_loc_FFT_PSD_eta   = loc_FFT_PSD_pinvM_M_eta ./ loc_f_pinvM_M; 
%  We dub 
%                 inv_loc_f_pinvM_M   = 1./loc_f_pinvM_M
%  the local FFT-PSD compensation vector of pinvM_M .
%
%  Accordingly, for perpM, we have:
%             loc_FFT_PSD_perpM_eta    = loc_FFT_PSD_eta .*loc_f_perpM;
%  where loc_FFT_PSD_perpM_eta denotes the local FFT-PSD of perpM_eta.
%  Therefore, an approximate of loc_FFT_PSD_eta can be computed as
%             approx_loc_FFT_PSD_eta   = loc_FFT_PSD_perpM_eta ./ loc_f_perpM 
%  We dub
%                 inv_loc_f_perpM      = 1./loc_f_perpM;
%  the local FFT-PSD compensation vector of perpM .
%
%  The local FFT-PSD compensation ratio is then
%         loc_FFT_f_rate  = loc_f_perpM./loc_f_pinvM_M; 
%                         = inv_loc_f_pinvM_M./inv_loc_f_perpM;
% ---------------------------------------------------------------------
%  
%   More details can be found in:
%   N. Eslahi and A. Foi, "Improved sparse Signal Recovery via Adaptive Correlated Noise Model",
%     submitted to IEEE TCI, 2022.
%
%
%  FUNCTION INTERFACE:
%   [loc_FFT_f_rate, loc_FFT_rootPSD_pinvM_eps] = PSD_Auxiliaries(samp_strat, D, locWin, std_eps, y)
%
%
% _________________________________________________________________________________________________
% INPUTS:                     |  TYPE:                  |    DESCRIPTION:
% -------------------------------------------------------------------------------------------------
%   samp_strat:               |(char)                   |  sampling strategy for each image/modality. 
%                             |                         |  Options       : 
%                             |                         | 'radial'       : radial line sampling 
%                             |                         |                  (of 2D FFT spectrum).
%                             |                         | 'lim_angle'    : limited angle sampling 
%                             |                         |                  (of 2D FFT spectrum).
%                             |                         | 'pseudo-random' OR
%                             |                         | 'pseudo-random': pseudo random sampling 
%                             |                         |                  (of 2D FFT spectrum).
%                             |                         | '3D_radial'    : 3D radial line sampling 
%                             |                         |                  (of 3D FFT spectrum).
%                             |                         | 'CSI'          : spectral multiplexing
%                             |                         |                  (To obtain single-shot 
%                             |                         |                   spectrally compressed measurements 
%                             |                         |                   given hyperspectral images).
%                             |                         | 'CTI'          : temporal multiplexing
%                             |                         |                  (To obtain multi-chunk temporally
%                             |                         |                   compressed measurements given a video).
% -------------------------------------------------------------------------------------------------
%   D:                        |(double/logical)         |  The sampling mask based on the sampling  
%                             |                         |  strategy "samp_strat". 
%                             |                         |  ** If "D" is not provided (i.e. it is set 
%                             |                         |  to "D=[]") and the sampling strategy is 
%                             |                         |  that of FFT spectrum subsampling, then "D" 
%                             |                         |  will be computed from the location of 
%                             |                         |  measurements "y" in the FFT plane.
% -------------------------------------------------------------------------------------------------
%   std_eps:                  |(double)                 |  standard deviation of the measurement error
%                             |                         | (used for computing the PSD of back-projected 
%                             |                         |  measurment error). 
%                             |                         |  ** If "std_eps" is not provided (i.e. it is set  
%                             |                         |  to "std_eps=[]"), the algorithm internally  
%                             |                         |  estimates it from given "y".
% -------------------------------------------------------------------------------------------------
%   locWin:                   |(double)                 |  size of blocks/cubes for local FFT-PSD estimation.
% -------------------------------------------------------------------------------------------------
%   y:                        |(double)                 |  measurements obtained based on the sampling 
%                             |                         |  strategy "samp_strat". 
%                             |                         |  **It is used to compute the std of measurement  
%                             |                         |  error "std_eps" (and/or to compute the binary  
%                             |                         |  FFT subsampling mask "D") if it is not provided.
%                             |                         |  ** we can set "y =[]" if "D" and "std_eps" are 
%                             |                         |  both provided.
%
%
%
% _________________________________________________________________________________________________
% OUTPUTS:                    |  TYPE:                  |    DESCRIPTION:
% -------------------------------------------------------------------------------------------------
%  loc_FFT_f_rate:            |(double)                 |  local FFT-PSD compensation vector ratio,   
%                             |                         |  computed from the sampling mask.
%                             |                         |  It is used to approximate the local FFT-PSD
%                             |                         |  of latent noise.
% -------------------------------------------------------------------------------------------------
%  loc_FFT_rootPSD_pinvM_eps: |(double)                 |  local FFT-root-PSD of the back-projected   
%                             |                         |  measurement error, computed from the  
%                             |                         |  sampling mask and the standard deviation 
%                             |                         |  of the measurement error.
%
%   EXTERNAL DEPENDENCY FUNCTIONS:
%    1) FFT_PSD_domain_convertor.m (used for samp_strat: 'radial', 'lim_angle', 'pseudo-random', 'pseudo_random', '3D_radial').            
%    2) Est_std.m  (used if std_eps=[]).
%      
%
%
%            Authors :         Nasser Eslahi & Alessandro Foi
%                                   Tampere University
%        Contact info:          firstname.lastname@tuni.fi
%
%                          Release ver. 1.0  (April 2022)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Copyright (c) 2006-2022 Tampere University.
% All rights reserved.
% This work (software, material, and documentation) shall only be used for nonprofit 
% noncommercial purposes.
% Any unauthorized use of this work for commercial or for-profit purposes is prohibited.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% check-shit to see everything is given properly!
%%% check-shit begins!

samp_modes  = {'radial', 'lim_angle', 'pseudo-random',...
                'pseudo_random', '3D_radial', ...
                'CSI', 'CTI'}; % available sampling modes (Don't touch!)
                     
check_strat = cellfun(@(c) strcmpi(c,samp_strat), samp_modes, ...
                         'UniformOutput',false);

checked_idx = sum(cell2mat(check_strat.'),1);


if checked_idx==0 
    error(['Your chosen sampling strategy ''%s'' does not exist.\n' ...
           'Please check the <a href="matlab:help %s">syntax</a>'], samp_strat, mfilename);
end



if ~exist('D','var')|| isempty(D)
    if sum(cell2mat(check_strat(1:5).'),1) && ~isempty(y)
        
        sprintf(['Since "D" (which is the binary sampling matrix of the FFT spectrum \n' ...
            'in the selected sampling strategy ''%s'') is not given, it is computed \n' ...
            'from the location of measurements "y" in the FFT plane. \n' ...
            'Please check the <a href="matlab:help %s">syntax</a>'], samp_strat, mfilename)

        D = true(y~=0);
        
    elseif sum(cell2mat(check_strat(1:5).'),1) && isempty(y)
        error('u:stuffed:it', ['Neither the sampling matrix "D" nor the measurements "y" are given! \n' ...
               'Please give either "D" or "y".'])
    else
        error(['The sampling matrix "D" should be given if the sampling strategy ' ...
               'is either of ''CSI'' and ''CTI''.'])
    end
end



if ~exist('std_eps','var')|| isempty(std_eps)
    sprintf(['Since "std_eps" (which is the standard deviation of the measurement error, \n' ...
            'distributed as i.i.d. Gaussian) is not given, the algorithm computes it \n' ...
            'internally from measurements "y". \n' ...
            'Please check the <a href="matlab:help %s">syntax</a>'], mfilename)
        
    estimate_std_eps = true;
else
    estimate_std_eps = false;
end



if ~exist('locWin','var')||isempty(locWin)
    if sum(cell2mat(check_strat(1:4).'),1)
        locWin  = [15 15];       % block size in the experiments of 2D tomography

    elseif cell2mat(check_strat(5))
        locWin  = [15 15 15];    % cube size in the experiments of 3D tomography
        
    elseif cell2mat(check_strat(6))
        locWin  = [31 31 13];    % cube size in the experiments of compressive spectral imaging
        
    elseif cell2mat(check_strat(7))
        locWin  = [31 31];       % block size in the experiments of compressive temporal imaging
    end
end

%%% check-shit ends!


FFT_Tr      = repelem({'fft'},size(locWin,2)); % the 2D/3D FFT for computing the local PSDs
         


if sum(cell2mat(check_strat(1:5).'),1)  
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%% when "M" is subsampling of FFT spectrum %%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
    
    Sz             = size(D);  % size of sampling mask
    pinvD_D        = D; pinvD_D(D~=0) = 1;  % = pinv(mask)*mask
    
    glob_f_pinvM_M = pinvD_D;       % the reciprocal of the global FFT-PSD compensation vector for pinvM(M(.))
    glob_f_perpM   = 1 - pinvD_D;   % the reciprocal of the global FFT-PSD compensation vector for perpM(.) = . - pinvM(M(.))
    
    %%% converting the reciprocal of global FFT-PSD compensation vectors to their local form
    loc_f_pinvM_M  = FFT_PSD_domain_convertor(glob_f_pinvM_M, locWin, FFT_Tr)...
                            .* (prod(Sz)/prod(locWin)); %% reciprocal of PSD compensation vector of pinvM(M(.))
                   
    loc_f_perpM    = FFT_PSD_domain_convertor(glob_f_perpM,    locWin, FFT_Tr)...
                            .* (prod(Sz)/prod(locWin)); %% reciprocal of PSD compensation vector of perpM(.)
    
    %%% the ratio between the local FFT-PSD compensation ratio
    loc_FFT_f_rate = loc_f_perpM./loc_f_pinvM_M; 
    loc_FFT_f_rate(isinf(loc_FFT_f_rate)) = 0; 
    loc_FFT_f_rate(isnan(loc_FFT_f_rate)) = 0; 

    if estimate_std_eps
        %%% if the measurement noise std ("std_eps") is not provided we can ...
        %%% estimate it as the following.
        
        pinvM = @(X) ifftn(X./(D+eps));

        %%% estimating the highest-frequency power of the noise
        Est_Tr   = repelem({'dct'},  size(locWin,2)); % basis for noise std estimation
        sigma    = Est_std(pinvM(y), ceil(locWin/2), Est_Tr, 'MAD'); % estimated noise std
        
        %%% compensation vector for the DCT-root-PSD
        Est_Win  = repelem(min(ceil(min(locWin)/2),8),  size(locWin,2));   
        dct_f_pinvM_M = sqrt( FFT_PSD_domain_convertor(glob_f_pinvM_M, Est_Win, Est_Tr)*prod(Sz) );
        
        %%% empirically estimated standard deviation of the measurement error
        std_eps  = (sigma/dct_f_pinvM_M(end))*sqrt(prod(Sz)); % estimated measurement error std
    end

    %%% analytically computing the local FFT-root-PSD of the back-projected measurement error ... 
    %%% "pinvM_eps = pinvM(real(epsilon))"
    loc_FFT_rootPSD_pinvM_eps = (std_eps ./sqrt(2)).*sqrt(loc_f_pinvM_M)*sqrt(prod(locWin)/prod(Sz));


% -------------------------------------------------------------------------------------------------  
      %%%% NOTE 1.1: "loc_FFT_rootPSD_pinvM_eps" can also be computed empirically from 
      %%%% multiple realizations of i.i.d. Gaussian noise generated with the same "std_eps" 
      %%%% (i.e. std of the measurement error) followed by computing the sample std (sstd)
      %%%% using a moving window over the realizations of "pinvM(eps)".
      %%%% Pragmatically, one can do as the following in MATLAB:
%
%  NumR = 100;                       %% number of realizations
%  n    = randn([Sz NumR]);          %% i.i.d. Gaussian noise
%  n    = (n-mean(n(:)))./std(n(:)); %% Standardization of noise
%  All_eps_hat = M( (std_eps/sqrt(prod(Sz))).*n ); %% computing the realizations of measurement error
%  pinvM_All_eps_hat = real( pinvM(All_eps_hat) ); %% because we filter the real part of noisy image ...
%                                                  %% note that M and pinvM are separately applied ...
%                                                  %% to each realizations of their respective input.
%  loc_FFT_rootPSD_pinvM_eps = sqrt(NumR)*Est_rootPSD(pinvM_All_eps_hat, ...
%                                             locWin, FFT_Tr, 0, 'sstd'); %% computing the FFT-root-PSD
%     
%
%
      %%%% NOTE 1.2: we alternatively can compute "loc_FFT_rootPSD_pinvM_eps" through ...
      %%%% downsampling the empirically computed global FFT-PSD of pinvM(eps).
      %%%% In MATLAB we can compute it as following:
% 
%  NumR = 100;                       %% number of realizations
%  n    = randn([Sz NumR]);          %% i.i.d. Gaussian noise
%  n    = (n-mean(n(:)))./std(n(:)); %% Standardization of noise
%  All_eps_hat = M( (std_eps/sqrt(prod(Sz))).*n ); %% computing the realizations of measurement error
%  pinvM_All_eps_hat = real( pinvM(All_eps_hat) ); %% because we filter the real part of noisy image ...
%                                                  %% note that M and pinvM are separately applied ...
%                                                  %% to each realizations of their respective input.
%  dim_rlz = ndims(pinvM_All_eps_hat); %% size of all back-projected noise realizations
%  FFT_pinvM_All_eps_hat  = ifft(fftn(pinvM_All_eps_hat) ,[], dim_rlz); %% FFT of each realization
%  glob_FFT_PSD_pinvM_eps = NumR*var(FFT_pinvM_All_eps_hat ,[], dim_rlz); %% global FFT-PSD of pinvM(eps)
%  loc_FFT_PSD_pinvM_eps  = ...
%          FFT_PSD_domain_convertor(glob_FFT_PSD_pinvM_eps, locWin, FFT_Tr); %% local FFT-PSD of pinvM(eps)
%  loc_FFT_rootPSD_pinvM_eps  = sqrt(loc_FFT_PSD_pinvM_eps); %% local FFT-root-PSD of pinvM(eps)
%     
%
%
      %%%% Note 1.3: "loc_f_pinvM_M" (resp. "loc_f_perpM") can also be computed empirically ...
      %%%% through generating multiple realizations of an i.i.d. Gaussian noise "e" and ... 
      %%%% then computing the mean of the ratios between the local FFT-PSDs ...
      %%%% of "pinvM(M(eta))" (resp. "perpM(eta)") and "eta".
      %%%% In MATLAB we can do as following:
%       
%  NumR = 100;                     %% number of realizations
%  eta  = randn([Sz NumR]);        %% i.i.d. Gaussian noise
%  eta  = (eta-mean(eta(:)))./std(eta(:)); %% Standardization of noise
%  pinvM_M_eta = pinvM(M(eta));   %% computing the realizations of measurement error
%  loc_FFT_rootPSD_pinvM_M_eta = sqrt(NumR)*Est_rootPSD(real(pinvM_M_eta), ...
%                                           locWin, FFT_Tr, 0, 'sstd'); %% computing the FFT-root-PSD of ...
%                                                                       %% the real part of pinvM_M_eta ...
%                                                                       %% (since we filter only this part)                                     
%  loc_FFT_PSD_pinvM_M_eta = loc_FFT_rootPSD_pinvM_M_eta.^2;
%  loc_f_pinvM_M = loc_FFT_PSD_pinvM_M_eta./prod(locWin); 
%
% ------------------------------------------------------------------------------------------------- 
 

    
elseif cell2mat(check_strat(6)) %%% 'CSI' %%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%% when "M" is spectral multiplexing %%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
    
          Est_loc_FFT_PSD = @(X) fast_Est_loc_FFT_PSD(X, locWin);
% %% %%   The above version of "Est_loc_FFT_PSD" is much faster than ...
% %% %%   the below version, due to using non-overlapping cube extraction.
% % % %   FFT_Tr  = repelem({'fft'},size(locWin,2));
% % % %   Est_loc_FFT_PSD = @(X) (Est_rootPSD(X, locWin, FFT_Tr, 0, 'sstd')).^2;

    save_dir = sprintf('./Data/compensation_vectors_CSI_%dx%dx%d.mat',locWin(1),locWin(2),locWin(3));
    
    if isfile(save_dir) 
    %%% loading the precomputed local FFT-PSD compensation ratio
         load(save_dir,'loc_FFT_f_rate')
        
    elseif isfile(['.',save_dir])
    %%% loading the precomputed local FFT-PSD compensation ratio
             load(['.',save_dir],'loc_FFT_f_rate')

    else
    %%% computing the local FFT-PSD compensation ratio   
        
        computing_approach = 1;
        
        rng(0,'v4');
        
        switch computing_approach
            
            case 1    %%% empirical estimation approach
            
                fprintf(['The local FFT-root-PSD of the back-projected measurement error and \n' ...
                         'the local FFT-PSD compensation ratio are being computed. \n']);
                
                M     = @(X) sum(X.*D,3);             % forward sampling operator
                pinvM = @(X) (X./max(sum(D,3),1)).*D; % pseudo-inverse of M
                NumR  = 50; % number of realizations
                loc_f_pinvM_M = 0;
                loc_f_perpM   = 0;
           
                for i=1:NumR
                    eta         = randn(size(D)); % i.i.d. Gaussian noise
                    eta         = (eta-mean(eta(:)))./std(eta(:)); % Standardization of noise
                    pinvM_M_eta = pinvM(M(eta));
                    perpM_eta   = eta - pinvM_M_eta;
                    clearvars eta
                    
                    
                    loc_f_pinvM_M = loc_f_pinvM_M + Est_loc_FFT_PSD(pinvM_M_eta);
                    loc_f_perpM   = loc_f_perpM   + Est_loc_FFT_PSD(perpM_eta);
  
                    
                    % % if std_eps~=0
                    % %     %%% computing the local FFT-root-PSD of the back-projected measurement error "pinvM_eps = pinvM(epsilon)"
                    % %     epsilon   = randn(size(D,1), size(D,2)); % i.i.d. Gaussian noise
                    % %     epsilon   = (epsilon-mean(epsilon(:)))./std(epsilon(:)); % Standardization of noise
                    % %     pinvM_eps = pinvM(epsilon);
                    % % 
                    % %     var_FFT_pinvM_eps = var_FFT_pinvM_eps + Est_loc_FFT_PSD(pinvM_eps);
                    % % end
                    
      
                end
                
                loc_f_pinvM_M  = loc_f_pinvM_M/NumR;
                loc_f_perpM    = loc_f_perpM/NumR;
                
                
                % % %%% the local FFT-root-PSD of the back-projected measurement error "pinvM_eps = pinvM(epsilon)"
                % % loc_FFT_rootPSD_pinvM_eps = std_eps.*sqrt(var_FFT_pinvM_eps/NumR);
                

                
            
            case 2    %%% Fast empirical approximation approach
                
                %%% Due to the stationary assumption of the noise and the structure of ...
                %%% the sliding mask, we can compute the local FFT-PSD compensation ...
                %%% vectors over a small version of the sampling operator (i.e. we ...
                %%% crop the sampling mask to the size of local 3D FFT-basis, and ...
                %%% create a smaller version of the sampling operator "M".) This is ...
                %%% due to avoid the computational burden of global FFT-PSD compensation.
                
                fprintf(['The local FFT-root-PSD of the back-projected measurement error and \n' ...
                         'the local FFT-PSD compensation ratio are being approximated. \n']);
                
                %%% cropping the 3D sampling mask to the size of local 3D FFT basis
                cropped_mask = D(1:locWin(1), 1:locWin(2), 1:locWin(3));
                %%% the smaller version of forward sampling operator ...
                %%% and its pseudo-inverse
                M_crop     = @(X) sum(X.*cropped_mask,3); 
                pinvM_crop = @(X) (X./max(sum(cropped_mask,3),1)).*cropped_mask;
                NumR = 1000;    % number of realizations
                eta  = randn([locWin NumR]); % i.i.d. Gaussian noise
                eta  = (eta-mean(eta(:)))./std(eta(:)); % Standardization of noise
                
                pinvM_M_eta = pinvM_crop(M_crop(eta));
                perpM_eta   = eta - pinvM_M_eta; 
                clearvars eta
                
                FFT_pinvM_M_eta = ifft(fftn(pinvM_M_eta),[],4); % 3D FFT over all 3D realizations
                FFT_perpM_eta   = ifft(fftn(perpM_eta),[],4); % 3D FFT over all 3D realizations
                clearvars pinvM_M_eta  perpM_eta
                
                loc_f_pinvM_M   = var(FFT_pinvM_M_eta,[],4);
                loc_f_perpM     = var(FFT_perpM_eta,[],4);
                
                
               
                % % %%% computing the local FFT-root-PSD of the back-projected measurement error "pinvM_eps = pinvM(epsilon)"
                % % if std_eps~=0
                % %     epsilon   = randn([locWin(1:2) 1 NumR]); % i.i.d. Gaussian noise
                % %     epsilon   = (epsilon-mean(epsilon(:)))./std(epsilon(:)); % Standardization of noise
                % %     pinvM_eps = pinvM_crop(epsilon);
                % %     clearvars epsilon
                % %     FFT_pinvM_eps = ifft(fftn(pinvM_eps),[],4);
                % %     loc_FFT_rootPSD_pinvM_eps = std_eps.* std(FFT_pinvM_eps,[],4);
                % % else
                % %     loc_FFT_rootPSD_pinvM_eps = 0;
                % % end
                
                
            case 3    %%% Analytical approximation approach
                
                %%% Due to the stationary assumption of the noise and the structure of ...
                %%% the sliding mask, we can compute the local FFT-PSD compensation ...
                %%% vectors over a small version of the sampling operator (i.e. we ...
                %%% crop the sampling mask to the size of local 3D FFT-basis, and ...
                %%% create a smaller version of the sampling operator "M".) This is ...
                %%% due to avoid the computational burden of global FFT-PSD compensation.
        
                fprintf(['The local FFT-root-PSD of the back-projected measurement error and \n' ...
                         'the local FFT-PSD compensation ratio are being approximated. \n' ...
                         '         Keep Calm! This might take a while. \n']);
                
                
                %%% cropping the 3D sampling mask to the size of local 3D FFT basis
                cropped_mask = D(1:locWin(1), 1:locWin(2), 1:locWin(3));
                cropped_mask = reshape(cropped_mask,[prod(locWin(1:2)) locWin(3)]);
                %%% computing the smaller version of 2D sampling matrix corresponding ...
                %%% to the cropped 3D sampling mask
                M_crop       = zeros(prod(locWin(1:2)), prod(locWin));
                for i=1:prod(locWin(1:2))
                    M_crop(i,(i-1)*locWin(3)+1:i*locWin(3)) = cropped_mask(i,:);
                end
                clear cropped_mask D i
                
                
                colon         = @(x) x(:);
                %%% computing the matrix corresponding to the local 3D IFFT
                loc_3D_IFFT_basis = eye(prod(locWin));
                for i=1:prod(locWin)
                    loc_3D_IFFT_basis(:,i) = colon(permute(ifftn(reshape(loc_3D_IFFT_basis(:,i), locWin)),[3 1 2]));
                end
                
                %%% computing the local FFT-PSD compensation of M^{\dagger}M
                pinvM_M_crop      = pinv(M_crop)*M_crop;
                inv_loc_f_pinvM_M = diag(pinv(diag(sum(abs(loc_3D_IFFT_basis'*pinvM_M_crop).^2,2)))); % this might take a while!
                inv_loc_f_pinvM_M = inv_loc_f_pinvM_M/numel(inv_loc_f_pinvM_M);
                inv_loc_f_pinvM_M = reshape(inv_loc_f_pinvM_M, locWin);
                loc_f_pinvM_M     = 1./inv_loc_f_pinvM_M;
                
                
                %%% computing the local FFT-PSD compensation of M_{\perp} = I - M^{\dagger}M
                perpM_crop      = eye(size(pinvM_M_crop)) - pinvM_M_crop;
                inv_loc_f_perpM = diag(pinv(diag(sum(abs(loc_3D_IFFT_basis'*perpM_crop).^2,2)))); % this might take a while!
                inv_loc_f_perpM = inv_loc_f_perpM/numel(inv_loc_f_perpM);
                inv_loc_f_perpM = reshape(inv_loc_f_perpM, locWin);
                loc_f_perpM     = 1./inv_loc_f_perpM;
        
        end
        
        %%% the ratio between the local FFT-PSD compensation vectors of "M^{\dagger}M" and "M_{\perp}"
        loc_FFT_f_rate  = loc_f_perpM./loc_f_pinvM_M;
        loc_FFT_f_rate(isinf(loc_FFT_f_rate)) = 0;
        loc_FFT_f_rate(isnan(loc_FFT_f_rate)) = 0;
        
        

        if isfolder('./Data')
            save(save_dir,'loc_f_pinvM_M','loc_f_perpM','loc_FFT_f_rate')
            
        elseif isfolder('../Data')
            save(['.',save_dir],'loc_f_pinvM_M','loc_f_perpM','loc_FFT_f_rate')
        else
            mkdir Data
            save(save_dir,'loc_f_pinvM_M','loc_f_perpM','loc_FFT_f_rate')
        end
        fprintf('The local FFT-PSD compensation ratio is saved in \n%s\n',...
            save_dir);
        
    end
    
    
    if estimate_std_eps
        %%% if the measurement noise std ("std_eps") is not provided we can ...
        %%% estimate it as the following.
        
        %%% estimating the highest-frequency power of the noise
        Est_Tr  = {'dct', 'dct'}; % basis for noise std estimation
        std_eps = Est_std(y, ceil(locWin/2), Est_Tr, 'MAD'); % estimated measurement error std
    end
    
    %%% computing the local FFT-root-PSD of the back-projected measurement error "pinvM_eps = pinvM(epsilon)"
    if std_eps~=0
        pinvM = @(X) (X./max(sum(D,3),1)).*D; % pseudo-inverse of M
        NumR  = 5; % number of realizations
        var_FFT_pinvM_eps = 0;
%         Est_loc_FFT_PSD = @(X) (Est_rootPSD(X, locWin, FFT_Tr, 0, 'sstd')).^2;
        for i=1:NumR
            
            epsilon   = randn(size(D,1), size(D,2)); % i.i.d. Gaussian noise
            epsilon   = (epsilon-mean(epsilon(:)))./std(epsilon(:)); % Standardization of noise
            pinvM_eps = pinvM(epsilon); % a realization of the back-projected standard AWGN

            var_FFT_pinvM_eps = var_FFT_pinvM_eps + Est_loc_FFT_PSD(pinvM_eps);
            
        end
       
         loc_FFT_rootPSD_pinvM_eps = std_eps.*sqrt(var_FFT_pinvM_eps/NumR);
    else
         loc_FFT_rootPSD_pinvM_eps = 0;
    end
    
    
    
    
    
elseif cell2mat(check_strat(7)) %%% 'CTI' %%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%% when "M" is temporal multiplexing %%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
    
          Est_loc_FFT_PSD = @(X) fast_Est_loc_FFT_PSD(X, locWin);
% %% %%   The above version of "Est_loc_FFT_PSD" is much faster than ...
% %% %%   the below version, due to using non-overlapping cube extraction.
% % % %   FFT_Tr  = repelem({'fft'},size(locWin,2));
% % % %   Est_loc_FFT_PSD = @(X) (Est_rootPSD(X, locWin, FFT_Tr, 0, 'sstd')).^2;
    
    save_dir = sprintf('./Data/compensation_vectors_CTI_%dx%d.mat',locWin(1),locWin(2));    
    if isfile(save_dir)
        %%% loading the precomputed local FFT-PSD compensation ratios for RND and FPN components
         load(save_dir,'loc_FFT_f_RND','loc_FFT_f_FPN')
        
    elseif isfile(['.',save_dir])
        %%% loading the precomputed local FFT-PSD compensation ratios for RND and FPN components
        load(['.',save_dir],'loc_FFT_f_RND','loc_FFT_f_FPN')      
    else
        %%% empirically computing the local FFT-PSD compensation ratios for RND and FPN components
        fprintf(['The local FFT-root-PSDs of RND and FPN components of the back-projected measurement error and \n' ...
                 'and local FFT-PSD compensation ratios for RND and FPN components are being computed. \n']);
        
       
        chunkSz = size(D,3);
        if exist('y','var')
            NumMeas = size(y,3);
        else
            NumMeas = 4;
        end
        
        loc_f_perpM_FPN   = 0; 
        loc_f_pinvM_M_FPN = 0;
        loc_f_perpM_RND   = 0; 
        loc_f_pinvM_M_RND = 0;
        
        NumR = 100; % number of realizations
        for i = 1:NumR

            RND = randn([size(D,1), size(D,2), chunkSz*NumMeas]);
            RND = (RND-mean(RND(:)))./(std(RND(:))); % standard white Gaussian noise
            
            FPN = randn([size(D,1), size(D,2)]);
            FPN = (FPN-mean(FPN(:)))./(std(FPN(:)));
            
            eta = RND + FPN; clear RND FPN
            
            M     = @(X) sum(X.*D,3);             % forward sampling for each chunk of the video
            pinvM = @(X) (X./max(sum(D,3),1)).*D; % pseudo-inverse operator of "M"
            
            pinvM_M_eta = zeros(size(eta));  % the pseudo-inverse of temporally compressed white noise "eta"
            for t = 1:NumMeas
                pinvM_M_eta(:,:,(t-1)*chunkSz+(1:chunkSz)) = pinvM( M( eta(:,:,(t-1)*chunkSz+(1:chunkSz)) ) );
            end
            pinvM_M_eta_FPN = mean(pinvM_M_eta,3);           % the FPN component of pinvM_M_eta
            pinvM_M_eta_RND = pinvM_M_eta - pinvM_M_eta_FPN; % the RND component of pinvM_M_eta
            
            perpM_eta       = eta - pinvM_M_eta;
            perpM_eta_FPN   = mean(perpM_eta,3);         % the FPN component of perpM_eta
            perpM_eta_RND   = perpM_eta - perpM_eta_FPN; % the RND component of perpM_eta
            clear eta  pinvM_M_eta  perpM_eta
            
            
            %%% computing the 2D FFT PSDs over via a local moving window approach
            loc_f_pinvM_M_FPN = loc_f_pinvM_M_FPN + Est_loc_FFT_PSD(pinvM_M_eta_FPN); clear pinvM_M_eta_FPN
            loc_f_pinvM_M_RND = loc_f_pinvM_M_RND + Est_loc_FFT_PSD(pinvM_M_eta_RND); clear pinvM_M_eta_RND
            loc_f_perpM_FPN = loc_f_perpM_FPN + Est_loc_FFT_PSD(perpM_eta_FPN); clear perpM_eta_FPN
            loc_f_perpM_RND = loc_f_perpM_RND + Est_loc_FFT_PSD(perpM_eta_RND); clear perpM_eta_RND
        end
        
        loc_f_pinvM_M_FPN = loc_f_pinvM_M_FPN/NumR;
        loc_f_pinvM_M_RND = loc_f_pinvM_M_RND/NumR;
        loc_f_perpM_FPN   = loc_f_perpM_FPN/NumR;
        loc_f_perpM_RND   = loc_f_perpM_RND/NumR;
         
        %%% local FFT-PSD compensation ratios for FPN component
        loc_FFT_f_FPN = loc_f_perpM_FPN./loc_f_pinvM_M_FPN;
        loc_FFT_f_FPN(isinf(loc_FFT_f_FPN)) = 0;
        loc_FFT_f_FPN(isnan(loc_FFT_f_FPN)) = 0;
        
        %%% local FFT-PSD compensation ratios for RND component
        loc_FFT_f_RND = loc_f_perpM_RND./loc_f_pinvM_M_RND;
        loc_FFT_f_RND(isinf(loc_FFT_f_RND)) = 0;
        loc_FFT_f_RND(isnan(loc_FFT_f_RND)) = 0;

        
        if isfolder('./Data')
            save(save_dir,'loc_FFT_f_FPN', 'loc_FFT_f_RND')
            
        elseif isfolder('../Data')
            save(['.',save_dir],'loc_FFT_f_FPN', 'loc_FFT_f_RND')
            
        else
            mkdir Data
            save(save_dir,'loc_FFT_f_FPN', 'loc_FFT_f_RND')
        end
        
        fprintf('The local FFT-PSD compensation ratios for RND and FPN components are saved in \n%s\n',...
            save_dir);
        
    end
    
    
    
    loc_FFT_f_rate        = zeros([locWin,2]);
    loc_FFT_f_rate(:,:,1) = loc_FFT_f_FPN;
    loc_FFT_f_rate(:,:,2) = loc_FFT_f_RND;

    
    
    %%%%% Estimating the local FFT-root-PSDs of RND and FPN components ...
    %%%%% within the back-projected (pseudo-inverse) of measurement error

    if estimate_std_eps
        %%% if the measurement noise std ("std_eps") is not provided we can ...
        %%% estimate it as the following.
        
        %%% estimating the highest-frequency power of the noise
        Est_Tr  = {'dct', 'dct', 'dct'}; % basis for noise std estimation
        std_eps = Est_std(y, ceil(locWin/2), Est_Tr, 'MAD'); % estimated measurement error std

    end

    %%% synthetic error with the same statistics as the measurement error
    if exist('y','var')
        eps_hat = randn(size(y));
        NumMeas = size(y,3);
    else
        NumMeas = 4;
        eps_hat = randn(size(D,1),size(D,2),NumMeas);
    end   
    eps_hat = (eps_hat-mean(eps_hat(:)))./(std(eps_hat(:))); % standard white Gaussian noise
    eps_hat = std_eps.*eps_hat;                                     

    
    pinvM   = @(X) (X./max(sum(D,3),1)).*D; % the pseudo-inverse operator of "M"
    chunkSz = size(D,3);

    pinvM_eps_hat  = zeros([size(D,1), size(D,2), NumMeas*chunkSz]);
    for t = 1:NumMeas
        % pseudo-inverse of the synthetic error for computing its root-PSDs
        pinvM_eps_hat(:,:,(t-1)*chunkSz+(1:chunkSz)) = pinvM( eps_hat(:,:,t) );
    end

    %%% estimating FFT root-PSDs of estimated "FPN" and "RND" components within "pinvM_eps_hat"   
    loc_FFT_rootPSD_pinvM_eps_FPN = sqrt( Est_loc_FFT_PSD(mean(pinvM_eps_hat,3)) );
    loc_FFT_rootPSD_pinvM_eps_RND = sqrt( Est_loc_FFT_PSD(pinvM_eps_hat - mean(pinvM_eps_hat,3)) );
    clear pinvM_eps_hat
    
    
    loc_FFT_rootPSD_pinvM_eps        = zeros(locWin);
    loc_FFT_rootPSD_pinvM_eps(:,:,1) = loc_FFT_rootPSD_pinvM_eps_FPN;
    loc_FFT_rootPSD_pinvM_eps(:,:,2) = loc_FFT_rootPSD_pinvM_eps_RND;
    
    
end



end








%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%  auxiliary function for extracting the blocks/cubes  %%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



function loc_FFT_PSD = fast_Est_loc_FFT_PSD(noise, locWin)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% fast_Est_loc_FFT_PSD rapidly computes the local 2D/3D FFT-PSD of the ...
% given input "noise" as the sample variance of its local FFT spectrum ...
% computed over a moving window of size "locWin", as each position of ...
% the window can be treated as a different realizations of the noise.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if size(noise,3)~=1 && (numel(locWin)==3 && locWin(3)~=1)
    %%% extracting cubes of size "locWin" from "noise"
    noise = PARSER( noise, locWin, [locWin(1:2) 1]);
    %%% reshaping (grouping) the extracted 3D cubes into a 4D array
    noise = reshape(noise, [locWin  size(noise,2)]);
    
    %%% computing the 3D FFT spectrum of all extracted cubes and then ...
    %%% computing the sample variance of 3D FFT spectra over the grouping (4th) dimension
    loc_FFT_PSD = var( ifft(fftn(noise),[],4) ,[],4);
    
elseif  numel(locWin)==2 || (numel(locWin)==3 && locWin(3)==1)
    if size(noise,3)~=1
        %%% extracting blocks of size "locWin" from "noise"
        noise = PARSER( noise, locWin, [locWin 1]);
    elseif size(noise,3)==1
        %%% extracting blocks of size "locWin" from "noise"
        noise = PARSER( noise, locWin, ceil(locWin/2));
    end
    %%% reshaping (grouping) the extracted 2D blocks into a 3D array
    noise = reshape(noise, [locWin  size(noise,2)]);
    
    %%% computing the 2D FFT spectrum of all extracted cubes and then ...
    %%% computing the sample variance of 2D FFT spectra over the grouping (3rd) dimension
    loc_FFT_PSD =  var( fft2(noise) ,[] ,3);
    
end      



function vec_parsed_image = PARSER( full_image, partition_size, step_size )
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PARSER extracts 2D blocks or 3D cubes of size "partition_size" from ...
% the input signal "full_image" with step size "step_size"
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if ~ismember(ndims(full_image),[2 3])
    error('the input image should be either 2D or 3D!')
end

if ~ismember(size(partition_size,2),[2 3])
    error('our partitioning is either block-based or cube-based!')
end

if nargin == 2 || isempty(step_size)
   step_size = max(round(partition_size./2),1); 
end

if ischar(step_size)
   if strcmpi(step_size,'distinct') % non-overlapping parsing
       step_size = partition_size;
       
   elseif strcmpi(step_size,'sliding')  % fully-overlapped parsing
       step_size = ones(1,size(partition_size,2));
       
   else
       error('available options for "step_size": ''%s'' and ''%s''', 'distinct', 'sliding')
       
   end
       
end


if size(step_size,2) == 2
   step_size(3) = 1; 
end

if size(partition_size,2) == 2
   partition_size(3) = 1; 
end


if  size(step_size,2) ~= size(partition_size,2)
    error('please check the step-size!')
end



s1    = partition_size(1); % horizontal size of partition
s2    = partition_size(2); % vertical size of partition
s3    = partition_size(3); % depth size of partition (if S3=1 then partitioning is block-based, otherwise cube-based)

step1 = step_size(1);      % step-size for sliding across horizontal dimension
step2 = step_size(2);      % step-size for sliding across vertical dimension
step3 = step_size(3);      % step-size for sliding across depth

[Sz1,Sz2,Sz3] = size(full_image);

% indices of the first element in the depth dimension, ...
% starting from 0 which indicates the first slice
ind_1st_element_depth = unique([0:step3:Sz3-s3, Sz3-s3])*Sz1*Sz2;

% indices of the first element within each partition (block or cube), ...
% where each index is the representative of the index of top left corner ...
% within each block or cube
ind_1st_element = ( unique([1:step1:Sz1-s1+1, Sz1-s1+1])' +...
                    unique([0:step2:Sz2-s2, Sz2-s2])*Sz1 )+...
                    reshape(ind_1st_element_depth, [1 1 length(ind_1st_element_depth)]); 
          
% computing the row indices of partitions
ind_plus_rows = permute((ind_1st_element(:)+(0:s1-1))', [1 3 2]); 


% computing the row-and-column indices of the partition
ind_both_rows_cols = reshape( ind_plus_rows +(0:s2-1)*Sz1, s1*s2, 1, []); 

% indices of partitioned elements 
all_parsing_indices = ind_both_rows_cols + Sz1*Sz2*(0:s3-1);
% indices of partitioned elements (vectorized)
all_parsing_indices = reshape(all_parsing_indices, [], size(all_parsing_indices,3));

% extract the vectorized partitioned image from the computed indices
vec_parsed_image = full_image(all_parsing_indices);
end



end




